#include <stdio.h>
#include <stdlib.h>   
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>

#include "../include/os.h"
#include "../include/graphics.h"
#include "../include/string.h"

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>

#include "../include/osw-x.h"
#include "../include/MwmUtil.h"

/* Shape Extension */
#include <X11/extensions/shape.h>

/* MIT Shared Memory Extension */
#ifdef USE_XSHM
# include <X11/extensions/XShm.h>
#endif

/* XPM Library */
#ifdef HAVE_LIBXPM
# include <X11/xpm.h>
#endif

#include <X11/keysym.h>


#include "../include/cursors/drag_unaccepted.xpm"
#include "../include/cursors/drag_move.xpm"
#include "../include/cursors/drag_copy.xpm"
#include "../include/cursors/drag_link.xpm"


/* Time */
time_t OSWTimeMS(void);

/* Initialize and shutdown */
static void OSWGenerateColorLookupTable(u_int32_t BitDepth);
int OSWGUIConnect(int argc, char **argv);
void OSWGUIDisconnect(void);

/* Keycodes */
static void OSWManageKeyEvent(
	osw_gui_struct *gui,
	event_t *event
);
keycode_t OSWGetKeyCodeFromKeyVal(const keyval_t keyval);
keyval_t OSWGetKeyValFromKeyCode(
	const keycode_t keycode,
	const key_state_t key_state
);
char OSWGetASCIIFromKeyCode(
	const keycode_t keycode,
	const key_state_t key_state
);
const char *OSWGetKeyCodeName(const keycode_t keycode);
const char *OSWGetKeyValName(const keyval_t keyval);
bool_t OSWIsModifierKey(const keycode_t keycode);

/* Keyboard Autorepeat */
void OSWKBAutoRepeatOff(void);
void OSWKBAutoRepeatOn(void);
bool_t OSWGetKBAutoRepeatState(void);

/* Focus */
void OSWSetInputFocus(
	win_t w,
	const int revert,
	const time_t t
);
void OSWGrabFocusTime(
	win_t w,
	const time_t t
);
void OSWGrabFocus(win_t w);

/* Flush & Sync */
void OSWFlush(void);
void OSWSync(const bool_t discard);

/* Pointer */
void OSWGetPointerCoords(
	win_t w,
	int *root_x_rtn, int *root_y_rtn,
	int *wx_rtn, int *wy_rtn
);
int OSWGrabPointer(
	win_t grab_w,
	const bool_t events_rel_grab_w,
	const event_mask_t eventmask,
	const int pointer_mode,
	const int keyboard_mode,
	win_t confine_w,
	cursor_t cursor
);
int OSWGrabPointerTime(
	win_t grab_w,
	const bool_t events_rel_grab_w,
	const event_mask_t eventmask,
	const int pointer_mode,
	const int keyboard_mode,
	win_t confine_w,
	cursor_t cursor,
	const time_t t
);
void OSWUngrabPointer(void);
void OSWUngrabPointerTime(const time_t t);
bool_t OSWIsPointerGrabbed(void);

/* Atoms */
atom_t OSWGetAtom(const char *name);

/* Selection */
static void OSWManageSelectionEvent(
	osw_gui_struct *gui,
	event_t *event
);
static void OSWDDEReportClear(
	osw_gui_struct *gui, win_t w,
	atom_t selection
);
char *OSWDDEGetString(win_t w);
void OSWDDESetString(win_t w, const char *s);
void *OSWDDEGetData(win_t w, atom_t target, int *len_rtn);
void OSWDDESetData(win_t w, atom_t target, const void *data, int len);

/* DND */
static void OSWManageDNDEvent(osw_gui_struct *gui, event_t *event);
void OSWDNDSetWindowTarget(win_t w);
bool_t OSWDNDQueryPointer(
	win_t *w_rtn,
	int *x_rtn, int *y_rtn,
	int *root_x_rtn, int *root_y_rtn
);
atom_t OSWDNDQueryAction(
	win_t w, atom_t def_action, dnd_action_flags_t action_flags
);
atom_t *OSWDNDGetActionsList(win_t w, int *nactions);
char **OSWDNDGetActionNamesList(win_t w, int *nactions);

void OSWDNDSendEnter(
	win_t w, win_t src_w,
	const atom_t *type_list, int ntypes
);
void OSWDNDSendPosition(
	win_t w, win_t src_w, atom_t action,
	int root_x, int root_y,
	time_t t
);
void OSWDNDSendStatus(
	win_t w, win_t tar_w, atom_t action,
	const rectangle_t *rect
);
void OSWDNDSendLeave(win_t w, win_t src_w);
void OSWDNDSendDrop(win_t w, win_t src_w, time_t t);
void OSWDNDSendFinished(
	win_t w, win_t tar_w,
	bool_t drop_processed_and_successful,
	atom_t action
);

void OSWDNDBegin(
	win_t src_w,
	atom_t target,			/* Source data type */
	const u_int8_t *data, int len,	/* Source data */
	atom_t def_action,
	dnd_action_flags_t action_flags,
	int x, int y,
	cursor_t cursor, win_t icon_w,
	time_t t
);
void OSWDNDSetCursor(cursor_t cursor, time_t t);
void OSWDNDSetIcon(win_t icon_w, int x, int y, time_t t);
u_int8_t *OSWDNDGetData(
	win_t tar_w,			/* Our target window */
	atom_t target,			/* Source data type */
	int *len_rtn
);
void OSWDNDIconAnimateReturn(
	win_t icon_w,
	int cur_x, int cur_y,
	int orig_x, int orig_y
);
void OSWDNDEnd(time_t t);


/* Visual */
visual_t *OSWGetVisualByCriteria(
	visual_info_mask_t vinfo_mask,
	visual_info_t criteria_vinfo
);
visual_t *OSWGetVisualByID(visual_id_t vid);

/* Fonts */
int OSWLoadFont(font_t **font_rtn, const char *name);
font_t *OSWQueryCurrentFont(void);
bool_t OSWGetCharBounds(
	font_t *font, char c,
	int *lbearing, int *rbearing,
	int *width,
	int *ascent, int *descent
);
bool_t OSWGetTextBounds(
	font_t *font, const char *text, int text_len,
	int *lbearing, int *rbearing,
	int *width,
	int *ascent, int *descent
);
void OSWSetFont(font_t *font);
void OSWUnloadFont(font_t **font);

/* Pixels (Colors) */
int OSWLoadPixelRGB(
	pixel_t *pix_rtn,
	u_int8_t r, u_int8_t g, u_int8_t b
);
int OSWLoadPixelHSL(
	pixel_t *pix_rtn,
	u_int8_t h, u_int8_t s, u_int8_t l
);
int OSWLoadPixelCLSP(pixel_t *pix_rtn, const char *clsp);
void OSWDestroyPixel(pixel_t *pix_ptr);
void OSWSetForeground(pixel_t pix);

/* Pointer Cursors */
cursor_t OSWCursorOpenID(cursor_id_t id);
cursor_t OSWCursorOpenXPMFile(
	const char *path,
	int hot_x, int hot_y
);
cursor_t OSWCursorOpenXPMData(
	char **xpmdata,
	int hot_x, int hot_y
);
void OSWSetWindowCursor(win_t w, cursor_t cur);
void OSWUnsetWindowCursor(win_t w);
void OSWDestroyCursor(cursor_t *cur_rtn);

/* Events */
int OSWEventsPending(void);
void OSWWaitNextEvent(event_t *event);
void OSWWaitPeakEvent(event_t *event);
bool_t OSWCheckMaskEvent(
	const event_mask_t eventmask,
	event_t *event
);
void OSWWaitWindowEvent(
	win_t w,
	const event_mask_t event_mask,
	event_t *event
);
bool_t OSWCheckWindowEvent(
	win_t w, event_mask_t event_mask, event_t *event
);
void OSWPutBackEvent(event_t *event);
int OSWSendEvent(
	event_mask_t mask,
	event_t *event,
	bool_t propagate
);
int OSWPurgeAllEvents(void);
int OSWPurgeTypedEvent(const int event_type);
int OSWPurgeWindowTypedEvent(win_t w, event_type_t event_type);
static void OSWTranslateBlockInputEvent(
	osw_gui_struct *gui,
	event_t *event
);
void OSWBlockInput(const bool_t block);
int OSWBlockInputLevel(void);
bool_t OSWEventIsTakeFocus(
	win_t w,
	event_t *event,
	time_t *t_rtn
);
bool_t OSWEventIsDeleteWindow(
	win_t w,
	event_t *event
);
bool_t OSWEventIsDNDEnter(
	win_t w, event_t *event,
	win_t *src_w_rtn,
	const atom_t *target, int ntargets
);
bool_t OSWEventIsDNDPosition(
	win_t w, event_t *event,
	win_t *src_w_rtn, win_t *tar_child_w_rtn,
	int *x_rtn, int *y_rtn,
	int *root_x_rtn, int *root_y_rtn,
	time_t *t,
	atom_t *action_rtn
);
bool_t OSWEventIsDNDStatus(
	win_t w, event_t *event,
	win_t *tar_w_rtn, atom_t *accepted_action_rtn,
	rectangle_t *rect_rtn
);
bool_t OSWEventIsDNDLeave(
	win_t w, event_t *event,
	win_t *src_w_rtn
);
bool_t OSWEventIsDNDDrop(
	win_t w, event_t *event,
	win_t *src_w_rtn, time_t *t
);
bool_t OSWEventIsDNDFinished(
	win_t w, event_t *event,
	win_t *tar_w_rtn,
	bool_t *drop_processed_and_successful,
	atom_t *action
);


/* Windows */
int OSWCreateWindow(
	win_t *w, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height
);
int OSWCreateInputWindow(
	win_t *w, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height
);
void OSWDestroyWindow(win_t *w);
bool_t OSWDrawableIsWindow(drawable_t d);
bool_t OSWWindowIsToplevel(win_t w);
static void OSWSetWindowIcon(
	win_t w,
	pixmap_t icon,
	const bool_t has_transparency
);
void OSWSetWindowWMProperties(
        win_t w,
        const char *title,
        const char *icon_title,
        pixmap_t icon,
        const bool_t wm_sets_coordinates,
        const int x, const int y,
        const unsigned int min_width, const unsigned int min_height,
        const unsigned int max_width, const unsigned int max_height,
        const int frame_style,		/* One of WindowFrameStyle* */
        const int argc, char **argv
);
void OSWSetWindowInput(
	win_t w,
	const event_mask_t eventmask
);
void OSWSetTransientFor(
	win_t w,
	win_t main_w
);
void OSWSetWindowTitle(
	win_t w,
	const char *title
);
void OSWClearWindow(win_t w);
void OSWSetWindowBkg(
	win_t w,
	pixel_t pix,
	pixmap_t pixmap
);
bool_t OSWGetWindowGeometry(
	win_t w,
	int *x, int *y,
	unsigned int *width, unsigned int *height,
	unsigned int *depth
);
bool_t OSWGetWindowAttributes(
	win_t w,
	win_attr_t *wattr
);
bool_t OSWTranslateCoordinatesToChild(
	win_t parent,
	win_t child,
	const int x, const int y,
	int *x_rtn, int *y_rtn
);
bool_t OSWTranslateCoordinatesToParent(
	win_t child,
	win_t parent,
	const int x, const int y,
	int *x_rtn, int *y_rtn
);
bool_t OSWGetWindowRootOrigin(win_t w, int *x, int *y);
win_t OSWGetWindowParent(win_t w);
bool_t OSWCheckWindowAncestory(win_t grand_parent, win_t grand_child);
static void OSWTranslateModalEvent(
	osw_gui_struct *gui,
	event_t *event
);
void OSWSetModal(win_t w);
void OSWMapWindow(win_t w);
void OSWMapRaised(win_t w);
void OSWMapSubwindows(win_t w);
void OSWUnmapWindow(win_t w);
void OSWIconifyWindow(win_t w);
void OSWReparentWindow(
	win_t w,
	win_t parent
);
void OSWMoveWindow(
	win_t w,
	const int x, const int y
);
void OSWResizeWindow(
	win_t w,
	const unsigned int width, const unsigned int height
);
void OSWMoveResizeWindow(
	win_t w,
	const int x, const int y,
	const unsigned int width, const unsigned int height
);

/* Bitmap */
int OSWCreateBitmap(
	bitmap_t *bitmap,
	const unsigned int width, const unsigned int height
);
void OSWDestroyBitmap(bitmap_t *bitmap);

/* Pixmap */
int OSWCreatePixmap(
	pixmap_t *pixmap, unsigned int width, unsigned int height
);
void OSWDestroyPixmap(pixmap_t *pixmap);
bool_t OSWDrawableIsPixmap(drawable_t d);
bool_t OSWGetPixmapGeometry(
	pixmap_t pixmap,
	unsigned int *width, unsigned int *height,
	unsigned int *depth
);
bool_t OSWGetPixmapAttributes(pixmap_t pixmap, pixmap_attr_t *pattr);
void OSWClearPixmap(
	pixmap_t pixmap,
	unsigned int width, unsigned int height,
	pixel_t pix
);

/* Image */
image_t *OSWCreateImageFromXImage(XImage *ximg);
int OSWCreateImage(
	image_t **img_rtn,
	unsigned int width, unsigned int height
);
void OSWDestroyImage(image_t **img_rtn);

/* Shared Images */
int OSWCreateSharedImage(
	shared_image_t **img_rtn,
	unsigned int width, unsigned int height
);
void OSWDestroySharedImage(shared_image_t **img_rtn);

/* Graphic Conversion */
pixmap_t OSWCreatePixmapFromImage(image_t *img);
bitmap_t OSWCreateMaskFromImage(image_t *img);

/* Graphics IO */
image_t *OSWGetImage(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height
);
void OSWPutImageToDrawable(image_t *img, drawable_t d);
void OSWPutImageToDrawablePos(
	image_t *img, drawable_t d,
	int tar_x, int tar_y
);
void OSWPutImageToDrawableSect(
	image_t *img, drawable_t d,
	int tar_x, int tar_y,
	int src_x, int src_y,
	unsigned int width, unsigned int height
);
void OSWPutSharedImageToDrawable(
	shared_image_t *img, drawable_t d
);
void OSWPutSharedImageToDrawablePos(
	shared_image_t *img, drawable_t d,
	int tar_x, int tar_y
);
void OSWPutSharedImageToDrawableSect(
	shared_image_t *img, drawable_t d,
	int tar_x, int tar_y,
	int src_x, int src_y,
	unsigned int width, unsigned int height
);
void OSWSyncSharedImage(shared_image_t *img, drawable_t d);
void OSWPutDrawableToWindow(win_t w, drawable_t d);
void OSWCopyDrawables(
	drawable_t tar_d, drawable_t src_d,
	unsigned int width, unsigned int height
);
void OSWCopyDrawablesCoord(
	drawable_t tar_d, drawable_t src_d,
	int tar_x, int tar_y,
	unsigned int width, unsigned int height,
	int src_x, int src_y
);

/* Drawing */
void OSWSetClipRectangle(const rectangle_t *rect);
void OSWSetClipMask(bitmap_t mask);
void OSWSetClipOrigin(int x, int y);
void OSWDrawString(drawable_t d, int x, int y, const char *s);
void OSWDrawStringLimited(
	drawable_t d, int x, int y, const char *s, int len
);
void OSWDrawLine(
	drawable_t d,
	int start_x, int start_y,
	int end_x, int end_y
);
void OSWDrawRectangle(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height
);
void OSWDrawSolidRectangle(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height
);
void OSWDrawArc(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height,
	float position_angle,
	float terminal_angle
);
void OSWDrawSolidArc(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height,
	float position_angle,
	float terminal_angle
);

/* X Error Callbacks */
static int OSWGUIErrorHandler(Display *dpy, XErrorEvent *error_event);
static int OSWGUISimpleErrorHandler(
	Display *dpy, XErrorEvent *error_event
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
#define IS_NUM_ODD(n)	((long)(n) & 1)

osw_gui_struct	osw_gui[1];
osw_atom_struct	osw_atom;


/*
 *	XShmGetEventBase() needs to be prototyped, it is apparently not
 *	declared in any of the available header files.
 */
extern "C" int XShmGetEventBase(Display *display);


/*
 *	Returns the current time in milliseconds.
 */
time_t OSWTimeMS(void)
{
	struct timeval tv[1];

	if(gettimeofday(tv, NULL) < 0)
	    return(0l);

	return((tv->tv_sec % 86400) * 1000l + tv->tv_usec / 1000l);
}


/*
 *	Graphics blitting lookup table generator.
 *
 *	Called by OSWGUIConnect().
 */
#if 0
u_int16_t gRLookup[256];
u_int16_t gGLookup[256];
u_int16_t gBLookup[256];
#endif
static void OSWGenerateColorLookupTable(u_int32_t BitDepth)
{
#if 0
	int i;

	switch(BitDepth)
	{
	  case 15:
	    for(i = 0; i < 256; i++)
	    {
		gRLookup[i] = (i & 0xF8) << 7;
		gGLookup[i] = (i & 0xF8) << 2;
		gBLookup[i] = i >> 3;
	    }
	    break;

	  case 16:
	    for(i = 0; i < 256; i++)
	    {
		gRLookup[i] = (i & 0xF8) << 8;
		gGLookup[i] = (i & 0xFC) << 3;
		gBLookup[i] = i >> 3;
	    }
	    break;
	}
#endif
}

/*
 *	Connect to GUI.
 */
int OSWGUIConnect(int argc, char **argv)
{
	osw_gui_struct *gui;
	int i, status;
	u_int8_t r, g, b;
	visual_id_t vid = 0x00;
	char *host = NULL;
	char *fg_pix_str = NULL;
	char *bg_pix_str = NULL;
	char *def_font_name = NULL;
	display_t *dpy;
	visual_t *vis;
	win_t root;
	depth_t depth = 0;
	XColor xc;


	/* Get pointer to gui connection structure */
	gui = &osw_gui[0];

	/* Reset values */
	gui->std_font = NULL;

	gui->display_width = 0;
	gui->display_height = 0;

	gui->black_pix = 0;
	gui->white_pix = 0;

	gui->std_cursor = None;
	gui->drag_unaccepted_cursor = None;
	gui->drag_move_cursor = None;
	gui->drag_copy_cursor = None;
	gui->drag_link_cursor = None;

	gui->root_win = None;

	gui->current_font = NULL;
	gui->current_fg_pix = 0;
	gui->current_bg_pix = 0;

	gui->def_font = NULL;
	gui->def_bg_pix = 0;
	gui->def_fg_pix = 0;
	gui->def_toplevel_x = 0;
	gui->def_toplevel_y = 0;
	gui->def_toplevel_width = 0;
	gui->def_toplevel_height = 0;
	gui->def_geometry_set = False;
#ifdef USE_XSHM
	gui->def_use_shm = True;
	gui->def_sync_shm = True;	/* Default to True, so mandatory syncing */
#else
	gui->def_use_shm = False;
	gui->def_sync_shm = False;
#endif


	/* Parse arguments */
	for(i = 0; i < argc; i++)
	{
	    if(argv[i] == NULL)
		continue;

	    /* Display */
	    if(!strcmp(argv[i], "--display") ||
	       !strcmp(argv[i], "--dpy") ||
	       !strcmp(argv[i], "-display") ||
	       !strcmp(argv[i], "-dpy")
	    )
	    {
		i++;
		if(i < argc)
		{
		    host = argv[i];
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Visual */
	    else if(!strcmp(argv[i], "--visual") ||
		    !strcmp(argv[i], "-visual")
	    )
	    {
		i++;
		if(i < argc)
		{
		    sscanf(argv[i], "%ld", &vid);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Depth */
	    else if(!strcmp(argv[i], "--depth") ||
		    !strcmp(argv[i], "-depth")
	    )
	    {
		i++;
		if(i < argc)
		{
		    depth = atoi(argv[i]);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Geometry */
	    else if(!strcmp(argv[i], "--geometry") ||
		    !strcmp(argv[i], "-geometry")
	    )
	    {
		i++;
		if(i < argc)
		{
		    XParseGeometry(argv[i],
			&gui->def_toplevel_x,
			&gui->def_toplevel_y,
			&gui->def_toplevel_width,
			&gui->def_toplevel_height
		    );
		    gui->def_geometry_set = True;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Default font (not the standard font) */
	    else if(!strcmp(argv[i], "--font") ||
		    !strcmp(argv[i], "-font") ||
		    !strcmp(argv[i], "--fn") || 
		    !strcmp(argv[i], "-fn")
	    )
	    {
		i++;
		if(i < argc)
		{
		    def_font_name = argv[i];
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Background pixel */
	    else if(!strcmp(argv[i], "--background") ||
		    !strcmp(argv[i], "-background") ||
		    !strcmp(argv[i], "--bg") ||
		    !strcmp(argv[i], "-bg")
	    )
	    {
		i++;
		if(i < argc)
		{
		    bg_pix_str = argv[i];
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }           
	    /* Foreground pixel */
	    else if(!strcmp(argv[i], "--foreground") ||
		    !strcmp(argv[i], "-foreground") ||
		    !strcmp(argv[i], "--fg") ||
		    !strcmp(argv[i], "-fg")      
	    )
	    {
		i++;
		if(i < argc)
		{
		    fg_pix_str = argv[i];
		}
		else
		{   
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* No shared memory */
	    else if(!strcmp(argv[i], "--no_shm") ||
		    !strcmp(argv[i], "-no_shm") ||
		    !strcmp(argv[i], "--noshm") ||
		    !strcmp(argv[i], "-no_shm") ||
		    !strcmp(argv[i], "--no_xshm") ||
		    !strcmp(argv[i], "-no_xshm") ||
		    !strcmp(argv[i], "--noxshm") ||
		    !strcmp(argv[i], "-noxshm")
	    )
	    {
		gui->def_use_shm = False;
	    }
	    /* Auto sync shared memory puts */
	    else if(!strcmp(argv[i], "--sync_xshm") ||
		    !strcmp(argv[i], "-sync_xshm") ||
		    !strcmp(argv[i], "--sync_shm") ||
		    !strcmp(argv[i], "-sync_shm")
	    )
	    {
		gui->def_sync_shm = True;
	    }
	}


	/* Open connection to display */
	status = -1;
	while(True)
	{
	    /* First try given input host address to connect to */
	    gui->display = XOpenDisplay(host);
	    if(gui->display != NULL)
	    {
		status = 0;	/* Success */
		break;
	    }

	    /* Check if host address was NULL (unspecified) */
	    if(host == NULL)
	    {
		/* Try enviroment valriable */
	        gui->display = XOpenDisplay(getenv("DISPLAY"));
	        if(gui->display != NULL)
	        {
		    status = 0;	/* Success */
		    break;
		}

		/* All else fails, try explicitly setting localhost */
		gui->display = XOpenDisplay("127.0.0.1:0.0");
		if(gui->display != NULL)
		{
		    status = 0;	/* Success */
		    break;
		}
	    }

	    /* Failed */
	    break;
	}
	/* Connect failed? */
	if(status)
	{
	    /* Print reason for failed connect */
	    fprintf(stderr, "Unable to connect to X server: ");
	    if(getenv("DISPLAY") == NULL)
		fprintf(stderr, "`DISPLAY' environment variable not set.");
	    else
		fprintf(stderr, "%s", XDisplayName(NULL));
	    fprintf(stderr, "\n");

	    /* Print verbose help */
	    fprintf(
		stderr,
"\nCheck to make sure that your X server is running and that your\n\
enviroment variable `DISPLAY' is set properly.\n"
	    );

	    return(1);
	}
	dpy = gui->display;


	/* Get the Screen */
	gui->scr_ptr = DefaultScreenOfDisplay(dpy);
	gui->scr_num = XScreenNumberOfScreen(gui->scr_ptr);


	/* Get the Depth */
	if(depth > 0)
	    gui->depth = depth;
	else
	    gui->depth = depth = XDefaultDepthOfScreen(gui->scr_ptr);


	/* Get the Visual */
	gui->visual = NULL;
	gui->visual_id = 0;
	if(vid != 0)
	{
	    /* Get the Visual that matches the specified visual ID */
	    int n;
	    XVisualInfo v, *vinfo_list;

	    v.visualid = vid;

	    vinfo_list = XGetVisualInfo(
		dpy,
		VisualIDMask,
		&v,
		&n
	    );
	    if((vinfo_list != NULL) && (n > 0))
	    {
		XVisualInfo *vinfo = &vinfo_list[0];
		gui->visual = vinfo->visual;
		gui->visual_id = vinfo->visualid;
	    }
	    else
	    {
		fprintf(
		    stderr,
"Visual %ld: No such Visual (the default Visual will be used).\n",
		    vid
		);
	    }
	    if(vinfo_list != NULL)
		XFree(vinfo_list);
	}
	else
	{
	    /* No specific Visual specified, get a suitable Visual */

	    visual_info_t vinfo;

	    if(XMatchVisualInfo(
		dpy, gui->scr_num, gui->depth, TrueColor,
		&vinfo
	    ))
	    {
		gui->visual = vinfo.visual;
		gui->visual_id = vinfo.visualid;
	    }
	    else if(XMatchVisualInfo(
		dpy, gui->scr_num, gui->depth, DirectColor,
		&vinfo
	    ))
	    { 
		gui->visual = vinfo.visual;
		gui->visual_id = vinfo.visualid;
	    }
	    else if(XMatchVisualInfo(
		dpy, gui->scr_num, gui->depth, PseudoColor,
		&vinfo
	    ))
	    { 
		gui->visual = vinfo.visual;
		gui->visual_id = vinfo.visualid;
	    }
	    if(gui->visual == NULL)
	    {
		fprintf(
		    stderr,
"OSWGUIConnect(): Warning: Unable to obtain a suitable Visual,\
 attempting to use the default Visual.\n"
		);
	    }
	}
	/* Did not get a Visual? */
	if(gui->visual == NULL)
	{
	    /* All else get the default Visual of the Screen */
	    gui->visual = DefaultVisualOfScreen(gui->scr_ptr);
	    gui->visual_id = XVisualIDFromVisual(gui->visual);
	}
	vis = gui->visual;
	if(vis == NULL)
	{
	    fprintf(
		stderr,
"OSWGUIConnect(): Warning: Unable to obtain a Visual.\n"
	    );
	}


	/* Get the bits per pixel by checking which Pixmap has the
	 * same depth as the default depth and what is its bits per
	 * pixel value
	 */
	gui->bits_per_pixel = 0;
	if(dpy != NULL)
	{
	    int n;
	    XPixmapFormatValues *list = XListPixmapFormats(dpy, &n);
	    if(list != NULL)
	    {
		int i;
		XPixmapFormatValues *v;
		for(i = 0; i < n; i++)
		{
		    v = &list[i];
		    if(v->depth == depth)
		    {
			gui->bits_per_pixel = v->bits_per_pixel;
			break;
		    }
		}
		XFree(list);
	    }
	}
	/* Unable to determine the number of bits per pixel? */
	if(gui->bits_per_pixel == 0)
	{
	    fprintf(
		stderr,
"OSWGUIConnect(): Warning: Unable to determine the number of\
 bits per pixel, assuming bits_per_pixel = depth = %i.\n",
		depth
	    );
	    gui->bits_per_pixel = depth;
	}
	gui->bytes_per_pixel = gui->bits_per_pixel / 8;
	if(gui->bytes_per_pixel < 1)
	{
	    fprintf(
		stderr,
"OSWGUIConnect(): Warning: bytes_per_pixel is too small, fixing to 1.\n"
	    );
	    gui->bytes_per_pixel = 1;
	}



	/* Get the color mask */
	gui->red_mask = vis->red_mask;
	gui->green_mask = vis->green_mask;
	gui->blue_mask = vis->blue_mask;


	/* Get the root Window (desktop) */
	gui->root_win = root = DefaultRootWindow(dpy);
	if(root == None)
	    fprintf(
		stderr,
"OSWGUIConnect(): Warning: Unable to obtain the root Window.\n"
	    );


	gui->toplevel = NULL;
	gui->ntoplevels = 0;

	/* Get the size of the Screen */
	gui->display_width = DisplayWidth(
	    dpy,
	    gui->scr_num
	);
	gui->display_height = DisplayHeight(
	    dpy,
	    gui->scr_num
	);


	/* Get the black and white pixels */
	gui->black_pix = BlackPixelOfScreen(gui->scr_ptr);
	gui->white_pix = WhitePixelOfScreen(gui->scr_ptr);


	/* Get the default graphics context */
	gui->gc = DefaultGC(dpy, gui->scr_num);


	/* Check if visual class is PseudoColor */
	// Dan S: The defined(__cplusplus) || defined(c_plusplus) causes redefinition
	// of "class" to be "c_class" (for C++ portability). There is no functional difference.
	if(vis->c_class == PseudoColor)
	{
	    /* Visual class is PseudoColor, we'll need to create
	     * a private colormap because the we need to handle
	     * image_t (XImages) in TrueColor.
	     */
	    switch(gui->depth)
	    {
	      case 8:
		gui->colormap = XCreateColormap(
		    dpy, root, vis, AllocNone
		);
		xc.flags = DoRed | DoGreen | DoBlue;
		for(i = 0; i < 256; i++)
		{
		    xc.blue = ((i & 0x03) << (5 + 8));
		    xc.green = ((i & 0x1C) << (3 + 8));
		    xc.red = ((i & 0xE0) << (0 + 8));
		    XAllocColor(dpy, gui->colormap, &xc);
		}
		XInstallColormap(dpy, gui->colormap);
		gui->we_created_colormap = True;
 		break;

	      /* Don't know how to handle other depths of PseudoColor,
	       * so just use default color map even though it may mess
	       * up colors
	       */
	      default:
		gui->colormap = DefaultColormap(dpy, gui->scr_num);
		gui->we_created_colormap = False;
		break;
	    }
	}
	else
	{
	    gui->colormap = DefaultColormap(dpy, gui->scr_num);
	    gui->we_created_colormap = False;
	}

	/* Load pointer cursors */
	gui->std_cursor = OSWCursorOpenID(XC_left_ptr);
	if(gui->std_cursor == None)
	    fprintf(
		stderr,
"OSWGUIConnect(): Warning: Unable to obtain the standard Cursor.\n"
	    );
	gui->drag_unaccepted_cursor = OSWCursorOpenXPMData(
	    drag_unaccepted_xpm,
	    0, 0
	);
	gui->drag_move_cursor = OSWCursorOpenXPMData(
	    drag_move_xpm,
	    0, 0
	);
	gui->drag_copy_cursor = OSWCursorOpenXPMData(
	    drag_copy_xpm,
	    0, 0
	);
	gui->drag_link_cursor = OSWCursorOpenXPMData(
	    drag_link_xpm,
	    0, 0
	);


#ifdef USE_XSHM
	/* Shared image put completion event code */
	gui->shm_completion_event_code = XShmGetEventBase(dpy) +
	    ShmCompletion;
#endif	/* USE_XSHM */


	/* ********************************************************* */
	/* Default foreground and background pixels */
	StringParseStdColor(
	    (bg_pix_str == NULL) ? "#000000" : bg_pix_str,
	    &r, &g, &b
	);
	OSWLoadPixelRGB(&gui->def_bg_pix, r, g, b);

	StringParseStdColor(
	    (fg_pix_str == NULL) ? "#ffffff" : fg_pix_str,
	    &r, &g, &b
	);
	OSWLoadPixelRGB(&gui->def_fg_pix, r, g, b);



	/* ********************************************************* */
	/* Load standard and default fonts */

	/* Standard font */
	if(OSWLoadFont(&gui->std_font, "7x14"))
	    fprintf(
		stderr,
"%s: Unable to load font.\n",
		"7x14"
	    );

	/* Default font (the one specified from command line -fn) */
	if(OSWLoadFont(
	    &gui->def_font,
	    (def_font_name == NULL) ? "7x14" : def_font_name
	))
	    fprintf(
		stderr,
"%s: Unable to load font.\n",
		(def_font_name == NULL) ? "7x14" : def_font_name
	    );
	/* Load something incase of error */
	if(gui->def_font == NULL)
	    OSWLoadFont(&gui->def_font, "7x14");


	/* ********************************************************* */
	/* Set up default foreground values */
	OSWSetForeground(gui->white_pix);
	OSWSetFont(gui->std_font);


	/* Reset block input level */
	gui->block_input_level = 0;

	/* Reset the grab pointer window */
	gui->grab_pointer_w = None;

	/* Reset the modal window */
	gui->modal_w = None;

	/* Reset selection */
	gui->selection_owner_w = None;
	gui->selection_data = NULL;
	gui->selection_len = 0;
	gui->selection_target = None;

	/* Reset DND */
	gui->dnd_in_progress = False;
	gui->dnd_src_w = None;
	gui->dnd_tar_w = None;
	gui->dnd_def_action = None;
	gui->dnd_action_flags = 0;
	gui->dnd_icon_w = None;
	gui->dnd_last_x = 0;
	gui->dnd_last_y = 0;


	/* ********************************************************* */
	/* Load property atoms (unique to X Window Systems) */

#define GET_ATOM(_s_)	XInternAtom(dpy, (_s_), False)

	/* Motif protocol$ */
	osw_atom._motif_wm_all_clients = GET_ATOM("_MOTIF_WM_ALL_CLIENTS");
	osw_atom._motif_wm_hints = GET_ATOM("_MOTIF_WM_HINTS");
	osw_atom._motif_wm_info = GET_ATOM("_MOTIF_WM_INFO");
	osw_atom._motif_wm_menu = GET_ATOM("_MOTIF_WM_MENU");
	osw_atom._motif_wm_messages = GET_ATOM("_MOTIF_WM_MESSAGES");
	osw_atom._motif_wm_offset = GET_ATOM("_MOTIF_WM_OFFSET");
	osw_atom._motif_wm_query = GET_ATOM("_MOTIF_WM_QUERY");

	/* Standard management protocols */
	osw_atom.wm_delete_window = GET_ATOM("WM_DELETE_WINDOW");
	osw_atom.wm_protocols = GET_ATOM("WM_PROTOCOLS");
	osw_atom.wm_save_yourself = GET_ATOM("WM_SAVE_YOURSELF");
	osw_atom.wm_state = GET_ATOM("WM_STATE");
	osw_atom.wm_take_focus = GET_ATOM("WM_TAKE_FOCUS");

	/* Selection */
	osw_atom.selection_property = GET_ATOM("OSW_SELECTION");
	osw_atom.selection_primary = XA_PRIMARY;
	osw_atom.target_string = XA_STRING;

	/* DND */
	osw_atom.dnd_action_ask = GET_ATOM("XdndActionAsk");
	osw_atom.dnd_action_copy = GET_ATOM("XdndActionCopy");
	osw_atom.dnd_action_description = GET_ATOM("XdndActionDescription");
	osw_atom.dnd_action_link = GET_ATOM("XdndActionLink");
	osw_atom.dnd_action_move = GET_ATOM("XdndActionMove");
	osw_atom.dnd_action_private = GET_ATOM("XdndActionPrivate");
	osw_atom.dnd_actions_list = GET_ATOM("XdndActionsList");
	osw_atom.dnd_aware = GET_ATOM("XdndAware");
	osw_atom.dnd_drop = GET_ATOM("XdndDrop");
	osw_atom.dnd_enter = GET_ATOM("XdndEnter");
	osw_atom.dnd_finished = GET_ATOM("XdndFinished");
	osw_atom.dnd_leave = GET_ATOM("XdndLeave");
	osw_atom.dnd_position = GET_ATOM("XdndPosition");
	osw_atom.dnd_proxy = GET_ATOM("XdndProxy");
	osw_atom.dnd_selection = GET_ATOM("XdndSelection");
	osw_atom.dnd_status = GET_ATOM("XdndStatus");
	osw_atom.dnd_type_list = GET_ATOM("XdndTypeList");

#undef GET_ATOM

	/* ******************************************************* */

	/* Set up graphics bit packing lookup table */
	OSWGenerateColorLookupTable(gui->depth);


	/* ******************************************************* */

	/* Set to standard error handler */
	XSetErrorHandler(OSWGUIErrorHandler);

	/* Reset the last error code */
	gui->last_error = 0;


	return(0);
}


/*
 *	Disconnect from GUI.
 */
void OSWGUIDisconnect(void)
{
	int i, n;
	pixel_t *pixel_array;
	osw_gui_struct *gui;

	if(IDC())
	{
	    gui = &osw_gui[0];

	    /* Purge all events */
	    OSWPurgeAllEvents();


	    /* Free resources */
	    OSWDestroyPixel(&gui->def_bg_pix);
	    OSWDestroyPixel(&gui->def_fg_pix);
	    OSWUnloadFont(&gui->def_font);

	    /* Free color map as needed */
	    if(gui->we_created_colormap)
	    {
		switch(gui->depth)
		{
		  case 8:
		    n = 256;
		    pixel_array = (pixel_t *)malloc(n * sizeof(pixel_t));
		    if(pixel_array != NULL)
		    {
		        for(i = 0; i < n; i++)
			    pixel_array[i] = (pixel_t)i;

			XFreeColors(
			    gui->display,
			    gui->colormap,
			    pixel_array,	/* Pixel array */
			    n,			/* Number of pixels */
			    0			/* Planes */
		        );
		        free(pixel_array);
		    }
		    break;
		}

		XUninstallColormap(gui->display, gui->colormap);
		XFreeColormap(gui->display, gui->colormap);
		gui->we_created_colormap = False;
	    }

	    /* Reset the block input level */
	    gui->block_input_level = 0;

	    /* Reset the modal window */
	    gui->modal_w = None;

	    /* Delete selection */
	    if(gui->selection_owner_w != None)
	    {
		if(XGetSelectionOwner(
		    gui->display, osw_atom.selection_primary
		) == gui->selection_owner_w)
		    XSetSelectionOwner(
			gui->display,
			osw_atom.selection_primary,
			None,
			CurrentTime
		    );
		gui->selection_owner_w = None;
	    }
	    free(gui->selection_data);
	    gui->selection_data = NULL;
	    gui->selection_len = 0;
	    gui->selection_target = None;

	    /* Reset the DND */
	    if(gui->dnd_in_progress)
		OSWDNDEnd(CurrentTime);

	    /* Ungrab the pointer as needed */
	    if(OSWIsPointerGrabbed())
		OSWUngrabPointer();

	    /* Delete list of toplevel windows */
	    free(gui->toplevel);
	    gui->toplevel = NULL;
	    gui->ntoplevels = 0;

	    OSWUnloadFont(&gui->std_font);

	    OSWDestroyCursor(&gui->std_cursor);
	    OSWDestroyCursor(&gui->drag_unaccepted_cursor);
	    OSWDestroyCursor(&gui->drag_move_cursor);
	    OSWDestroyCursor(&gui->drag_copy_cursor);
	    OSWDestroyCursor(&gui->drag_link_cursor);

	    /* Turn keyboard autorepeat back on */
	    OSWKBAutoRepeatOn();


	    /* Close the connection to the server */
	    XSetCloseDownMode(gui->display, DestroyAll);
	    XCloseDisplay(gui->display);

	    gui->def_use_shm = False;
	    gui->def_geometry_set = False;
	    gui->def_toplevel_width = 0;
	    gui->def_toplevel_height = 0;
	    gui->def_toplevel_x = 0;
	    gui->def_toplevel_y = 0;

	    gui->red_mask = 0l;
	    gui->green_mask = 0l;
	    gui->blue_mask = 0l;

	    gui->display_width = 0;
	    gui->display_height = 0;
	    gui->black_pix = 0;
	    gui->white_pix = 0;
	    gui->root_win = None;
	    gui->current_font = NULL;
	    gui->current_fg_pix = 0;
	    gui->current_bg_pix = 0;

	    gui->key_state = 0l;

	    gui->display = NULL;
	    gui->scr_num = 0;
	    gui->scr_ptr = NULL;
	    gui->visual = NULL;
	    gui->visual_id = 0;
	    gui->depth = 0;
	    gui->bits_per_pixel = 0;
	    gui->bytes_per_pixel = 0;

	    memset(&gui->colormap, 0x00, sizeof(colormap_t));
	    memset(&gui->gc, 0x00, sizeof(gc_t));
	}
}


/*
 *	KeyPress and KeyRelease event manager.
 *
 *	Called internally to manage KeyPress and KeyRelease events.
 */
static void OSWManageKeyEvent(
	osw_gui_struct *gui,
	event_t *event
)
{
	keycode_t keycode;
	event_key_t *key;

	if(event == NULL)
	    return;

	switch(event->type)
	{
	  case KeyPress:
	    key = &event->xkey;
	    keycode = key->keycode;

	    /* Update the exact key states from X */
	    gui->key_state = key->state;

	    /* Get modifiers separately, they are not reported by X
	     * until a subseqent event with a modifier occures
	     */
	    if(OSWIsModifierKey(keycode))
	    {
		switch(XKeycodeToKeysym(
		    gui->display,
		    keycode,
		    0
		))
		{
		  case XK_Meta_L:
		  case XK_Meta_R:
		    break;
		  case XK_Alt_L:
		  case XK_Alt_R:
		    gui->key_state |= OSW_KEY_STATE_ALT;
		    break;
		  case XK_Super_L:
		  case XK_Super_R:
		  case XK_Hyper_L:
		  case XK_Hyper_R:
		    break;
		  case XK_Control_L:
		  case XK_Control_R:
		    gui->key_state |= OSW_KEY_STATE_CTRL;
		    break;
		  case XK_Shift_L:
		  case XK_Shift_R:
		    gui->key_state |= OSW_KEY_STATE_SHIFT;
		    break;
		}
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    keycode = key->keycode;

	    /* Update the exact key states from X */
	    gui->key_state = key->state;

	    /* Get modifiers separately, they are not reported by X
	     * until a subseqent event with a modifier occures
	     */
	    if(OSWIsModifierKey(keycode))
	    {
		switch(XKeycodeToKeysym(
		    gui->display,
		    keycode,
		    0
		))
		{
		  case XK_Meta_L:
		  case XK_Meta_R:
		    break;
		  case XK_Alt_L:
		  case XK_Alt_R:
		    gui->key_state &= ~OSW_KEY_STATE_ALT;
		    break;
		  case XK_Super_L:
		  case XK_Super_R:
		  case XK_Hyper_L:
		  case XK_Hyper_R:
		    break;
		  case XK_Control_L:
		  case XK_Control_R:
		    gui->key_state &= ~OSW_KEY_STATE_CTRL;
		    break;
		  case XK_Shift_L:
		  case XK_Shift_R:
		    gui->key_state &= ~OSW_KEY_STATE_SHIFT;
		    break;
		}
	    }
	    break;
	}
}

/*
 *	Returns the keycode from the specified keyval.
 *
 *	The keyval must be one of OSW_KEY_* defined in oswkeyvals.h.
 */
keycode_t OSWGetKeyCodeFromKeyVal(const keyval_t keyval)
{
	if(keyval == OSW_KEY_None)
	    return(0);

	return(XKeysymToKeycode(
	    osw_gui[0].display,
	    keyval
	));
}

/*
 *	Returns the keyval from the specified keycode and key state.
 *
 *	The returned value will be one of OSW_KEY_* defined in
 *	oswkeyvals.h.
 */
keyval_t OSWGetKeyValFromKeyCode(
	const keycode_t keycode,
	const key_state_t key_state
)
{
	const int index = ((key_state & OSW_KEY_STATE_SHIFT) ||
			  (key_state & OSW_KEY_STATE_CAPS_LOCK)) ? 1 : 0;
	KeySym keysym;

	if(keycode == 0)
	    return(0x0000);

	keysym = XKeycodeToKeysym(osw_gui[0].display, keycode, index);  
	return((keysym != NoSymbol) ? (keyval_t)keysym : 0x0000);
}

/*
 *	Returns an ASCII char from the given keycode or '\0' if
 *	no match could be found.
 */
char OSWGetASCIIFromKeyCode(
	const keycode_t keycode,
	const  key_state_t key_state
)
{
	const bool_t num_lock_state = (key_state & OSW_KEY_STATE_NUM_LOCK) ? True : False;
	const int index = ((key_state & OSW_KEY_STATE_SHIFT) ||
			  (key_state & OSW_KEY_STATE_CAPS_LOCK)) ? 1 : 0;
	KeySym keysym;
	osw_gui_struct *gui = &osw_gui[0];

	if((gui->display == NULL) || (keycode == 0))
	    return('\0');

	keysym = XKeycodeToKeysym(gui->display, keycode, index);
	if(keysym == NoSymbol)
	    return('\0');

	/* Handle special keys */
	switch(keysym)
	{
	  case XK_KP_Space:
	    return(' ');
	    break;
	  case XK_KP_Tab:
	    return('\t');
	    break;
	  case XK_KP_Equal:
	    return('=');
	    break;
	  case XK_KP_Multiply:
	    return('*');
	    break;
	  case XK_KP_Add:
	    return('+');
	    break;
	  case XK_KP_Subtract:
	    return('-');
	    break;
	  case XK_KP_Decimal:
	    return('.');
	    break;
	  case XK_KP_Divide:
	    return('/');     
	    break;
	  case XK_KP_Home:
	    return(num_lock_state ? '7' : '\0');
	    break;
	  case XK_KP_Left:
	    return(num_lock_state ? '4' : '\0');
	    break;
	  case XK_KP_Up:
	    return(num_lock_state ? '8' : '\0');
	    break;
	  case XK_KP_Right:
	    return(num_lock_state ? '6' : '\0');
	    break;
	  case XK_KP_Down:
	    return(num_lock_state ? '2' : '\0');
	    break;
	  case XK_KP_Page_Up:
	    return(num_lock_state ? '9' : '\0');
	    break;
	  case XK_KP_Page_Down:
	    return(num_lock_state ? '3' : '\0');
	    break;
	  case XK_KP_End:
	    return(num_lock_state ? '1' : '\0');
	    break;
	  case XK_KP_Begin:
	    return(num_lock_state ? '5' : '\0');
	    break;
	  case XK_KP_Insert:
	    return(num_lock_state ? '0' : '\0');
	    break;
	  case XK_KP_Delete:
	    return(num_lock_state ? '.' : '\0');
	    break;
	  case XK_KP_Separator:
	    return(',');
	    break;
	  case XK_KP_0:
	    return('0');
	    break;
	  case XK_KP_1:
	    return('1');
	    break;
	  case XK_KP_2:
	    return('2');
	    break;
	  case XK_KP_3:
	    return('3');
	    break;
	  case XK_KP_4:
	    return('4');
	    break;
	  case XK_KP_5:
	    return('5');
	    break;
	  case XK_KP_6:
	    return('6');
	    break;
	  case XK_KP_7:
	    return('7');
	    break;
	  case XK_KP_8:
	    return('8');
	    break;
	  case XK_KP_9:
	    return('9');
	    break;
	}

	/* If any bits other than the first 8 are set then this is not
	 * an ASCII value that we can return
	 */
	if((u_int32_t)keysym & 0xFFFFFF00)
	    return('\0');

	return((char)((u_int32_t)keysym & 0x000000FF));
}

/*
 *	Returns a statically allocated string describing the name of
 *	the key specified keycode.
 */
const char *OSWGetKeyCodeName(const keycode_t keycode)
{
	if(keycode > 0)
	    return(OSWGetKeyValName(XKeycodeToKeysym(
	        osw_gui[0].display,
		keycode,
		0
	    )));
	else
	    return("#0");
}

/*
 *	Returns a statically allocated string describing the name of
 *	the key specified by keyval.
 */
const char *OSWGetKeyValName(const keyval_t keyval)
{
	if(keyval != 0x0000)
	    return(XKeysymToString(keyval));
	else
	    return("#0");
}


/*
 *	Get pointer coordinates.
 */
void OSWGetPointerCoords(
	win_t w,
	int *root_x_rtn, int *root_y_rtn,
	int *wx_rtn, int *wy_rtn
)
{
	int		root_x, root_y,
			wx, wy;
	unsigned int mask_return;
	win_t root_return, child_return;

	if(root_x_rtn != NULL)
	    *root_x_rtn = 0;
	if(root_y_rtn != NULL)
	    *root_y_rtn = 0;
	    
	if(wx_rtn != NULL)
	    *wx_rtn = 0;
	if(wy_rtn != NULL)
	    *wy_rtn = 0;

	if(w == None)
	    return;

	XQueryPointer(
	    osw_gui[0].display,
	    w, 
	    &root_return,
	    &child_return,
	    &root_x, &root_y,
	    &wx, &wy,
	    &mask_return
	);

	if(root_x_rtn != NULL)
	    *root_x_rtn = root_x;
	if(root_y_rtn != NULL)
	    *root_y_rtn = root_y;
  
	if(wx_rtn != NULL)
	    *wx_rtn = wx;
	if(wy_rtn != NULL)
	    *wy_rtn = wy;
}

/*
 *	Checks if the keycode is a modifier key.
 *
 *	Modifier keys are considered as follows:
 *
 *	alt, ctrl, and shift
 *
 *	Warning, Return may be incorrect if not connected to GUI.
 */
bool_t OSWIsModifierKey(const keycode_t keycode)
{
#if defined(IsModifierKey)
	return(IsModifierKey(XKeycodeToKeysym(
	    osw_gui[0].display,
	    keycode,
	    0
	)));
#elif defined(IsModiferKey)
	return(IsModiferKey(XKeycodeToKeysym(
	    osw_gui[0].display,
	    keycode,
	    0
	)));
#else
#warning OSWIsModifierKey(): IsModifierKey and IsModiferKey are not defined
	return(False);
#endif
}


/*
 *	Turns off keyboard autorepeat.
 */
void OSWKBAutoRepeatOff(void)
{
	XAutoRepeatOff(osw_gui[0].display);
}

/*
 *	Turns on keyboard autorepeat.
 */
void OSWKBAutoRepeatOn(void)
{
	XAutoRepeatOn(osw_gui[0].display);
}

/*
 *	Returns the keyboard autorepeat state.
 */
bool_t OSWGetKBAutoRepeatState(void)
{
	XKeyboardState kb;
	osw_gui_struct *gui = &osw_gui[0];
	XGetKeyboardControl(
	    gui->display,
	    &kb
	);
	return((kb.global_auto_repeat == AutoRepeatModeOn) ? True : False);
}


/*
 *	Sets the input focus.
 */
void OSWSetInputFocus(
	win_t w,
	const int revert,
	const time_t t
)
{
	XSetInputFocus(
	    osw_gui[0].display,
	    w,
	    revert,
	    (Time)t
	);
}

/*
 *	Same as OSWSetInputFocus() with revert set to RevertToParent.
 */
void OSWGrabFocusTime(
	win_t w,
	const time_t t
)
{
	XSetInputFocus(
	    osw_gui[0].display,
	    w,
	    RevertToParent,
	    (Time)t
	);
}

void OSWGrabFocus(win_t w)
{
	OSWGrabFocusTime(
	    w,
	    CurrentTime
	);
}


/*
 *	Flushes the GUI.
 */
void OSWFlush(void)
{
	XFlush(osw_gui[0].display);
}

/*
 *	Flush events and sync with GUI.
 */
void OSWSync(const bool_t discard)
{
	XSync(
	    osw_gui[0].display,
	    discard
	);
}


/*
 *	Grabs the pointer.
 *
 *	The grab_w specifies the grab window.
 *
 *	If events_rel_grab_w is True then all events will be reported
 *	relative to the grab window during the grab.
 *
 *	The eventmask specifies which events to report during the
 *	grab.
 *
 *	The pointer_mode specifies the pointer syncronization, which
 *	can be either GrabModeSync or GrabModeAsync.
 *
 *	The keyboard_mode specifies the keyboard syncronization, which
 *	can be either GrabModeSync or GrabModeAsync.
 *
 *	The confine_w specifies the window to confine the pointer to.
 *	If confine_w is None then the pointer will not be confined to
 *	any window.
 *
 *	The cursor specifies the cursor to display during the grab.
 *
 *	If grab_w or confine_w is not viewable or if confine_w lies
 *	completely outside of the boundaries of the root window
 *	then XGrabPointer fails and returns GrabNotViewable.
 *
 *	If the pointer is actively grabbed by another client then it
 *	fails and returns AlreadyGrabbed.
 *
 *	If the pointer is frozen by an active grab of another client
 *	then it fails and returns GrabFrozen.
 *
 *	If the specified time is earlier than the last pointer grab
 *	time or later than the current X server time then it fails and
 *	returns GrabInvalidTime.
 */
int OSWGrabPointer(
	win_t grab_w,
	const bool_t events_rel_grab_w,
	const event_mask_t eventmask,
	const int pointer_mode,
	const int keyboard_mode,
	win_t confine_w,
	cursor_t cursor
)
{
	return(OSWGrabPointerTime(
	    grab_w,
	    events_rel_grab_w,
	    eventmask,
	    pointer_mode,
	    keyboard_mode,
	    confine_w,
	    cursor,
	    CurrentTime
	));
}

/*
 *	Grabs the pointer.
 *
 *	The grab_w specifies the grab window.
 *
 *	If events_rel_grab_w is True then all events will be reported
 *	relative to the grab window during the grab.
 *
 *	The eventmask specifies which events to report during the
 *	grab.
 *
 *	The pointer_mode specifies the pointer syncronization, which
 *	can be either GrabModeSync or GrabModeAsync.
 *
 *	The keyboard_mode specifies the keyboard syncronization, which
 *	can be either GrabModeSync or GrabModeAsync.
 *
 *	The confine_w specifies the window to confine the pointer to.
 *	If confine_w is None then the pointer will not be confined to
 *	any window.
 *
 *	The cursor specifies the cursor to display during the grab.
 *
 *	The t specifies the current time. If this value is not
 *	available then CurrentTime can be used.
 *
 *	If grab_w or confine_w is not viewable or if confine_w lies
 *	completely outside of the boundaries of the root window
 *	then XGrabPointer fails and returns GrabNotViewable.
 *
 *	If the pointer is actively grabbed by another client then it
 *	fails and returns AlreadyGrabbed.
 *
 *	If the pointer is frozen by an active grab of another client
 *	then it fails and returns GrabFrozen.
 *
 *	If the specified time is earlier than the last pointer grab
 *	time or later than the current X server time then it fails and
 *	returns GrabInvalidTime.
 */
int OSWGrabPointerTime(
	win_t grab_w,
	const bool_t events_rel_grab_w,
	const event_mask_t eventmask,
	const int pointer_mode,
	const int keyboard_mode,
	win_t confine_w,
	cursor_t cursor,
	const time_t t
)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(grab_w == None)
	    return(!GrabSuccess);

	gui->grab_pointer_w = grab_w;
	return(XGrabPointer(
	    gui->display,
	    grab_w,
	    !events_rel_grab_w,
	    eventmask,
	    pointer_mode,
	    keyboard_mode,
	    confine_w,
	    cursor,
	    t
	));
}

/*
 *	Ungrabs the pointer.
 */
void OSWUngrabPointer(void)
{
	OSWUngrabPointerTime(CurrentTime);
}

void OSWUngrabPointerTime(const time_t t)
{
	osw_gui_struct *gui = &osw_gui[0];
	XUngrabPointer(
	    gui->display,
	    (Time)t
	);
	gui->grab_pointer_w = None;
}

/*
 *	Returns the pointer grab state.
 */
bool_t OSWIsPointerGrabbed(void)
{
	return((osw_gui[0].grab_pointer_w != None) ? True : False);
}


/*
 *	Gets the atom specified by name.
 */
atom_t OSWGetAtom(const char *name)
{
	if(name == NULL)
	    return(None);

	return(XInternAtom(osw_gui[0].display, name, False));
}


/*
 *	Manages the SelectionNotify, SelectionRequest, or
 *	SelectionClear event.
 *
 *	This for the processing of responding to SelectionRequest
 *	events and to watch for when someone else has taken the
 *	selection in SelectionClear events.
 */
static void OSWManageSelectionEvent(
	osw_gui_struct *gui,
	event_t *event
)
{
	XSelectionEvent *sel_notify;
	XSelectionRequestEvent *sel_request;
	XSelectionClearEvent *sel_clear;

	switch(event->type)
	{
	  case SelectionNotify:
	    sel_notify = &event->xselection;
	    break;

	  case SelectionRequest:
	    sel_request = &event->xselectionrequest;
	    /* Requesting from our primary selection? */
	    if((sel_request->owner == gui->selection_owner_w) &&
	       (sel_request->selection == osw_atom.selection_primary)
	    )
	    {
		XSelectionEvent *re = (XSelectionEvent *)calloc(
		    1, sizeof(XSelectionEvent)
		);

		/* Put/copy our data to the requesting window */
		XChangeProperty(
		    sel_request->display, sel_request->requestor,
		    sel_request->property, sel_request->target,
		    8, PropModeReplace,
		    (unsigned char *)gui->selection_data,
		    gui->selection_len
		);

		/* Send response to requestor */
		re->type = SelectionNotify;
		re->serial = sel_request->serial;
		re->send_event = sel_request->send_event;
		re->display = sel_request->display;
		re->requestor = sel_request->requestor;
		re->selection = sel_request->selection;
		re->target = sel_request->target;
		re->property = sel_request->property;
		re->time = sel_request->time;
		XSendEvent(
		    sel_request->display, sel_request->requestor,
		    False, 0l, (XEvent *)re
		);
		free(re);
	    }
	    /* Requesting from our DND source window's selection? */
	    else if((sel_request->owner == gui->dnd_src_w) &&
		    (sel_request->selection == osw_atom.dnd_selection)
	    )
	    {
		XSelectionEvent *re = (XSelectionEvent *)calloc(
		    1, sizeof(XSelectionEvent)
		);

		/* Put/copy our data on to the DND target window */
		XChangeProperty(
		    sel_request->display, sel_request->requestor,
		    sel_request->property, sel_request->target,
		    8, PropModeReplace,
		    (unsigned char *)gui->dnd_data,
		    gui->dnd_data_len
		);

		/* Send response to the DND target window */
		re->type = SelectionNotify;
		re->serial = sel_request->serial;
		re->send_event = sel_request->send_event;
		re->display = sel_request->display;
		re->requestor = sel_request->requestor;
		re->selection = sel_request->selection;
		re->target = sel_request->target;
		re->property = sel_request->property;
		re->time = sel_request->time;
		XSendEvent(
		    sel_request->display, sel_request->requestor,   
		    False, 0l, (XEvent *)re
		);
		free(re);
	    }
	    else
	    {
		/* Send response to requestor indicating error */
		XSelectionEvent *re = (XSelectionEvent *)calloc(
		    1, sizeof(XSelectionEvent)
		);
		re->type = SelectionNotify;
		re->serial = sel_request->serial;
		re->send_event = sel_request->send_event;
		re->display = sel_request->display;
		re->requestor = sel_request->requestor;
		re->selection = sel_request->selection;
		re->target = sel_request->target;
		re->property = None;
		re->time = sel_request->time;
		XSendEvent(
		    sel_request->display, sel_request->requestor,
		    False, 0l, (XEvent *)re
		);
		free(re);
	    }
	    break;

	  case SelectionClear:
	    sel_clear = &event->xselectionclear;
	    if(sel_clear->window == gui->selection_owner_w)
	    {
		gui->selection_owner_w = None;
		free(gui->selection_data);
		gui->selection_data = NULL;
		gui->selection_len = 0;
		gui->selection_target = None;
	    }
	    break;
	}
}

/*
 *	Checks if the current selection owner of the specified selection
 *	is different from the specified window. If it is then it sends
 *	a XSelectionClear event to the current selection owner if the
 *	current selection owner is a window managed by us.
 */
static void OSWDDEReportClear(
	osw_gui_struct *gui, win_t w,
	atom_t selection
)
{
	/* If the current selection owner (if any) is different from
	 * the specified selection owner and it is one of our Windows 
	 * then we need to send our window a XSelectionClear event
	 * because it is not automatically sent in this case
	 */
	win_t current_owner_w = XGetSelectionOwner(gui->display, selection);
	if((current_owner_w != None) &&
	   (current_owner_w != w) &&
	   (current_owner_w == gui->selection_owner_w)
	)
	{
	    XSelectionClearEvent *e = (XSelectionClearEvent *)calloc(
		1, sizeof(XSelectionClearEvent)
	    );
	    e->type = SelectionClear;
	    e->serial = 0l;
	    e->send_event = False;
	    e->display = gui->display;
	    e->window = current_owner_w;
	    e->selection = selection;
	    e->time = CurrentTime;
	    XSendEvent(
		e->display, e->window,
		False, 0l, (XEvent *)e
	    );
	    free(e);
	}
}

/*
 *	Gets the DDE string value.
 *
 *	The calling function must delete the returned string.
 */
char *OSWDDEGetString(win_t w)
{
	const time_t start_time = time(NULL);
	Atom	selection = osw_atom.selection_primary,
		selection_target = osw_atom.target_string,
		selection_property = osw_atom.selection_property;
	event_t event;
	char *s = NULL;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return(NULL);

	/* Request value from the primary selection */
	XConvertSelection(
	    gui->display, selection,
	    selection_target,		/* Target */
	    selection_property,		/* Property */
	    w,
	    CurrentTime
	);

	/* Wait up to 30 seconds for response */
	while((time(NULL) - start_time) <= 30l)
	{
	    /* If a SelectionRequest is received then handle it, since
	     * it may be a request for a selection from one of our
	     * Windows
	     */
	    if(XCheckTypedEvent(
		gui->display, SelectionRequest, &event
	    ))
		OSWManageSelectionEvent(gui, &event);

	    /* Check if there is a SelectionNotify for our requesting
	     * Window
	     */
	    if(XCheckTypedWindowEvent(
		gui->display, w, SelectionNotify, &event
	    ))
	    {
		/* The requested data has arrived */
		XSelectionEvent *sel_notify = &event.xselection;

		/* Selection value available? */
		if(sel_notify->property != None)
		{
		    /* Get the selection value */
		    Atom v_target;
		    int v_format, bytes_read = 0;
		    unsigned long nitems, bytes_after;
		    unsigned char *v;

		    do {
			if(XGetWindowProperty(
			    sel_notify->display,
			    sel_notify->requestor,
			    sel_notify->property,
			    (long)(bytes_read / 4),	/* Offset */
			    (long)(1024 / 4),		/* Length */
			    True,			/* Delete */
			    AnyPropertyType,
			    &v_target,			/* Type return */
			    &v_format,			/* Element size (in bits) */
			    &nitems,
			    &bytes_after,
			    &v				/* Value return */
			) != Success)
			    break;

			/* Make sure that the received type and format
			 * matches our requested type and format
			 */
			if((v_target == selection_target) &&
			   (v_format == 8) && (nitems > 0)
			)
			{
			    s = strinsstr(s, -1, (const char *)v);
			}

			/* Delete the received selection segment value */
			if(v != NULL)
			    XFree(v);

			bytes_read += (int)nitems * (v_format / 8);

		    } while(bytes_after > 0);
		}
		break;
	    }
	}

	return(s);
}

/*
 *	Sets the DDE to the specified string value.
 */
void OSWDDESetString(win_t w, const char *s)
{
	Atom	selection = osw_atom.selection_primary,
		selection_target = osw_atom.target_string;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	OSWDDEReportClear(gui, w, selection);

	/* Copy the data to the selection */
	free(gui->selection_data);
	gui->selection_data = STRDUP(s);
	gui->selection_len = STRLEN(s);
	gui->selection_target = selection_target;	/* Type */

	/* Set the new selection owner */
	XSetSelectionOwner(
	    gui->display, selection,
	    w, CurrentTime
	);

	/* Verify that the selection owner has been set */
	if(XGetSelectionOwner(gui->display, selection) == w)
	    gui->selection_owner_w = w;
}

/*
 *	Gets the DDE data.
 *
 *	The calling function must delete the returned data.
 */
void *OSWDDEGetData(win_t w, atom_t target, int *len_rtn)
{
	const time_t start_time = time(NULL);
	Atom    selection = osw_atom.selection_primary,
		selection_property = osw_atom.selection_property;
	event_t event;
	void *data = NULL;
	int len = 0;
	osw_gui_struct *gui = &osw_gui[0];

	if((w == None) || (len_rtn == NULL))
	    return(NULL);

	*len_rtn = len;

	/* Request value from the primary selection */
	XConvertSelection(
	    gui->display, selection,
	    target,			/* Type */
	    selection_property,		/* Property */
	    w,
	    CurrentTime
	);

	/* Wait up to 30 seconds for response */
	while((time(NULL) - start_time) <= 30l)
	{
	    /* If a SelectionRequest is received then handle it, since
	     * it may be a request for a selection from one of our
	     * Windows
	     */
	    if(XCheckTypedEvent(
		gui->display, SelectionRequest, &event
	    ))
		OSWManageSelectionEvent(gui, &event);

	    /* Check if there is a SelectionNotify for our requesting
	     * Window
	     */
	    if(XCheckTypedWindowEvent(
		gui->display, w, SelectionNotify, &event
	    ))
	    { 
		/* The requested data has arrived */
		XSelectionEvent *sel_notify = &event.xselection;

		/* Selection value available? */
		if(sel_notify->property != None)
		{
		    /* Get the selection value */
		    Atom v_target;
		    int v_format, bytes_read = 0;
		    unsigned long nitems, bytes_after;
		    unsigned char *v;

		    do {
			if(XGetWindowProperty(
			    sel_notify->display,
			    sel_notify->requestor,
			    sel_notify->property,
			    (long)(bytes_read / 4),	/* Offset */
			    (long)(1024 / 4),		/* Length */
			    True,			/* Delete */
			    AnyPropertyType,
			    &v_target,			/* Type return */
			    &v_format,			/* Element size (in bits) */
			    &nitems,
			    &bytes_after,
			    &v				/* Value return */
			) != Success)
			    break;

			/* Make sure that the received type and format
			 * matches our requested type and format
			 */
			if((v_target == target) && (nitems > 0))
			{
			    const int	i = len,
					seg_len = (int)nitems * (v_format / 8);
			    len += seg_len;
			    data = realloc(data, len);
			    if(data != NULL)
				memcpy((u_int8_t *)data + i, v, seg_len);
			    else
				len = 0;
			}

			/* Delete the received selection segment value */
			if(v != NULL)
			    XFree(v);

			bytes_read += (int)nitems * (v_format / 8);

		    } while(bytes_after > 0);
		}
		break;
	    }
	}

	*len_rtn = len;

	return(data);
}

/*
 *	Sets the DDE to the specified data.
 */
void OSWDDESetData(win_t w, atom_t target, const void *data, int len)
{
	Atom selection = osw_atom.selection_primary;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	OSWDDEReportClear(gui, w, selection);

	/* Copy the data to the selection */
	free(gui->selection_data);
	gui->selection_len = 0;
	gui->selection_data = malloc(len);
	if(gui->selection_data != NULL)
	{
	    memcpy(gui->selection_data, data, len);
	    gui->selection_len = len;
	}
	gui->selection_target = target;

	/* Set the new selection owner */
	XSetSelectionOwner(
	    gui->display, selection,
	    w, CurrentTime
	);

	/* Verify that the selection owner has been set */
	if(XGetSelectionOwner(gui->display, selection) == w)
	    gui->selection_owner_w = w;
}


/*
 *	Manages the DND event for the DND source window.
 *
 *	The drag must be in progress and that we started it.
 */
static void OSWManageDNDEvent(osw_gui_struct *gui, event_t *event)
{
	bool_t drop_processed_and_successful;
	atom_t action;
	win_t w, src_w = gui->dnd_src_w;
	event_button_t *button;
	event_motion_t *motion;
	XCrossingEvent *crossing;

	/* DND not set in progress by us or the source window has not
	 * been set?
	 */
	if(!gui->dnd_in_progress || (src_w == None))
	    return;

	w = event->xany.window;
	switch(event->type)
	{
	  case ButtonRelease:
	    button = &event->xbutton;
	    if(button->subwindow == gui->dnd_tar_w)
	    {
		if(gui->dnd_tar_action != None)
		    OSWDNDSendDrop(w, src_w, (time_t)button->time);
		else
		    OSWDNDEnd((time_t)button->time);
printf("DND drop on DND window %ui\n", (unsigned int)w);
	    }
	    else
	    {
printf("DND operation canceled\n");
		OSWDNDEnd((time_t)button->time);
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(motion->subwindow == gui->dnd_tar_w)
	    {
		w = motion->subwindow;

		/* Move the DND icon window to follow the pointer */
		if(gui->dnd_icon_w != None)
		{
		    int x_offset = 8,
			y_offset = 8;
		    OSWMoveWindow(
			gui->dnd_icon_w,
			(int)motion->x_root + x_offset,
			(int)motion->y_root + y_offset
		    );
		}

		/* Send and update position */
		OSWDNDSendPosition(
		    w, src_w,
		    gui->dnd_def_action,
		    (int)motion->x_root, (int)motion->y_root,
		    (time_t)motion->time
		);
		gui->dnd_last_x = (int)motion->x_root;
		gui->dnd_last_y = (int)motion->y_root;

printf(
 "DND motion on window %ui %i %i\n",
 (unsigned int)w, (int)motion->x_root, (int)motion->y_root
);
	    }
	    break;

	  case EnterNotify:
	    crossing = &event->xcrossing;
	    if(crossing->subwindow != gui->dnd_tar_w)
	    {
		w = crossing->subwindow;
printf("DND Entered new window %ui\n", (unsigned int)w);

		/* Is this window is DND aware? */
		if(OSWDNDQueryPointer(
		    &w, NULL, NULL, NULL, NULL
		))
		{
		    atom_t types_list[1];
		    types_list[0] = gui->dnd_src_target;
		    OSWDNDSendEnter(
			w, src_w,
			types_list,
			sizeof(types_list) / sizeof(atom_t)
		    );
printf("DND window accepts DND\n");

		    /* Set the current DND target window */
		    gui->dnd_tar_w = w;
		    gui->dnd_tar_action = None;
		}
	    }
	    break;

	  case LeaveNotify:
	   crossing = &event->xcrossing;
	    if(crossing->subwindow == gui->dnd_tar_w)
	    {
		w = crossing->subwindow;

		OSWDNDSetCursor(
		    gui->drag_unaccepted_cursor,
		    (time_t)crossing->time
		);
		OSWDNDSendLeave(w, src_w);

printf("DND Leaving DND window %ui\n", (unsigned int)w);
		/* Unset the current DND target window and action */
		gui->dnd_tar_w = None;
		gui->dnd_tar_action = None;
	    }
	    break;

	  case ClientMessage:
	    /* DND Status? */
	    if(OSWEventIsDNDStatus(
		src_w, event,
		&w, &action,
		NULL
	    ))
	    {
printf("DND received status from DND window %ui\n", (unsigned int)w);
		if(action != None)
		{
		    /* Target accepts drop */
		    cursor_t cur;
		    if(action == osw_atom.dnd_action_link)
			cur = gui->drag_link_cursor;
		    else if(action == osw_atom.dnd_action_copy)
			cur = gui->drag_copy_cursor;
		    else if(action == osw_atom.dnd_action_move)
			cur = gui->drag_move_cursor;
		    else
			cur = gui->drag_unaccepted_cursor;
		    OSWDNDSetCursor(cur, CurrentTime);
		}
		else
		{
		    OSWDNDSetCursor(
			gui->drag_unaccepted_cursor,
			CurrentTime
		    );
		}
		gui->dnd_tar_action = action;
	    }
	    /* DND Finished? */
	    else if(OSWEventIsDNDFinished(
		src_w, event,
		&w,
		&drop_processed_and_successful,
		&action
	    ))
	    {
printf("DND Finished\n");
		OSWDNDEnd(CurrentTime);
	    }
	    break;
	}
}

/*
 *	Sets the window as a DND target.
 *
 *	This makes the specified window "DND aware" and allows it to
 *	accept DND drops.
 *
 *	The specified window must be a toplevel window.
 *
 *	The DND protocol will be set as xdnd protocol version 5.
 */
void OSWDNDSetWindowTarget(win_t w)
{
	unsigned char v[] = {
		0x05			/* Version */
	};
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	/* Set the "DND aware" property value on this window */
	XChangeProperty(
	    gui->display, w,
	    osw_atom.dnd_aware,		/* Property */
	    XA_ATOM,			/* Type */
	    32,				/* Bits */
	    PropModeReplace,
	    v,				/* Value */
	    sizeof(v) / sizeof(unsigned char)
	);
}

/*
 *	Checks if the pointer is currently over a window that is set as
 *	a DND target.
 *
 *	The window must have the xdnd protocol version 4 property set.
 *
 *	The *w_rtn is the child window of the toplevel that is set as
 *	a DND target that the pointer is currently over (in accordance
 *	with the xdnd protocol version 3 or newer).
 *
 *	Returns True if the pointer is currently over a window that is
 *	set as a DND target.
 */
bool_t OSWDNDQueryPointer(
	win_t *w_rtn,
	int *x_rtn, int *y_rtn,
	int *root_x_rtn, int *root_y_rtn
)
{
	int x, y, rx, ry, format;
	unsigned int mask;
	unsigned long nitems, bytes_remaining;
	unsigned char *v;
	atom_t type;
	win_t w, parent, root;
	osw_gui_struct *gui = &osw_gui[0];

	if(w_rtn != NULL)
	    *w_rtn = 0;
	if(x_rtn != NULL)
	    *x_rtn = 0;
	if(y_rtn != NULL)
	    *y_rtn = 0;
	if(root_x_rtn != NULL)
	    *root_x_rtn = 0;
	if(root_y_rtn != NULL)
	    *root_y_rtn = 0;

	/* Get the window that the pointer is currently over */
	if(!XQueryPointer(
	    gui->display, gui->root_win,
	    &root, &w,
	    &rx, &ry,
	    &x, &y,
	    &mask
	))
	    return(False);

	if(w == None)
	    return(False);

	/* In xdnd protocol version 3 or newer, the toplevel window
	 * (not a child window) must be set as a DND target, so
	 * traverse the window's ancestry until we find a window that
	 * is set as a DND target (excluding the root window)
	 */
	parent = w;
	while((parent != None) && (parent != root))
	{
	    /* Check if this window has the "DND aware" property set
	     * on it and, if so, get its value
	     */
	    if(XGetWindowProperty(
		gui->display, parent,
		osw_atom.dnd_aware,
		0l,
		(long)(1024 / 4),
		False,
		XA_ATOM,
		&type,
		&format,
		&nitems,
		&bytes_remaining,
		&v
	    ) == Success)
	    {
		/* This window has the "DND aware" property set,
		 * therefore defining it as a DND target, now check if
		 * its version is set to xdnd 3 or newer
		 */
		bool_t status = False;
		if((v != NULL) && (nitems >= 1))
		{
		    int version = 0;
		    switch(format)
		    {
		      case 8:
			version = (int)(*(u_int8_t *)v);
		        break;
		      case 16:
		        version = (int)(*(u_int16_t *)v);
			break;
		      case 32:
		        version = (int)(*(u_int32_t *)v);
			break;
		    }
		    status = (version >= 3) ? True : False;
		}

		if(v != NULL)
		    XFree(v);

		if(status)
		{
		    if(w_rtn != NULL)
			*w_rtn = w;
		    if(x_rtn != NULL)
			*x_rtn = x;
		    if(y_rtn != NULL)
			*y_rtn = y;
		    if(root_x_rtn != NULL)
			*root_x_rtn = rx;
		    if(root_y_rtn != NULL)
			*root_y_rtn = ry;
		}
		return(status);
	    }

	    parent = OSWGetWindowParent(parent);
	}

	return(False);
}

/*
 *	Returns the action based on the current modifier keys compared
 *	with the allowed actions specified in action_flags.
 *
 *	If no modifier keys are held down then def_action is returned.
 */
atom_t OSWDNDQueryAction(
	win_t w, atom_t def_action, dnd_action_flags_t action_flags
)
{
	int x, y, root_x, root_y;
	unsigned int state;
	win_t root, child;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    w = gui->root_win;
	if(w == None)
	    return(None);

	/* Get the current modifier keys */
	if(XQueryPointer(
	    gui->display, w,
	    &root, &child,
	    &x, &y, &root_x, &root_y,
	    &state
	))
	{
	    if((state & OSW_KEY_STATE_CTRL) &&
	       (state & OSW_KEY_STATE_SHIFT)
	    )
		return((action_flags & OSW_DND_ACTION_LINK) ?
		    osw_atom.dnd_action_link : None
		);
	    else if(state & OSW_KEY_STATE_CTRL)
		return((action_flags & OSW_DND_ACTION_COPY) ?
		    osw_atom.dnd_action_copy : None
		);
	    else if(state & OSW_KEY_STATE_SHIFT)
		return((action_flags & OSW_DND_ACTION_MOVE) ?
		    osw_atom.dnd_action_move : None  
		);
	    else
		return(def_action);
	}
	else
	{
	    return(None);
	}
}

/*
 *	Gets the DND actions list from the source window specified by
 *	w.
 *
 *	The source window is expected to have this list if it sent a
 *	DND position event to the target window with an action that
 *	specifies dnd_action_ask.
 *
 *	The calling function must delete the returned actions list.
 */
atom_t *OSWDNDGetActionsList(win_t w, int *nactions)
{
	int format;
	unsigned long nitems, bytes_remaining;
	unsigned char *v;
	atom_t type, *list = NULL;
	osw_gui_struct *gui = &osw_gui[0];

	if((w == None) || (nactions == NULL))
	    return(list);

	*nactions = 0;

	/* Get the DND actions list property value from the source
	 * window
	 */
	if(XGetWindowProperty(
	    gui->display, w,
	    osw_atom.dnd_actions_list,
	    0l,
	    (long)(65536 / 4),
	    False,
	    XA_ATOM,
	    &type,
	    &format,
	    &nitems, &bytes_remaining,
	    &v
	) != Success)
	    return(list);

	/* Did we get the actions list property value and is it an
	 * array of atoms?
	 */
	if((v != NULL) && (nitems >= 1) &&
	   (type == XA_ATOM) && (format == 32)
	)
	{
	    *nactions = nitems;
	    list = (atom_t *)realloc(
		list, nitems * sizeof(atom_t)
	    );
	    if(list != NULL)
	    {
		int i;
		const u_int32_t *v32 = (u_int32_t *)v;
		for(i = 0; i < (int)nitems; i++)
		    list[i] = (atom_t)v32[i];
	    }
	    else
	    {
		*nactions = 0;
	    }
	}

	if(v != NULL)
	    XFree(v);

	return(list);
}


/*
 *	Same as OSWDNDGetActionsList() except that it returns a list
 *	of strings describing the respective actions obtained in
 *	OSWDNDGetActionsList().
 *
 *	The calling function must delete the returned strings list and
 *	each string.
 */
char **OSWDNDGetActionNamesList(win_t w, int *nactions)
{
	int format;
	unsigned long nitems, bytes_remaining;
	unsigned char *v;
	atom_t type;
	char **list = NULL;
	osw_gui_struct *gui = &osw_gui[0];

	if((w == None) || (nactions == NULL))
	    return(list);

	*nactions = 0;

	/* Get the DND actions list property value from the source
	 * window
	 */
	if(XGetWindowProperty(
	    gui->display, w,
	    osw_atom.dnd_actions_list,
	    0l,
	    (long)(65536 / 4),
	    False,
	    XA_STRING,
	    &type,
	    &format,
	    &nitems, &bytes_remaining,
	    &v
	) != Success)
	    return(list);

	if((v != NULL) && (nitems >= 1) &&
	   (type == XA_STRING) && (format == 8)
	)
	{
	    int i = 0;
	    const char	*vc = (char *)v,
			*vc_end = vc + nitems;
	    while(vc < vc_end)
	    {
		/* Create a new string */
		i = *nactions;
		*nactions = i + 1;
		list = (char **)realloc(list, (*nactions) * sizeof(char *));
		if(list == NULL)
		    break;
		list[i] = strdup(vc);

		/* Seek vc to the next string segment by finding the
		 * next null character (and seek past that)
		 */
		while(vc < vc_end)
		{
		    if(*vc == '\0')
		    {
			vc++;
		        break;
		    }
		    vc++;
		}
	    }
	}    

	if(v != NULL)
	    XFree(v);

	return(list);
}


/*
 *	Sends a DND enter event to the target window specified by w.
 */
void OSWDNDSendEnter(
	win_t w, win_t src_w,
	const atom_t *type_list, int ntypes
)
{
	const unsigned char version = 0x03;	/* Min version */
	long flags;
	event_client_message_t *e;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	e = (event_client_message_t *)calloc(
	    1, sizeof(event_t)
	);
	if(e == NULL)
	    return;

	flags = (ntypes > 3) ? (1 << 0) : (0 << 0);
	flags |= (version << 24);

	e->type = ClientMessage;
	e->serial = 0l;
	e->send_event = False;
	e->display = gui->display;
	e->window = w;
	e->message_type = osw_atom.dnd_enter;
	e->format = 32;
	e->data.l[0] = (long)src_w;
	e->data.l[1] = flags;
	e->data.l[2] = (ntypes > 0) ? (long)type_list[0] : 0l;
	e->data.l[3] = (ntypes > 1) ? (long)type_list[1] : 0l;
	e->data.l[4] = (ntypes > 2) ? (long)type_list[2] : 0l;
	XSendEvent(
	    gui->display, w, False, 0l, (XEvent *)e
	);
	free(e);
}

/*
 *	Sends a DND position event to the target window specified by w.
 *
 *	Tbe action specifies the action requested by the user. The
 *	action can also be dnd_action_ask, in which case the target
 *	window should call OSWDNDGetActionsList(src_w, ...) to get the
 *	actions list from the source window.
 *
 *	The specified position must be relative to the root window.
 */
void OSWDNDSendPosition(
	win_t w, win_t src_w, atom_t action,
	int root_x, int root_y,
	time_t t
)
{
	event_client_message_t *e;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)                
	    return;

	e = (event_client_message_t *)calloc(
	    1, sizeof(event_t)
	);
	if(e == NULL)
	    return;

	e->type = ClientMessage;
	e->serial = 0l;
	e->send_event = False;
	e->display = gui->display;
	e->window = w;
	e->message_type = osw_atom.dnd_position;
	e->format = 32;
	e->data.l[0] = (long)src_w;
	e->data.l[1] = 0l;		/* Flags, reserved */
	e->data.l[2] = ((long)root_x << 16) | (long)root_y;
	e->data.l[3] = (long)t;
	e->data.l[4] = (long)action;
	XSendEvent(
	    gui->display, w, False, 0l, (XEvent *)e
	);
	free(e);
}

/*
 *	Sends a DND status event to the source window specified by w.
 *
 *	If action is None then a "drop not accepted" status will be
 *	sent, any other action specifies that the drop is accepted and
 *	that the drop action must be the specified action.
 *
 *	If rect is NULL then a rectangle of 0 size will be sent.
 */
void OSWDNDSendStatus(
	win_t w, win_t tar_w, atom_t action,
	const rectangle_t *rect
)
{
	long flags;
	event_client_message_t *e;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	e = (event_client_message_t *)calloc(
	    1, sizeof(event_t)
	);
	if(e == NULL)
	    return;

	flags = (action != None) ? (1 << 0) : (0 << 0);
	if(rect != NULL)
	{
	    if((rect->width == 0) && (rect->height == 0))
		flags |= (1 << 1);
	}
	else
	    flags |= (1 << 1);

	e->type = ClientMessage;
	e->serial = 0l;
	e->send_event = False;
	e->display = gui->display;
	e->window = w;
	e->message_type = osw_atom.dnd_status;
	e->format = 32;
	e->data.l[0] = (long)tar_w;
	e->data.l[1] = flags;
	if(rect != NULL)
	{
	    e->data.l[2] = (((long)rect->x << 16) | (long)rect->y);
	    e->data.l[3] = (((long)rect->width << 16) | (long)rect->height);
	}
	e->data.l[4] = (long)action;
	XSendEvent(
	    gui->display, w, False, 0l, (XEvent *)e
	);
	free(e);
}

/*
 *	Sends a DND leave event to the target window specified by w.
 */
void OSWDNDSendLeave(win_t w, win_t src_w)
{
	event_client_message_t *e;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	e = (event_client_message_t *)calloc(
	    1, sizeof(event_t)
	);
	if(e == NULL)
	    return;

	e->type = ClientMessage;
	e->serial = 0l;
	e->send_event = False;
	e->display = gui->display;
	e->window = w;
	e->message_type = osw_atom.dnd_leave;
	e->format = 32;
	e->data.l[0] = (long)src_w;
	e->data.l[1] = 0l;		/* Reserved */
	XSendEvent(
	    gui->display, w, False, 0l, (XEvent *)e
	);
	free(e);
}

/*
 *	Sends a DND drop event to the target window specified by w.
 */
void OSWDNDSendDrop(win_t w, win_t src_w, time_t t)
{
	event_client_message_t *e;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	e = (event_client_message_t *)calloc(
	    1, sizeof(event_t)  
	);
	if(e == NULL)
	    return;

	e->type = ClientMessage;
	e->serial = 0l;
	e->send_event = False;
	e->display = gui->display;
	e->window = w;
	e->message_type = osw_atom.dnd_drop;
	e->format = 32;
	e->data.l[0] = (long)src_w;
	e->data.l[1] = 0l;		/* Reserved */
	e->data.l[2] = (long)t;
	XSendEvent(
	    gui->display, w, False, 0l, (XEvent *)e
	);
	free(e);
}

/*
 *	Sends a DND finished event to the source window specified by w.
 *
 *	The action specifies the action performed by the target or
 *	0 if drop_processed_and_successful is False.
 */
void OSWDNDSendFinished(
	win_t w, win_t tar_w,
	bool_t drop_processed_and_successful,
	atom_t action
)
{
	long flags;
	event_client_message_t *e;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	e = (event_client_message_t *)calloc(
	    1, sizeof(event_t)
	);
	if(e == NULL)
	    return;

	flags = drop_processed_and_successful ? (1 << 0) : (0 << 0);

	e->type = ClientMessage;
	e->serial = 0l;
	e->send_event = False;
	e->display = gui->display;
	e->window = w;
	e->message_type = osw_atom.dnd_finished;
	e->format = 32;
	e->data.l[0] = (long)tar_w;
	e->data.l[1] = flags;
	e->data.l[2] = drop_processed_and_successful ?
	    (long)action : 0l;
	XSendEvent(
	    gui->display, w, False, 0l, (XEvent *)e
	);
	free(e);
}

/*
 *	Begins a DND procedure.
 *
 *	The src_w specifies the source window, that is the window that
 *	the drag started on (need not be a toplevel window).
 *
 *	The type, data, and len, specifies the data to be set on the
 *	source window. The data will be set on the source window under
 *	the osw_atom.dnd_selection property.
 *
 *	The def_action specifies the default action.
 *
 *	The action_flags specifies all the allowed actions.
 */
void OSWDNDBegin(
	win_t src_w,
	atom_t target,			/* Source data type */
	const u_int8_t *data, int len,	/* Source data */
	atom_t def_action,
	dnd_action_flags_t action_flags,
	int x, int y,
	cursor_t cursor, win_t icon_w,
	time_t t
)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(src_w == None)
	    return;

	/* Stop the previous DND as needed */
	OSWDNDEnd(t);

	/* Set new DND values */
	gui->dnd_in_progress = True;
	gui->dnd_src_w = src_w;
	gui->dnd_def_action = def_action;
	gui->dnd_action_flags = action_flags;
	gui->dnd_last_x = x;
	gui->dnd_last_y = y;

	/* Set the data */
	gui->dnd_src_target = target;
	if((data != NULL) && (len > 0))
	{
	    /* Copy the specified DND data locally */
	    gui->dnd_data_len = len;
	    gui->dnd_data = (u_int8_t *)malloc(len * sizeof(u_int8_t));
	    if(gui->dnd_data != NULL)
		memcpy(gui->dnd_data, data, len);
	    else
		gui->dnd_data_len = 0;

	    /* Set the DND data on the source window */
	    XChangeProperty(
		gui->display,
		src_w,
		osw_atom.dnd_selection,
		target,
		8,				/* 8 bits */
		PropModeReplace,
		data, len
	    );
	}

	/* Grab the pointer during the entire drag with respect to the
	 * source window
	 */
	XGrabPointer(
	    gui->display,
	    src_w,
	    False,
	    ButtonReleaseMask | PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask,
	    GrabModeAsync, GrabModeAsync,
	    None,
	    cursor,
	    t
	);

	/* Set the DND icon */
	OSWDNDSetIcon(icon_w, x, y, t);
}

/*
 *	Sets the DND cursor.
 */
void OSWDNDSetCursor(cursor_t cursor, time_t t)
{
	XChangeActivePointerGrab(
	    osw_gui[0].display,
	    ButtonReleaseMask | PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask,
	    cursor,
	    t
	);
}

/*
 *	Sets the DND icon.
 */
void OSWDNDSetIcon(win_t icon_w, int x, int y, time_t t)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(gui->dnd_icon_w != icon_w)
	{
	    OSWDestroyWindow(&gui->dnd_icon_w);
	    gui->dnd_icon_w = icon_w;
	    if(icon_w != None)
	    {
		int	x_offset = 8,
			y_offset = 8;
		OSWMoveWindow(icon_w, x + x_offset, y + y_offset);
		OSWMapRaised(icon_w);
	    }
	}
}

/*
 *	Gets the DND data from the current DND selection on the source
 *	window.
 *
 *	The target specifies the data type that we want from the
 *	DND source window.
 *
 *	The calling function must delete the returned data.
 */
u_int8_t *OSWDNDGetData(
	win_t tar_w,		/* Our target window */
	atom_t target,		/* Source data type */
	int *len_rtn,
	time_t t
)
{
	const time_t start_time = time(NULL);
	event_t event;
	u_int8_t *data = NULL;
	int len = 0;
	osw_gui_struct *gui = &osw_gui[0];

	if(len_rtn != NULL)
	    *len_rtn = 0;

	if(tar_w == None)
	    return(NULL);

	/* Request value from the dnd selection */
	XConvertSelection(
	    gui->display,
	    osw_atom.dnd_selection,		/* Selection */
	    target,				/* Type */
	    osw_atom.selection_property,	/* Property */
	    tar_w,
	    t
	);

	/* Wait up to 30 seconds for response */
	while((time(NULL) - start_time) <= 30l)
	{
	    /* If a SelectionRequest is received then handle it, since
	     * it may be a request for a selection from one of our
	     * Windows
	     */
	    if(XCheckTypedEvent(
		gui->display, SelectionRequest, &event
	    ))
		OSWManageSelectionEvent(gui, &event);

	    /* Check if there is a SelectionNotify for our requesting
	     * Window
	     */
	    if(XCheckTypedWindowEvent(
		gui->display, tar_w, SelectionNotify, &event
	    ))
	    {
		/* The requested data has arrived */
		XSelectionEvent *sel_notify = &event.xselection;

		/* Selection value available? */
		if(sel_notify->property != None)
		{
		    /* Get the selection value */
		    Atom v_target;
		    int v_format, bytes_read = 0;
		    unsigned long nitems, bytes_after;
		    unsigned char *v;

		    do {
			if(XGetWindowProperty(
			    sel_notify->display,
			    sel_notify->requestor,
			    sel_notify->property,
			    (long)(bytes_read / 4),     /* Offset */
			    (long)(1024 / 4),           /* Length */
			    True,                       /* Delete */
			    AnyPropertyType,
			    &v_target,                  /* Type return */
			    &v_format,                  /* Element size (in bits) */
			    &nitems,
			    &bytes_after,
			    &v                          /* Value return */
			) != Success)
			    break;

			/* Make sure that the received type and format
			 * matches our requested type and format
			 */
			if((v_target == target) &&
			   (v_format == 8) && (nitems > 0)
			)
			{
			    const int pos = len;

			    len += (int)nitems * (v_format / 8);
			    data = (u_int8_t *)realloc(
				data,
				nitems * sizeof(u_int8_t)
			    );
			    if(data != NULL)
				memcpy(data + pos, v, nitems);
			    else
				len = 0;
			}

			/* Delete the received selection segment value */
			if(v != NULL)
			    XFree(v);

			bytes_read += (int)nitems * (v_format / 8);

		    } while(bytes_after > 0);
		}
		break;
	    }
	}

	if(len_rtn != NULL)
	    *len_rtn = len;

	return(data);
}

/*
 *	Moves the icon window specified by icon_w from its current
 *	position to its original position.
 *
 *	The original position is the coordinates of the origin of the
 *	drag relative to the root window.
 *
 *	The current position is the current coordinates of the drag
 *	cursor relative to the root window.
 */
void OSWDNDIconAnimateReturn(
	win_t icon_w,
	int cur_x, int cur_y, 
	int orig_x, int orig_y
)
{
	int x, y, dx, dy;
	time_t t, t2;
	const time_t dt = 500l, t0 = OSWTimeMS();
	float c;

	if(icon_w == None)
	    return;

	x = cur_x;
	y = cur_y;
	dx = orig_x - x;
	dy = orig_y - y;
	t = t0;
	do {

	    t2 = t;

	    OSWMoveWindow(icon_w, x, y);
	    OSWFlush();
	    usleep(1000l);

	    t = OSWTimeMS();
	    c = (float)(t - t2) / (float)dt;
	    x += (int)(dx * c);
	    y += (int)(dy * c);

	} while((t0 + dt) > t);
}

/*
 *	Ends a DND procedure.
 */
void OSWDNDEnd(time_t t)
{
	osw_gui_struct *gui = &osw_gui[0];

	/* Ungrab the pointer */
	XUngrabPointer(gui->display, (Time)t);

	/* Delete the DND data from the source window */
	if(gui->dnd_src_w != None)
	    XDeleteProperty(
		gui->display,
		gui->dnd_src_w,
		osw_atom.dnd_selection
	    );

	gui->dnd_in_progress = False;
	gui->dnd_src_w = None;
	gui->dnd_tar_w = None;
	gui->dnd_src_target = None;
	free(gui->dnd_data);		/* Delete the local DND data */
	gui->dnd_data = NULL;
	gui->dnd_data_len = 0;
	gui->dnd_def_action = None;
	gui->dnd_tar_action = None;
	gui->dnd_action_flags = 0;
	OSWDestroyWindow(&gui->dnd_icon_w);
	gui->dnd_last_x = 0;
	gui->dnd_last_y = 0;
}


/*
 *	Returns a visual pointer by a match given in the criteria
 *	for a visual info. 
 */
visual_t *OSWGetVisualByCriteria(
	visual_info_mask_t vinfo_mask,
	visual_info_t criteria_vinfo
)
{               
	int num_rtn;
	visual_t *visual_rtn;
	visual_info_t *vinfo_rtn = XGetVisualInfo(
	    osw_gui[0].display,
	    vinfo_mask,
	    &criteria_vinfo,
	    &num_rtn
	);

	if((num_rtn <= 0) || (vinfo_rtn == NULL))
	    return(NULL);

	/* Got match */
	visual_rtn = vinfo_rtn->visual;

	/* Need to free the vinfo return */        
	XFree(vinfo_rtn);

	return(visual_rtn);
}           

/*
 *	Get visual information by ID.
 */
visual_t *OSWGetVisualByID(visual_id_t vid)
{
	int x;
	visual_info_t vinfo_ref;
	visual_info_t *vinfo_rtn;
	visual_t *vptr;

	vinfo_ref.visualid = vid;
	vinfo_rtn = XGetVisualInfo(
	    osw_gui[0].display,
	    VisualIDMask,
	    &vinfo_ref,
	    &x
	);
	if(vinfo_rtn == NULL)
	    return(NULL);

	/* Get visual pointer */
	vptr = vinfo_rtn->visual;

	/* Free visual info return */
	XFree(vinfo_rtn);

	return(vptr);
}


/*
 *	Loads the font by the specified name.
 */
int OSWLoadFont(
	font_t **font_rtn,
	const char *name
)
{
	font_t *font;
	XCharStruct *xchar;
	XFontStruct *xfont;
	osw_gui_struct *gui = &osw_gui[0];

	if((font_rtn == NULL) || (name == NULL))
	    return(-1);

	*font_rtn = font = (font_t *)calloc(1, sizeof(font_t));
	if(font == NULL)
	    return(-3);

	/* Load the xfont */
	font->xfont = xfont = XLoadQueryFont(gui->display, name);
	if(xfont == NULL)
	{
	    /* No such font */
	    free(font);
	    *font_rtn = font = NULL;
	    return(-2);
	}

	xchar = &xfont->max_bounds;
	font->max_lbearing = xchar->lbearing;
	font->max_rbearing = xchar->rbearing;
	font->max_cwidth = xchar->width;
	font->ascent = xfont->ascent;
	font->descent = xfont->descent;

	return(0);
}


/*
 *	Returns the current font pointer set in the graphics
 *	context foreground or NULL on error.
 */
font_t *OSWQueryCurrentFont(void)
{
	return(osw_gui[0].current_font);
}

/*
 *	Returns the geometry of the specified character.
 */
bool_t OSWGetCharBounds(
	font_t *font, char c,
	int *lbearing, int *rbearing,
	int *width,
	int *ascent, int *descent
)
{
	XFontStruct *xfont;
	XCharStruct *cbounds;

	if(ascent != NULL)
	    *ascent = 0;
	if(descent != NULL)
	    *descent = 0;
	if(lbearing != NULL)
	    *lbearing = 0;
	if(rbearing != NULL)
	    *rbearing = 0;
	if(width != NULL)
	    *width = 0;

	if(font == NULL)
	    return(False);

	/* No font? */
	xfont = font->xfont;
	if(xfont == NULL)
	    return(False);

	if((xfont->min_byte1 == 0) &&         
	   (xfont->max_byte1 == 0) &&
	   ((unsigned int)c >= xfont->min_char_or_byte2) &&
	   ((unsigned int)c <= xfont->max_char_or_byte2)
	)
	{
	    XCharStruct *list = xfont->per_char;
	    if(list != NULL)
		cbounds = &list[c - xfont->min_char_or_byte2];
	    else
		cbounds = &xfont->max_bounds;
	}
	else
	{
	    cbounds = &xfont->max_bounds;
	}

	if(ascent != NULL)
	    *ascent = font->ascent;
	if(descent != NULL)
	    *descent = font->descent;
	if(lbearing != NULL)
	    *lbearing = (int)cbounds->lbearing;
	if(rbearing != NULL)
	    *rbearing = (int)cbounds->rbearing;
	if(width != NULL)
	    *width = (int)cbounds->width;

	return(True);
}

/*
 *	Returns the geometry of the specified text.
 */
bool_t OSWGetTextBounds(
	font_t *font, const char *text, int text_len,
	int *lbearing, int *rbearing,
	int *width,
	int *ascent, int *descent
)
{
	int direction, font_ascent, font_descent;
	XFontStruct *xfont;
	XCharStruct overall;
	
	if(ascent != NULL)
	    *ascent = 0;
	if(descent != NULL)
	    *descent = 0;
	if(lbearing != NULL)
	    *lbearing = 0;
	if(rbearing != NULL)                  
	    *rbearing = 0;
	if(width != NULL)
	    *width = 0;
	
	if((font == NULL) || (text == NULL))
	    return(False);

	/* Calculate entire length of text? */
	if(text_len < 0)
	    text_len = strlen(text);

	/* No font? */
	xfont = font->xfont;
	if(xfont == NULL)
	    return(False);

	/* If length is 0 then just return the font height */
	if(text_len == 0)
	{
	    /* Use the font's ascent and descent as the return */
	    if(ascent != NULL)
		*ascent = font->ascent;
	    if(descent != NULL)
		*descent = font->descent;
	    return(True);
	}

	if((xfont->min_byte1 == 0) && (xfont->max_byte1 == 0))
	    XTextExtents(
		xfont, text, text_len,
		&direction, &font_ascent, &font_descent,
		&overall
	    );
	else
	    XTextExtents16(
		xfont, (XChar2b *)text, text_len / 2,
		&direction, &font_ascent, &font_descent,
		&overall
	    );

	/* Use the font's ascent and descent as the return */
	if(ascent != NULL)
	    *ascent = font->ascent;
	if(descent != NULL)
	    *descent = font->descent;

	if(lbearing != NULL)
	    *lbearing = (int)overall.lbearing;
	if(rbearing != NULL)
	    *rbearing = (int)overall.rbearing;
	if(width != NULL)
	    *width = (int)overall.width;

	return(True);
} 


/*
 *	Set font for the graphics context foreground.
 */
void OSWSetFont(font_t *font)
{
	XFontStruct *xfont;
	osw_gui_struct *gui = &osw_gui[0];

	if(font == NULL)
	    return;

	xfont = font->xfont;
	if(xfont == NULL)
	    return;

	/* No change in font pointers? */
	if(font == gui->current_font)
	    return;

	/* Set new font */
	XSetFont(
	    gui->display,
	    gui->gc,
	    xfont->fid
	);

	/* Record current font */
	gui->current_font = font;
}

/*
 *	Unload font.
 */
void OSWUnloadFont(font_t **font)
{
	XFontStruct *xfont;
	osw_gui_struct *gui = &osw_gui[0];

	if(font == NULL)
	    return;

	if(*font == NULL)
	    return;

	/* Unset current_font as needed */
	if(*font == gui->current_font)
	    gui->current_font = NULL;

	/* Unload font */
	xfont = (*font)->xfont;
	if(xfont != NULL)
	    XFreeFont(gui->display, xfont);

	free(*font);
	*font = NULL;
}


/*
 *	Load a color pixel by RGB values.
 *
 *	Valid values for r, g, and b range from 0 to 255.
 *
 *      Returns 0 on success, -1 on general error, and -2 on
 *      ambiguous or missing data.
 */
int OSWLoadPixelRGB(
	pixel_t *pix_rtn,
	u_int8_t r,
	u_int8_t g,
	u_int8_t b
)
{
	int i;
	XColor c;
	osw_gui_struct *gui = &osw_gui[0];

	if(pix_rtn == NULL)
	    return(-1);

	if(gui->we_created_colormap)
	{
	    /* We created color map, fetch pixel from our values */
	    switch(gui->depth)
	    {
	      case 8:
		*pix_rtn = gui->black_pix;

		for(i = 0; i < 256; i++)
		{
		    c.blue = ((i & 0x03) << (5));
		    c.green = ((i & 0x1C) << (3));
		    c.red = ((i & 0xE0) << (0));

		    if((c.blue >= b) &&
		       (c.green >= g) &&
		       (c.red >= r)
		    )
		    {
			*pix_rtn = (pixel_t)i;
			break;
		    }
		}
		if(i >= 256)
		    *pix_rtn = 255;
		return(0);
		break;

	      default:
		*pix_rtn = gui->black_pix;
		return(0);
		break;
	    }
	}
	else
	{
	    /* Using default color map */

	    /* Set colors to be allocated */
	    c.flags	= DoRed | DoGreen | DoBlue;
	    c.red	= ((u_int16_t)r << 8) | (u_int16_t)r;
	    c.green	= ((u_int16_t)g << 8) | (u_int16_t)g;
	    c.blue	= ((u_int16_t)b << 8) | (u_int16_t)b;

	    /* Allocate color */
	    if(XAllocColor(gui->display, gui->colormap, &c))
	    {
		/* True, successful pixel allocation */
		*pix_rtn = c.pixel;
		return(0);
	    }

	    /* False, failed allocation, use global black_pix */
	    *pix_rtn = gui->black_pix;
	}

	return(-1);
}


/*
 *      Load a color pixel by HSL values.
 *
 *      Valid values for h, s, and l range from 0 to 255.
 * 
 *      Returns 0 on success, -1 on general error, an\nd -2 on
 *      ambiguous or missing data.
 */
int OSWLoadPixelHSL(
	pixel_t *pix_rtn,
	u_int8_t h,
	u_int8_t s,
	u_int8_t l 
)
{


	return(-1);
}



/*
 *	Load pixel by given color specification string (CLSP).
 *	Returns zero on success, -1 on
 *
 *	The format for clsp is "rgbi:%lf/%lf/%lf".
 *	(The input format for clsp is native to Xlib.)
 *
 *	Returns 0 on success, -1 on general error, an\nd -2 on
 *	ambiguous or missing data.
 */
int OSWLoadPixelCLSP(pixel_t *pix_rtn, const char *clsp)
{
	XColor c;

	if(pix_rtn == NULL)
	    return(-1);

	if(clsp == NULL)
	{
	    *pix_rtn = 0;
	    return(-1);
	}

	switch(XParseColor(
	    osw_gui[0].display,
	    osw_gui[0].colormap,
	    clsp, &c
	))
	{
	  case BadColor:
	    fprintf(stderr,
		"OSWLoadPixelCLSP(): %s: BadColor\n",
		clsp
	    );
	    *pix_rtn = 0;
	    return(-1);
	    break;

	  case BadValue:
	    fprintf(stderr,
		"OSWLoadPixelCLSP(): %s: BadValue\n",
		clsp
	    );
	    *pix_rtn = 0;
	    return(-1);
	    break;

	  default:
	    return(OSWLoadPixelRGB(
		pix_rtn,
		(c.red >> 8),
		(c.green >> 8),
		(c.blue >> 8)
	    ));
	    break;
	}

	return(0);
}

/*
 *	Sets the graphic context's foreground pixel.
 */
void OSWSetForeground(pixel_t p)
{
	osw_gui_struct *gui = &osw_gui[0];
	XSetForeground(gui->display, gui->gc, p);
	gui->current_fg_pix = p;
}

/*
 *	Deletes the pixel.
 */
void OSWDestroyPixel(pixel_t *pix_ptr)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(pix_ptr == NULL)
	    return;

	/* Leave default black and white pixels alone */
	if((*pix_ptr == osw_gui[0].black_pix) ||
	   (*pix_ptr == osw_gui[0].white_pix)
	)
	    return;

	/* Unset current pixels as needed */
	if(osw_gui[0].current_fg_pix == *pix_ptr)
	    osw_gui[0].current_fg_pix = 0;

	/* Free pixel */
	if(*pix_ptr != 0)
	{
	    /* When we create the colormap, the pixels are not
	     * allocated, therefore we do not need to free them.
	     * See function OSWLoadPixelRGB() for example.
	     */
	    if(!gui->we_created_colormap)
		XFreeColors(
		    osw_gui[0].display,
		    osw_gui[0].colormap,
		    pix_ptr,	/* Pixel array */
		    1,		/* Number of pixels */
		    0		/* Planes */
	        );

	    /* Reset pixel value to 0 */
	    *pix_ptr = 0;
	}
}


/*
 *	Opens the cursor from a standard GUI internally defined data
 *	specified by id.
 */
cursor_t OSWCursorOpenID(cursor_id_t id)
{
	return(XCreateFontCursor(osw_gui[0].display, id));
}

/*
 *	Opens the cursor from the XPM file.
 */
cursor_t OSWCursorOpenXPMFile(
	const char *path,
	int hot_x, int hot_y 
)
{
#ifdef HAVE_LIBXPM
	XColor	fg_xc = { 0l, 0xffff, 0xffff, 0xffff },
		bg_xc = { 0l, 0x0000, 0x0000, 0x0000 };
	osw_gui_struct *gui = &osw_gui[0];
	int status;
	unsigned int    width, height,
			req_width, req_height;
	char *dpath;
	display_t *dpy = gui->display;
	cursor_t cur;
	bitmap_t mask;
	pixmap_t pixmap;
	win_t w = gui->root_win;
	XpmAttributes xpmattr;

	if((w == None) || (path == NULL))
	    return(None);

	/* Set up XPM attributes */
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.closeness = 40000;
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;

	/* We need to load the XPM data to a bitmap in order to
	 * create a cursor, so the depth must be 1 bit
	 */
	xpmattr.depth = 1;

	/* Open the XPM file to a pixmap and mask */
	dpath = strdup(path);
	status = XpmReadFileToPixmap(
	    dpy, w,
	    dpath,
	    &pixmap,
	    &mask,
	    &xpmattr
	);
	free(dpath);
	if(status != XpmSuccess)
	    return(None);

	/* Get size of loaded pixmap and mask */
	width = xpmattr.width;
	height = xpmattr.height;

	/* Get the recommend cursor size */
	if(!XQueryBestCursor(
	    dpy, w,
	    width, height,
	    &req_width, &req_height
	))
	{  
	    req_width = width;
	    req_height = height;
	}

	/* Need to reduce size of pixmap and mask? */
	if((width > req_width) || (height > req_height))
	{
	    bitmap_t tmp_mask;
	    pixmap_t tmp_pixmap;
	    gc_val_t gcv;
	    gc_t gc;

	    /* Create resized pixmap and mask */
	    OSWCreateBitmap(&tmp_mask, req_width, req_height);
	    OSWCreatePixmap(&tmp_pixmap, req_width, req_height);

	    /* Copy to the resized pixmap */
	    if(tmp_pixmap != None)
	    { 
		XCopyArea(
		    dpy, 
		    pixmap,             /* Src */
		    tmp_pixmap,         /* Tar */
		    gui->gc,  
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyPixmap(&pixmap);
		pixmap = tmp_pixmap;
	    }

	    /* Create gc for the mask and copy to the resized mask */
	    gcv.function = GXcopy;
	    gcv.plane_mask = 1;		/* Monochrome */
	    gcv.foreground = gui->white_pix;
	    gcv.background = gui->black_pix;
	    gc = XCreateGC(
		dpy,
		mask,
		GCFunction | GCPlaneMask |
		GCForeground | GCBackground,
		&gcv
	    );
	    if((tmp_mask != None) && (gc != None))
	    {
		XCopyArea(
		    dpy,
		    mask,               /* Source */
		    tmp_mask,           /* Target */
		    gc,  
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyBitmap(&mask);
		mask = tmp_mask;
	    }

	    XFreeGC(dpy, gc);
	}
	 
	/* Create the cursor */
	cur = XCreatePixmapCursor(
	    dpy,
	    pixmap,
	    mask,
	    &fg_xc,
	    &bg_xc,
	    hot_x, hot_y
	);

	/* Delete the pixmap and mask */
	OSWDestroyPixmap(&pixmap);
	OSWDestroyBitmap(&mask);

	return(cur);
#else
#warning OSWCursorOpenXPMFile() disabled, HAVE_LIBXPM was not defined
	return(None);
#endif
}

/*
 *	Opens the cursor from the XPM data.
 */
cursor_t OSWCursorOpenXPMData(
	char **xpmdata,
	int hot_x, int hot_y 
)
{
#ifdef HAVE_LIBXPM
	XColor	fg_xc = { 0l, 0xffff, 0xffff, 0xffff },
		bg_xc = { 0l, 0x0000, 0x0000, 0x0000 };
	osw_gui_struct *gui = &osw_gui[0];
	int status;
	unsigned int    width, height,
			req_width, req_height;
	display_t *dpy = gui->display;
	cursor_t cur;
	bitmap_t mask;
	pixmap_t pixmap;
	win_t w = gui->root_win;
	XpmAttributes xpmattr;

	if((dpy == NULL) || (w == None) || (xpmdata == NULL))
	    return(None);

	/* Set up XPM attributes */
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.closeness = 40000;
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;

	/* We need to load the XPM data to a bitmap in order to
	 * create a cursor, so the depth must be 1 bit
	 */
	xpmattr.depth = 1;

	/* Load the XPM data to a pixmap and mask */
	status = XpmCreatePixmapFromData(
	    dpy, w,
	    xpmdata,
	    &pixmap,
	    &mask,
	    &xpmattr
	);
	if(status != XpmSuccess)
	    return(None);

	/* Get size of loaded pixmap and mask */
	width = xpmattr.width;
	height = xpmattr.height;

	/* Get the recommend cursor size */
	if(!XQueryBestCursor(
	    dpy, w,
	    width, height,
	    &req_width, &req_height
	))
	{
	    req_width = width;
	    req_height = height;
	}
	 
	/* Need to reduce size of pixmap and mask? */
	if((width > req_width) || (height > req_height))
	{
	    bitmap_t tmp_mask;
	    pixmap_t tmp_pixmap;
	    gc_val_t gcv;
	    gc_t gc;

	    /* Create resized pixmap and mask */
	    OSWCreateBitmap(&tmp_mask, req_width, req_height);
	    OSWCreatePixmap(&tmp_pixmap, req_width, req_height);

	    /* Copy to the resized pixmap */
	    if(tmp_pixmap != None)
	    { 
		XCopyArea(
		    dpy,
		    pixmap,             /* Src */
		    tmp_pixmap,         /* Tar */
		    gui->gc,  
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyPixmap(&pixmap);
		pixmap = tmp_pixmap;
	    }


	    /* Create gc for the mask and copy to the resized mask */
	    gcv.function = GXcopy;
	    gcv.plane_mask = 1;		/* Monochrome */
	    gcv.foreground = gui->white_pix;
	    gcv.background = gui->black_pix;
	    gc = XCreateGC(
		dpy,
		mask,
		GCFunction | GCPlaneMask |
		GCForeground | GCBackground,
		&gcv
	    );
	    if((tmp_mask != None) && (gc != None))
	    {
		XCopyArea(
		    dpy,  
		    mask,               /* Source */
		    tmp_mask,           /* Target */
		    gc,
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyBitmap(&mask);
		mask = tmp_mask;
	    }

	    XFreeGC(dpy, gc);
	}
	 
	/* Create the cursor */
	cur = XCreatePixmapCursor(
	    dpy,
	    pixmap,
	    mask,
	    &fg_xc,
	    &bg_xc,
	    hot_x, hot_y
	);

	/* Delete the pixmap and mask */
	OSWDestroyPixmap(&pixmap);
	OSWDestroyBitmap(&mask);

	return(cur);
#else
#warning OSWCursorOpenXPMFile() disabled, HAVE_LIBXPM was not defined
	return(None);
#endif
}

/*
 *	Sets the window's cursor.
 */
void OSWSetWindowCursor(win_t w, cursor_t cur)
{
	if(w == None)
	    return;

	if(cur != None)
	    XDefineCursor(osw_gui[0].display, w, cur);
	else
	    XUndefineCursor(osw_gui[0].display, w);
}

/*
 *	Unset cursor on window, window now gets cursor defined on its
 *	parent.
 */
void OSWUnsetWindowCursor(win_t w)
{
	if(w == None)
	    return;

	XUndefineCursor(osw_gui[0].display, w);
}

/*
 *	Destroys the cursor.
 */
void OSWDestroyCursor(cursor_t *cur_rtn)
{
	if(cur_rtn == NULL)
	    return;

	if(*cur_rtn != None)
	{
	    XFreeCursor(osw_gui[0].display, *cur_rtn);
	    *cur_rtn = None;
	}
}


/*
 *	Get the number of pending events.
 */
int OSWEventsPending(void)
{
	return(XPending(osw_gui[0].display));
}


/*
 *	Preprocesses the event, allows our internal routines to
 *	manage this event before giving it to the calling function
 */
#define GUI_PREPROCESS_EVENT(_gui_,_event_)	{	\
 OSWManageKeyEvent((_gui_), (_event_));			\
 OSWManageSelectionEvent((_gui_), (_event_));		\
 OSWManageDNDEvent((_gui_), (_event_));			\
 OSWTranslateModalEvent((_gui_), (_event_));		\
 OSWTranslateBlockInputEvent((_gui_), (_event_));	\
}

/*
 *	Blocks until the next event is recieved.
 */
void OSWWaitNextEvent(event_t *event)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(event == NULL)
	    return;

	XNextEvent(gui->display, event);
	GUI_PREPROCESS_EVENT(gui, event);
}

/*
 *	Blocks until and peak at the next event.
 */
void OSWWaitPeakEvent(event_t *event)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(event == NULL)
	    return;

	XPeekEvent(gui->display, event);
	GUI_PREPROCESS_EVENT(gui, event);
}

/*
 *	Checks the event queue for an event type matching the
 *	specified eventmask.
 *
 *	If an event is found, it will be removed from the queue
 *	and coppied to the event, returning True.
 *	If no event is found, then False is returned.
 */
bool_t OSWCheckMaskEvent(
	const event_mask_t eventmask,
	event_t *event
)
{
	bool_t status;
	osw_gui_struct *gui = &osw_gui[0];

	if(event == NULL)
	    return(False);

	status = XCheckMaskEvent(
	    gui->display,
	    eventmask,
	    event
	);
	if(status)
	{
	    GUI_PREPROCESS_EVENT(gui, event);
	}
	else
	{
	    memset(event, 0x00, sizeof(event_t));
	}

	return(status);
}       

/*
 *	The OSWWaitWindowEvent function searches the event queue for an event
 *	that matches both the specified window and event mask.  When it finds a
 *      match, OSWWaitWindowEvent removes that event from the queue and copies
 *	it into the specified event_t structure.  The other events stored in
 *	the queue are not discarded.  If a matching event is not in the queue,
 *	OSWWaitWindowEvent flushes the output buffer and blocks until one is
 *	received.
 */
void OSWWaitWindowEvent(
	win_t w,
	const event_mask_t event_mask,
	event_t *event
)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(event == NULL)
	    return;

	if(w == None)
	{
	    memset(event, 0x00, sizeof(event_t));
	    return;
	}

	XWindowEvent(
	    gui->display,
	    w,
	    event_mask,
	    event
	);
	GUI_PREPROCESS_EVENT(gui, event);
}

/*
 *	Checks event matching event_mask on window w.
 *	Returns True if matched event is found.
 */
bool_t OSWCheckWindowEvent(win_t w, event_mask_t event_mask, event_t *event)
{
	bool_t status;
	osw_gui_struct *gui = &osw_gui[0];

	if(event == NULL)
	    return(False);

	if(w == None)
	{
	    memset(event, 0x00, sizeof(event_t));
	    return(False);
	}

	status = XCheckWindowEvent(
	    gui->display,
	    w,
	    event_mask,
	    event
	);
	if(status)
	{
	    GUI_PREPROCESS_EVENT(gui, event);
	}
	else
	{
	    memset(event, 0x00, sizeof(event_t));
	}

	return(status);
} 

#undef GUI_PREPROCESS_EVENT

/*
 *	Puts the given event which was fetched earlier back onto
 *	the queue.   The given event should have been originally
 *	fetched from one of the event recieving functions and not
 *	just `made up'.
 */
void OSWPutBackEvent(event_t *event)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(event == NULL)
	    return;

	XPutBackEvent(gui->display, event);
}

/*
 *	Sends an event to the GUI, which can then be recieved by
 *	any of the event fetching functions.
 *
 *	Event members xany.window and type needs to be specified
 *	properly in the event_t structure.
 */
int OSWSendEvent(
	event_mask_t mask,
	event_t *event,
	bool_t propagate
)
{
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];

	if(event == NULL)
	    return(-1);

	w = event->xany.window;
	if(w == None)
	    return(-1);

	return(XSendEvent(
	    gui->display,
	    w,
	    propagate,
	    mask,
	    event
	) ? 0 : -1);
}

/*
 *	Purges all events (if any) in the queue.
 */
int OSWPurgeAllEvents(void)
{
	int nevents_purged = 0;
	event_t event;

	while(OSWEventsPending() > 0)
	{
	    OSWWaitNextEvent(&event);
	    nevents_purged++;
	}

	return(nevents_purged);
}

/*
 *	Purges event table only of events matching event_type.
 */
int OSWPurgeTypedEvent(const int event_type)
{
	int nevents_purged = 0;
	event_t event;

	while(XCheckTypedEvent(
	    osw_gui[0].display,
	    event_type,
	    &event
	))
	    nevents_purged++;

	return(nevents_purged);
}

/*      
 *	Purges the event table only of events matching event_type for
 *	a specific window.
 */
int OSWPurgeWindowTypedEvent(
	win_t w,
	const event_type_t event_type
)
{
	int nevents_purged = 0;
	event_t event;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return(nevents_purged);

	while(XCheckTypedWindowEvent(
	    gui->display,
	    w,
	    event_type,
	    &event
	))
	    nevents_purged++;

	return(nevents_purged);
}

/* 
 *	If the block input level is > 0 then this function will
 *	modify any; KeyPress, KeyRelease, ButtonPress, ButtonRelease,
 *	MotionNotify, FocusIn, and EnterNotify events to be relative
 *	to the root window.
 *
 *	This effectively make functions processing this even think that
 *	it is not for them.
 */
static void OSWTranslateBlockInputEvent(
	osw_gui_struct *gui,
	event_t *event
)
{
	int x, y;
	win_t root = gui->root_win;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	event_crossing_t *crossing;

	/* Not blocking input? */
	if(gui->block_input_level <= 0)
	    return;

	switch(event->type)
	{
	  case KeyPress:
	  case KeyRelease:
	    key = &event->xkey;
	    XTranslateCoordinates(
		key->display,
		key->window,
		root,
		key->x, key->y,
		&x, &y,
		&key->subwindow
	    );
	    key->x = x;
	    key->y = y;
	    key->window = root;
	    key->keycode = 0;			/* Clear the key value */
	    key->state = 0;			/* Clear the key state */
	    break;

	  case ButtonPress:
	  case ButtonRelease:
	    button = &event->xbutton;
	    XTranslateCoordinates(
		button->display,
		button->window,
		root,
		button->x, button->y,
		&x, &y,
		&button->subwindow
	    );
	    button->x = x;
	    button->y = y;
	    button->window = root;
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    XTranslateCoordinates(
		motion->display,
		motion->window,
		root,
		motion->x, motion->y,
		&x, &y,
		&motion->subwindow
	    );
	    motion->x = x;
	    motion->y = y;
	    motion->window = root;
	    break;

	  case EnterNotify:
	    crossing = &event->xcrossing;
	    XTranslateCoordinates(
		crossing->display,
		crossing->window,
		root,
		crossing->x, crossing->y,
		&x, &y,
		&crossing->subwindow 
	    );
	    crossing->x = x;
	    crossing->y = y;
	    crossing->window = root;
	    break;
	}
}

/*
 *	Increases or decreases the block input level.
 */
void OSWBlockInput(const bool_t block)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(block)
	{
	    gui->block_input_level++;
	}
	else
	{
	    gui->block_input_level--;
	    if(gui->block_input_level < 0)
		gui->block_input_level = 0;
	}
}

/*
 *	Returns the block input level.
 */
int OSWBlockInputLevel(void)
{
	return(osw_gui[0].block_input_level);
}

/*
 *	Checks if the ClientMessage event specified by event is a
 *	request to take focus of the window specified by w.
 *
 *	The specified event must be a ClientMessage event.
 */
bool_t OSWEventIsTakeFocus(
	win_t w,
	event_t *event,
	time_t *t_rtn
)
{
	event_client_message_t *client;

	if(t_rtn != NULL)
	    *t_rtn = 0l;

	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;

	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.wm_protocols) &&
	   (client->format == 32)
	)
	{
	    if((atom_t)client->data.l[0] == osw_atom.wm_take_focus)
	    {
		if(t_rtn != NULL)
		    *t_rtn = (time_t)client->data.l[1];
		return(True);
	    }
	    else
		return(False);
	}
	else
	    return(False);
}

/*
 *	Checks if the ClientMessage event specified by event is a
 *	request to delete the window specified by w.
 *
 *	The specified event must be a ClientMessage event.
 */
bool_t OSWEventIsDeleteWindow(
	win_t w,
	event_t *event
)
{
	event_client_message_t *client;

	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;
 
	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.wm_protocols) &&
	   (client->format == 32)
	)
	    return(
		((atom_t)client->data.l[0] == osw_atom.wm_delete_window) ?
		True : False
	    );
	else
	    return(False);
}

/*
 *	Checks if the specified event is a dnd enter event and if
 *	one of its targets matches the specified target.
 *
 *	The target specifies the acceptable drag data target type,
 *	if the specified target is not found on the source window's
 *	list of target types then False is returned.
 *
 *	The specified event must be a ClientMessage event.
 */
bool_t OSWEventIsDNDEnter(
	win_t w, event_t *event,
	win_t *src_w_rtn,
	const atom_t *target, int ntargets
)
{
	event_client_message_t *client;
	osw_gui_struct *gui = &osw_gui[0];

	if(src_w_rtn != NULL)
	    *src_w_rtn = None;

	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;

	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.dnd_enter) &&
	   (client->format == 32)
	)
	{
	    bool_t target_accepted = False;
	    win_t src_w = (win_t)client->data.l[0];
	    const long flags = client->data.l[1];

#define CHECK_TARGET(_a_)	{	\
 if(!target_accepted) {			\
  int _i;				\
  for(_i = 0; _i < ntargets; _i++) {	\
   if(target[_i] == (atom_t)(_a_)) {	\
    target_accepted = True;		\
    break;				\
   }					\
  }					\
 }					\
}

	    if(src_w_rtn != NULL)
		*src_w_rtn = src_w;

	    /* More than 3 target types? */
	    if((flags & (1 << 0)) && (src_w != None))
	    {
		atom_t type;
		int format;
		unsigned long nitems, bytes_remaining;
		unsigned char *v;

		/* Get the target types list from the source window */
		if(XGetWindowProperty(
		    gui->display,
		    src_w,
		    osw_atom.dnd_type_list,
		    0l,
		    (long)(65536 / 4),
		    False,
		    XA_ATOM,
		    &type,
		    &format,
		    &nitems, &bytes_remaining,
		    &v
		) == Success)
		{
		    /* Got the target types list, now check if the
		     * specified target is in the list
		     */
		    if(v != NULL)
		    {
			if((type == XA_ATOM) && (format == 32))
			{
			    int i;
			    atom_t a;
			    const u_int32_t *v32 = (u_int32_t *)v;
			    for(i = 0; i < (int)nitems; i++)
			    {
				a = (atom_t)v32[i];
				CHECK_TARGET(a);
				if(target_accepted)
				    break;
			    }
			}
			XFree(v);
		    }
		    return(target_accepted);
		}
		return(False);
	    }

	    /* Check if one of the targets matches the specified target */
	    CHECK_TARGET(client->data.l[2]);
	    CHECK_TARGET(client->data.l[3]);
	    CHECK_TARGET(client->data.l[4]);

	    return(target_accepted);
#undef CHECK_TARGET
	}
	else             
	    return(False);
}

/*
 *	Checks if the specified event is a dnd position event.
 *
 *	The specified event must be a ClientMessage event.
 */
bool_t OSWEventIsDNDPosition(
	win_t w, event_t *event, 
	win_t *src_w_rtn, win_t *tar_child_w_rtn,
	int *x_rtn, int *y_rtn,
	int *root_x_rtn, int *root_y_rtn,
	time_t *t,
	atom_t *action_rtn
)
{
	event_client_message_t *client;
	osw_gui_struct *gui = &osw_gui[0];

	if(src_w_rtn != NULL)
	    *src_w_rtn = None;
	if(tar_child_w_rtn != NULL)
	    *tar_child_w_rtn = None;
	if(x_rtn != NULL)
	    *x_rtn = 0;
	if(y_rtn != NULL)
	    *y_rtn = 0;
	if(root_x_rtn != NULL)
	    *root_x_rtn = 0;
	if(root_y_rtn != NULL)
	    *root_y_rtn = 0;
	if(t != NULL)
	    *t = 0l;
	if(action_rtn != NULL)
	    *action_rtn = None;
			   
	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;

	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.dnd_position) &&
	   (client->format == 32)
	)
	{
	    const long coord = client->data.l[2];
	    int16_t	root_x = (int16_t)((coord >> 16) & 0x0000ffff),
			root_y = (int16_t)((coord >> 0) & 0x0000ffff);

	    if(src_w_rtn != NULL)
		*src_w_rtn = (win_t)client->data.l[0];

	    if(root_x_rtn != NULL)
		*root_x_rtn = (int)root_x;
	    if(root_y_rtn != NULL)
		*root_y_rtn = (int)root_y;

	    if((x_rtn != NULL) || (y_rtn != NULL) || (tar_child_w_rtn != NULL))
	    {
		int x, y;
		win_t child = w, next_child = None;

		do {

		    if(!XTranslateCoordinates(
			gui->display,
			gui->root_win,		/* Source window */
			child,			/* Target window */
			(int)root_x, (int)root_y,	/* Source coordinates */
			&x, &y,			/* Target coordinates */
			&next_child
		    ))
			break;

		    if(x_rtn != NULL)
			*x_rtn = x;
		    if(y_rtn != NULL)
		        *y_rtn = y;
		    if(tar_child_w_rtn != NULL)
			*tar_child_w_rtn = child;

		    child = next_child;

		} while(child != None);
	    }

	    if(t != NULL)
		*t = (time_t)client->data.l[3];

	    if(action_rtn != NULL)
		*action_rtn = (atom_t)client->data.l[4];

	    return(True);
	}
	else
	    return(False);
}

/*
 *	Checks if the specified event is a dnd status event and if
 *
 *	The specified event must be a ClientMessage event.
 *
 *	If *accepted_action_rtn is None then it means the drop is not
 *	accepted.
 *
 *	If the width and height of rect_rtn are 0 then it means no
 *	rectangle has been defined and to always send positions.
 */
bool_t OSWEventIsDNDStatus(
	win_t w, event_t *event,
	win_t *tar_w_rtn, atom_t *accepted_action_rtn,
	rectangle_t *rect_rtn
)
{
	event_client_message_t *client;

	if(tar_w_rtn != NULL)
	    *tar_w_rtn = None;
	if(accepted_action_rtn != NULL)
	    *accepted_action_rtn = None;
	if(rect_rtn != NULL)
	    memset(rect_rtn, 0x00, sizeof(rectangle_t));

	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;

	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.dnd_status) &&
	   (client->format == 32)
	)
	{
	    const long flags = client->data.l[1];
	    bool_t drop_acceptable;

	    if(tar_w_rtn != NULL)
		*tar_w_rtn = (win_t)client->data.l[0];

	    /* If bit 0 is set then it means that a drop is acceptable */
	    drop_acceptable = (flags & (1 << 0)) ? True : False;

	    /* If bit 1 is NOT set then it means a rectangle has been
	     * specified that indicates to NOT send XdndPosition
	     * ClientMessage events while the pointer is this rectangle
	     */
	    if((rect_rtn != NULL) && !(flags & (1 << 1)))
	    {
		const long	coord = client->data.l[2],
				size = client->data.l[3];
		int16_t	x = (int16_t)((coord >> 16) & 0x0000ffff),
			y = (int16_t)((coord >> 0) & 0x0000ffff);
		u_int16_t	width = (u_int16_t)((size >> 16) & 0x0000ffff),
				height = (u_int16_t)((size >> 0) & 0x0000ffff);
		rect_rtn->x = (short)x;
		rect_rtn->y = (short)y;
		rect_rtn->width = (unsigned short)width;
		rect_rtn->height = (unsigned short)height;
	    }

	    if((accepted_action_rtn != NULL) && drop_acceptable)
		*accepted_action_rtn = (atom_t)client->data.l[4];

	    return(True);
	}
	else
	    return(False);
}

/*
 *	Checks if the specified event is a dnd leave event.
 *
 *	The specified event must be a ClientMessage event.
 */
bool_t OSWEventIsDNDLeave(
	win_t w, event_t *event,
	win_t *src_w_rtn
)
{
	event_client_message_t *client;

	if(src_w_rtn != NULL)
	    *src_w_rtn = None;
			   
	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;

	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.dnd_leave) &&
	   (client->format == 32)
	)
	{
	    if(src_w_rtn != NULL)
		*src_w_rtn = (win_t)client->data.l[0];
	    return(True);
	}
	else
	    return(False);
}

/*
 *	Checks if the specified event is a dnd drop event.
 *
 *	The specified event must be a ClientMessage event.
 */
bool_t OSWEventIsDNDDrop(
	win_t w, event_t *event,
	win_t *src_w_rtn, time_t *t
)
{
	event_client_message_t *client;

	if(src_w_rtn != NULL)
	    *src_w_rtn = None;
	if(t != NULL)
	    *t = 0l;

	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;
				 
	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.dnd_drop) &&
	   (client->format == 32)
	)
	{
	    if(src_w_rtn != NULL)
		*src_w_rtn = (win_t)client->data.l[0];
	    if(t != NULL)
		*t = (time_t)client->data.l[2];
	    return(True);
	}
	else
	    return(False);
}

/*
 *	Checks if the specified event is a dnd finished event.
 *
 *	The specified event must be a ClientMessage event.
 */
bool_t OSWEventIsDNDFinished(
	win_t w, event_t *event,
	win_t *tar_w_rtn,
	bool_t *drop_processed_and_successful,
	atom_t *action
)
{
	event_client_message_t *client;

	if(tar_w_rtn != NULL)
	    *tar_w_rtn = None;
	if(drop_processed_and_successful != NULL)
	    *drop_processed_and_successful = False;
	if(action != NULL)
	    *action = None;

	if((w == None) || (event == NULL))
	    return(False);

	if(event->type != ClientMessage)
	    return(False);

	client = &event->xclient;
				 
	if((event->xany.window == w) &&
	   (client->message_type == osw_atom.dnd_finished) &&
	   (client->format == 32)
	)
	{
	    const long flags = client->data.l[1];
	    if(tar_w_rtn != NULL)
		*tar_w_rtn = (win_t)client->data.l[0];
	    if(drop_processed_and_successful != NULL)
		*drop_processed_and_successful = (flags & (1 << 0)) ? True : False;
	    if(action != NULL)
		*action = (atom_t)client->data.l[2];
	    return(True);
	}
	else
	    return(False);
}



/*
 *	Creates a window.
 */
int OSWCreateWindow(
	win_t *w, win_t parent,  
	int x, int y,
	unsigned int width, unsigned int height
)
{
	osw_gui_struct *gui = &osw_gui[0];
	XSetWindowAttributes swattr;

	if((w == NULL) || (width == 0) || (height == 0))
	    return(-2);

	if(parent == None)
	    parent = gui->root_win;

	/* Set up the window attributes */
	memset(&swattr, 0x00, sizeof(XSetWindowAttributes));

	swattr.background_pixmap = ParentRelative;
	swattr.background_pixel = gui->black_pix;
	swattr.border_pixel = gui->white_pix;
	swattr.bit_gravity = NorthWestGravity;
	swattr.win_gravity = NorthWestGravity;
	swattr.backing_store = WhenMapped;	/* NotUseful, WhenMapped, Always */
	swattr.backing_planes = AllPlanes;
	swattr.backing_pixel = gui->black_pix;
	swattr.save_under = False;
	swattr.event_mask = 0l;
	swattr.do_not_propagate_mask = 0l;
	swattr.override_redirect = False;
	if(gui->we_created_colormap)
	    swattr.colormap = gui->colormap;
	else
	    swattr.colormap = CopyFromParent;
	swattr.cursor = CopyFromParent;
/*	swattr.cursor = gui->std_cursor; */


	/* Create the window */
	*w = XCreateWindow(
	    gui->display,
	    parent,
	    x, y,
	    width, height,
	    0,			/* Border */
	    gui->depth,		/* Operative bit depth */
	    InputOutput,	/* InputOutput or InputOnly? */
	    gui->visual,	/* Should we use CopyFromParent? */
#if 1
	    CWBitGravity | CWWinGravity | CWColormap |
	    CWOverrideRedirect | CWSaveUnder,
#else
	    CWBackPixel  | CWBorderPixel | CWBackPixmap | CWBorderPixmap|
	    CWBitGravity | CWWinGravity  | CWColormap   | CWBackingStore|
	    CWBackingPlanes | CWBackingPixel | CWOverrideRedirect |
	    CWSaveUnder | CWCursor,
#endif
	    &swattr		/* Set attributes */
	);
	if(*w == None)
	    return(-1);

	/* Is this a toplevel window? */
	if(parent == gui->root_win)
	{
	    int i;

	    /* Set WM protocols */
	    atom_t a[2];
	    a[0] = osw_atom.wm_delete_window;
	    a[1] = osw_atom.wm_take_focus;
	    XSetWMProtocols(
		gui->display,
		*w,
		a, sizeof(a) / sizeof(atom_t)
	    );

	    /* Record this toplevel window */
	    i = MAX(gui->ntoplevels, 0);
	    gui->ntoplevels = i + 1;
	    gui->toplevel = (win_t *)realloc(
		gui->toplevel, gui->ntoplevels * sizeof(win_t)
	    );
	    if(gui->toplevel != NULL)
	    {
		gui->toplevel[i] = *w;
	    }
	    else
	    {
		gui->ntoplevels = 0;
	    }

	}

	return(0);
}

/*
 *	Create an input only window.
 */
int OSWCreateInputWindow(
	win_t *w, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	osw_gui_struct *gui = &osw_gui[0];
	XSetWindowAttributes swattr;

	if((w == NULL) || (parent == None) || (width == 0) || (height == 0))
	    return(-1);

	/* Set up the window attributes */
	memset(&swattr, 0x00, sizeof(XSetWindowAttributes));

	swattr.background_pixmap = ParentRelative;
	swattr.background_pixel = gui->black_pix;
	swattr.border_pixel = gui->white_pix;
	swattr.bit_gravity = NorthWestGravity;
	swattr.win_gravity = NorthWestGravity;
	swattr.backing_store = WhenMapped;	/* NotUseful, WhenMapped, Always */
	swattr.backing_planes = AllPlanes;
	swattr.backing_pixel = gui->black_pix;
	swattr.save_under = False;
	swattr.event_mask = 0l;
	swattr.do_not_propagate_mask = 0l;
	swattr.override_redirect = False;
	if(gui->we_created_colormap)
	    swattr.colormap = gui->colormap;
	else
	    swattr.colormap = CopyFromParent;
	swattr.cursor = CopyFromParent;
/*      swattr.cursor = gui->std_cursor; */


	/* Create an input only window */
	*w = XCreateWindow(
	    gui->display,
	    parent,
	    x, y,
	    width, height,
	    0,			/* Border (internal, we don't want it) */
	    0,			/* Operative bit depth (must be 0) */
	    InputOnly,		/* InputOutput or InputOnly? */
	    gui->visual,	/* Should we use CopyFromParent? */
	    0,			/* Set attributes mask */
	    &swattr		/* Set attributes */
	);
	if(*w == None)
	    return(-1);

	return(0);
}

/*
 *	Destroy window.
 */
void OSWDestroyWindow(win_t *w)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(w == NULL)
	    return;

	if(*w != None)
	{
	    /* Destroy this window */
	    XDestroyWindow(gui->display, *w);

	    /* If this is a toplevel window then remove it from the
	     * toplevel windows list
	     */
	    if(gui->ntoplevels > 0)
	    {
		int i;
		for(i = 0; i < gui->ntoplevels; i++)
		{
		    if(gui->toplevel[i] == *w)
		    {
			int n;

			gui->ntoplevels--;
			for(n = i; n < gui->ntoplevels; n++)
			    gui->toplevel[n] = gui->toplevel[n + 1];

			if(gui->ntoplevels > 0)
			{
			    gui->toplevel = (win_t *)realloc(
				gui->toplevel,
				gui->ntoplevels * sizeof(win_t)
			    );
			    if(gui->toplevel == NULL)
				gui->ntoplevels = 0;
			}
			else
			{
			    free(gui->toplevel);
			    gui->toplevel = NULL;
			    gui->ntoplevels = 0;
			}

			i--;
		    }
		}
	    }

	    /* If this window was the modal window then unset the
	     * modal window
	     */
	    if(*w == gui->modal_w)
		gui->modal_w = None;

	    *w = None;
	}
}

/*
 *	Checks if the drawable is window.
 */
bool_t OSWDrawableIsWindow(drawable_t d)
{
	int status;
	osw_gui_struct *gui = &osw_gui[0];
	win_attr_t wattr;

	if(d == None)
	    return(False);

	/* Use simple error handler for this operation */
	gui->last_error = 0;
	XSetErrorHandler(OSWGUISimpleErrorHandler);

	/* Get window attributes of d, if this fails then
	 * d is a pixmap_t. If this is successful then d is
	 * a win_t.
	 */
	XGetWindowAttributes(
	    gui->display,
	    d,
	    &wattr
	);
	status = gui->last_error;

	/* Set back error handler */
	XSetErrorHandler(OSWGUIErrorHandler);

	return(status ? False : True);
}

/*
 *	Checks if the window is a toplevel window.
 */
bool_t OSWWindowIsToplevel(win_t w)
{
	int i;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return(False);

	for(i = 0; i < gui->ntoplevels; i++)
	{
	    if(gui->toplevel[i] == w)
		return(True);
	}

	return(False);
}

/*
 *	GUI wrapper's set window icon function.
 */
static void OSWSetWindowIcon(
	win_t w,
	const pixmap_t icon,
	const bool_t has_transparency
)
{
	Display *dpy;
	image_t *img;
	long supplied_return;

	XIconSize *icon_sizes;
	int count, i;
	unsigned int width = 64, height = 64;

	XWMHints wm_hints;
	XClassHint class_hints;
	XSizeHints sz_hints;
	osw_gui_struct *gui = &osw_gui[0];

	if((w == None) || (icon == None))
	    return;

	dpy = gui->display;

	/* Set new icon referance in wm hints structure */
	wm_hints.icon_pixmap = icon;

	/* Set class hint names */
	class_hints.res_name = "Eterm"; /* Must call it that */
	class_hints.res_class = "Eterm";


	/* Get original size hints */
	XGetWMNormalHints(
	    dpy, w,
	    &sz_hints,
	    &supplied_return
	);
	/* Update values */
	if(!(sz_hints.flags & PResizeInc))
	{
	    sz_hints.flags |= PResizeInc;
	    sz_hints.width_inc = 1;
	    sz_hints.height_inc = 1;
	}
	if(!(sz_hints.flags & PBaseSize))
	{
	    sz_hints.flags |= PBaseSize;
	    sz_hints.base_width = 0;
	    sz_hints.base_height = 0;
	}

	/* Get valid/suggested icon sizes */
	if(XGetIconSizes(
	    dpy, gui->root_win,
	    &icon_sizes, &count
	))
	{
	    for(i = 0; i < count; i++)
	    {
		width = MIN(icon_sizes[i].max_width, 64);
		height = MIN(icon_sizes[i].max_height, 64);
	    }
	    XFree(icon_sizes);
	}

	/* Create mask from image */
	img = OSWGetImage(icon, 0, 0, width, height);
	wm_hints.icon_mask = (Pixmap)OSWCreateMaskFromImage(img);
	OSWDestroyImage(&img);

	/* Create window to be used for icon */
	wm_hints.icon_window = XCreateSimpleWindow(
	    dpy, w,
	    0, 0,
	    width, height,
	    0,
	    0L, 0L
	);

	XShapeCombineMask(
	    dpy, wm_hints.icon_window,
	    ShapeBounding,
	    0, 0,
	    wm_hints.icon_mask,
	    ShapeSet
	);
	XSetWindowBackgroundPixmap(
	    dpy, wm_hints.icon_window,
	    wm_hints.icon_pixmap
	);

	/* Extra stuff, may be needed */
	wm_hints.window_group = w;
	wm_hints.input = True;
	wm_hints.initial_state = NormalState;

	wm_hints.flags = (
	    InputHint |
	    StateHint |
	    WindowGroupHint |
	    IconPixmapHint |
	    IconWindowHint |
	    IconMaskHint
	);


	/* Set new properties */
	XSetWMProperties(
	    dpy, w,
	    NULL, NULL,
	    (char **)NULL, 0,
	    &sz_hints,		/* Reset original size hints */
	    &wm_hints,		/* WM icon hints */
	    &class_hints	/* Nessasary gibberish */
	);
}

/*
 *	Sets standard window manager properties for window w.
 */
void OSWSetWindowWMProperties(
        win_t w,
        const char *title,
        const char *icon_title,
        pixmap_t icon,
        const bool_t wm_sets_coordinates,
        const int x, const int y,
        const unsigned int min_width, const unsigned int min_height,
        const unsigned int max_width, const unsigned int max_height,
        const int frame_style,		/* One of WindowFrameStyle* */
        const int argc, char **argv
)
{
	sizehints_t sizehints;
	PropMwmHints mwm_prop;		/* WM window decoration control */
	XSetWindowAttributes swattr;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	if(title == NULL)
	    title = "Untitled";
	if(icon_title == NULL) 
	    icon_title = "Untitled";

	memset(&sizehints, 0x00, sizeof(sizehints_t));
	if(wm_sets_coordinates)
	{
	    /* Let window manager set coordinates */
	    sizehints.flags = PSize | PMinSize | PMaxSize |
		PResizeInc | PBaseSize;
	    sizehints.x = 0;
	    sizehints.y = 0;
	}
	else
	{
	    /* Use given input coordinates */
	    sizehints.flags = USPosition | PSize | PMinSize |
		PMaxSize | PResizeInc | PBaseSize;
	    sizehints.x = x;
	    sizehints.y = y;
	}

	/* Set the size hints */
	sizehints.min_width = min_width;
	sizehints.min_height = min_height;
	sizehints.max_width = max_width;
	sizehints.max_height = max_height;
	sizehints.width_inc = 1;
	sizehints.height_inc = 1;
	sizehints.base_width = 0;
	sizehints.base_height = 0;

	XSetWMNormalHints(
	    gui->display,
	    w,
	    &sizehints
	);

	/* Set the standard properties */
	XSetStandardProperties(
	    gui->display,
	    w,
	    title,
	    icon_title,
	    icon,
	    argv, argc,
	    &sizehints
	);

	/* Is this a toplevel window? */
	if(OSWWindowIsToplevel(w))
	{
	    /* Set up the Motif WM hints */
	    mwm_prop.flags = MWM_HINTS_FUNCTIONS |
			     MWM_HINTS_DECORATIONS;
/*
			     MWM_HINTS_INPUT_MODE |
			     MWM_HINTS_STATUS;
 */

	    mwm_prop.functions = (MWM_FUNC_RESIZE   |
				  MWM_FUNC_MOVE     |
				  MWM_FUNC_MINIMIZE |
				  MWM_FUNC_MAXIMIZE |
				  MWM_FUNC_CLOSE
	    );

	    mwm_prop.decorations = (MWM_DECOR_BORDER   |
				    MWM_DECOR_RESIZEH  |
				    MWM_DECOR_TITLE    |
				    MWM_DECOR_MENU     |
				    MWM_DECOR_MINIMIZE |
				    MWM_DECOR_MAXIMIZE
	    );

	    mwm_prop.inputMode = 0;
	    mwm_prop.status = 0;

	    /* Set the frame style */
	    switch(frame_style)
	    {
	      case OSW_WINDOW_FRAME_NONE:
		/* Override redirect, WM neither decorates it nor
		 * manages it
		 */
		swattr.override_redirect = True;
		XChangeWindowAttributes(
		    gui->display,
		    w,
		    CWOverrideRedirect,
		    &swattr
		);
		break;

	      case OSW_WINDOW_FRAME_DIALOG:
		mwm_prop.functions = MWM_FUNC_MOVE |
				     MWM_FUNC_MINIMIZE |
				     MWM_FUNC_CLOSE;
		mwm_prop.decorations = MWM_DECOR_BORDER |
				       MWM_DECOR_TITLE |
				       MWM_DECOR_MENU |
				       MWM_DECOR_MINIMIZE;
		XChangeProperty(
		    gui->display,
		    w,
		    osw_atom._motif_wm_hints,
		    osw_atom._motif_wm_hints,
		    32,				/* Bits */
		    PropModeReplace,
		    (unsigned char *)&mwm_prop,
		    PROP_MWM_HINTS_ELEMENTS
		);
		break;

	      case OSW_WINDOW_FRAME_DIALOG_NOTITLE:
		mwm_prop.functions = MWM_FUNC_MOVE |
				     MWM_FUNC_MINIMIZE |
				     MWM_FUNC_CLOSE;
		mwm_prop.decorations = MWM_DECOR_BORDER;
		XChangeProperty(
		    gui->display,
		    w,
		    osw_atom._motif_wm_hints,
		    osw_atom._motif_wm_hints,
		    32,				/* Bits */
		    PropModeReplace,
		    (unsigned char *)&mwm_prop,
		    PROP_MWM_HINTS_ELEMENTS
		);
		break;

	      /* No close button or menu item */
	      case OSW_WINDOW_FRAME_NO_CLOSE:
		mwm_prop.functions &= ~(MWM_FUNC_CLOSE);
		XChangeProperty(
		    gui->display,
		    w,
		    osw_atom._motif_wm_hints,
		    osw_atom._motif_wm_hints,
		    32,				/* Bits */
		    PropModeReplace,
		    (unsigned char *)&mwm_prop,
		    PROP_MWM_HINTS_ELEMENTS
		);
		break;

	      default:				/* OSW_WINDOW_FRAME_STANDARD */
		break;
	    }
	}

	/* Set the WM icon */
	OSWSetWindowIcon(
	    w,
	    icon,
	    True				/* Has transparency */
	);
}

/*
 *	Select events for window w to recieve.
 */
void OSWSetWindowInput(
	win_t w,
	const event_mask_t eventmask
)
{
	if(w == None)
	    return;

	XSelectInput(
	    osw_gui[0].display,
	    w,
	    eventmask
	);
}

/*
 *	Sets the window specified by w to be a transient of the
 *	window master.
 */
void OSWSetTransientFor(
	win_t w,
	win_t main_w
)
{
	if(w == None)
	    return;

	XSetTransientForHint(
	    osw_gui[0].display,
	    w,
	    main_w
	);
}

/*
 *	Sets the window's title.
 */
void OSWSetWindowTitle(
	win_t w,
	const char *title
)
{
	if(w == None)
	    return;

	XStoreName(
	    osw_gui[0].display,
	    w,
	    (title != NULL) ? title : ""
	);
}

/*
 *	Clears the window.
 */
void OSWClearWindow(win_t w)
{
	if(w == None)
	    return;

	XClearWindow(osw_gui[0].display, w);
}

/*
 *	Sets window's background color or tile.
 */
void OSWSetWindowBkg(win_t w, pixel_t pix, pixmap_t pixmap)
{
	if(w == None)
	    return;

	if(pixmap == None)
	    XSetWindowBackground(osw_gui[0].display, w, pix);
	else
	    XSetWindowBackgroundPixmap(osw_gui[0].display, w, pixmap);
}

/*
 *	Gets the window's geometry.
 */
bool_t OSWGetWindowGeometry(
	win_t w,
	int *x, int *y, unsigned int *width, unsigned int *height,
	unsigned int *depth
)
{
	int cx = 0, cy = 0;
	unsigned int	cwidth = 0, cheight = 0,
			cborder = 0, cdepth = 0;
	Status status;
	win_t root = None;

	if(x != NULL)
	    *x = cx;
	if(y != NULL)
	    *y = cy;
	if(width != NULL)
	    *width = cwidth;
	if(height != NULL)
	    *height = cheight;
	if(depth != NULL)
	    *depth = cdepth;

	if(w == None) 
	    return(False);

	status = XGetGeometry(
	    osw_gui[0].display, (Drawable)w,
	    &root,
	    &cx, &cy, &cwidth, &cheight, &cborder, &cdepth
	);
	if(!status)
	    return(False);

	if(x != NULL)
	    *x = cx;
	if(y != NULL)
	    *y = cy;
	if(width != NULL)
	    *width = cwidth;
	if(height != NULL) 
	    *height = cheight;
	if(depth != NULL)    
	    *depth = cdepth;

	return(True);
}

/*
 *	Gets window attributes, returns True if successful.
 */
bool_t OSWGetWindowAttributes(win_t w, win_attr_t *wattr)
{
	int status;
	osw_gui_struct *gui = &osw_gui[0];

	if(wattr == NULL)
	    return(False);

	if(w == None)
	{
	    memset(wattr, 0x00, sizeof(win_attr_t));
	    return(False);
	}

	/* Use simple error handler for this operation */
	gui->last_error = 0;
	XSetErrorHandler(OSWGUISimpleErrorHandler);

	/* Get window attributes */
	XGetWindowAttributes(gui->display, w, wattr);
	status = gui->last_error;

	/* Set back to use standard error handler */
	XSetErrorHandler(OSWGUIErrorHandler);

	return(status ? False : True);
}

/*
 *	Calculates the position of the specified coordinates on the
 *	parent window to the coordinates on the child window.
 *
 *	The x and y specify the coordinates on the parent window.
 */
bool_t OSWTranslateCoordinatesToChild(       
	win_t parent,
	win_t child,
	const int x, const int y,
	int *x_rtn, int *y_rtn
)
{
	bool_t status;
	int x2, y2;
	win_t youngest;

	if(x_rtn != NULL)
	    *x_rtn = 0;
	if(y_rtn != NULL)
	    *y_rtn = 0;

	if((child == None) || (parent == None))
	    return(False);

	status = XTranslateCoordinates(
	    osw_gui[0].display,
	    parent, child,
	    x, y,
	    &x2, &y2,
	    &youngest
	);
	if(status)
	{
	    if(x_rtn != NULL)
		*x_rtn = x2;
	    if(y_rtn != NULL)
		*y_rtn = y2;
	}

	return(status);
}

/*
 *	Calculates the position of the specified coordinates on the
 *	child window to the coordinates on the parent window.
 *
 *      The x and y specify the coordinates on the child window.
 */
bool_t OSWTranslateCoordinatesToParent(
	win_t child,
	win_t parent,
	const int x, const int y,
	int *x_rtn, int *y_rtn
)
{
	bool_t status;
	int x2, y2;
	win_t youngest;

	if(x_rtn != NULL)
	    *x_rtn = 0;
	if(y_rtn != NULL)
	    *y_rtn = 0;

	if((child == None) || (parent == None))
	    return(False);

	status = XTranslateCoordinates(
	    osw_gui[0].display,
	    child,
	    parent,
	    x, y,
	    &x2, &y2,
	    &youngest
	);
	if(status)
	{
	    if(x_rtn != NULL)
		*x_rtn = x2;
	    if(y_rtn != NULL)
		*y_rtn = y2;
	}

	return(status);
}

/*
 *	Gets the toplevel window's position relative to the root window.
 *
 *	Returns True if the position was obtained or False on error.
 */
bool_t OSWGetWindowRootOrigin(
	win_t w,
	int *x_rtn, int *y_rtn
)
{
	bool_t status;
	int cx, cy, rx = 0, ry = 0;
	win_t parent, root = osw_gui[0].root_win;
/*	win_t child; */

	if(x_rtn != NULL)
	    *x_rtn = rx;
	if(y_rtn != NULL)
	    *y_rtn = ry;

	if(w == None)
	    return(False);

#if 0
/* Does not work, WM problem */
	status = XTranslateCoordinates(
	    osw_gui[0].display, w, root,
	    0, 0,
	    &rx, &ry,
	    &child
	);
#else
/* Workaround, assume w is a toplevel and only get the
 * position of the root's child, this is needed because some
 * WM position things oddly
 */
	status = True;
	do
	{
	    parent = OSWGetWindowParent(w);
	    if(parent == root)
	    {
		if(!OSWGetWindowGeometry(
		    w,
		    &cx, &cy,
		    NULL, NULL,
		    NULL
		))
		{ 
		    status = False;
		    break;
		}

		rx += cx;
		ry += cy;
	    }

	    w = parent;

	} while((w != None) && (w != root));
#endif
#if 0
/* Does not work, WM problem */
	do
	{
	    if(!OSWGetWindowGeometry(
		w,
		&cx, &cy,
		NULL, NULL,
		NULL
	    ))
	    {
		status = False;
		break;
	    }

	    rx += cx;
	    ry += cy;

	    parent = OSWGetWindowParent(w);
	    w = parent;

	} while((w != None) && (w != root));
#endif

	if(x_rtn != NULL)
	    *x_rtn = rx;
	if(y_rtn != NULL)
	    *y_rtn = ry;

	return(status);
}

/*
 *	Gets the parent window of the specified window.
 *
 *	Can return 0 on error.
 */
win_t OSWGetWindowParent(win_t w)
{
	Status status;
	unsigned int nchildren;
	win_t	root = osw_gui[0].root_win,
		parent,
		root_rtn,
		*children;

	if((w == None) || (root == None))
	    return(0);

	/* Check if w is already the root window? */
	if(w == root)
	    return(root);

	/* Get the window tree */
	children = NULL;
	status = XQueryTree(
	    osw_gui[0].display,
	    w,
	    &root_rtn,
	    &parent,
	    &children,
	    &nchildren
	);

	/* Failed to get the tree? */
	if(status == 0)
	    return(None);

	/* Need to free children_return */
	if(children != NULL)
	    XFree(children);

	return(parent);
}

/*
 *	Checks if the child window is a grandchild of the parent
 *	window.
 *
 *	The grand_parent specifies the parent window. The grand_parent
 *	may not be the root window or else False will be returned.
 *
 *	The grand_child specifies the child window.
 *
 *	Returns True if grand_child is a child of the grand_parent.
 *	Will also return True if grand_child == grand_parent.
 */
bool_t OSWCheckWindowAncestory(
	win_t grand_parent,
	win_t grand_child
)
{
	int count;
	win_t		w,
			root_w = osw_gui[0].root_win;

	if(root_w == None)
	    return(False);

	else if((grand_child == root_w) || (grand_child == None))
	    return(False);

	else if((grand_parent == root_w) || (grand_parent == None))
	    return(False);

	/* Begin checking for ancestry */
	w = grand_child;
	for(count = 0; count < 256; count++)
	{
	    /* Is this window the parent? */
	    if(w == grand_parent)
		return(True);

	    /* Get the next parent */
	    w = OSWGetWindowParent(w);
	    if(w == None)   
		return(False);
	    else if(w == root_w)
		return(False);
	}

	/* Too many recursions, return False */
	return(False);
}

/*
 *	If a window is set modal and if this event is not for that
 *	window or one of its children and if this event is a;
 *	KeyPress, KeyRelease, ButtonPress, ButtonRelease, MotionNotify,
 *	FocusIn, or EnterNotify event then this event will be translated
 *	to be relative to the modal window.
 */
static void OSWTranslateModalEvent(
	osw_gui_struct *gui,
	event_t *event
)
{
	int x, y;
	win_t   w,
		modal_w = gui->modal_w,
		root = gui->root_win;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	event_focus_t *focus;
	event_crossing_t *crossing;

	/* No window set modal? */
	if(modal_w == None)
	    return;

	/* If the pointer is grabbed then we need to report events to
	 * that window even though a modal window has been set
	 */
	if(gui->grab_pointer_w != None)
	    return;

	/* Check if the event is one that should be translated and if
	 * so then check if the window that the event is for is not
	 * the modal window and not a child of the modal window
	 */
	switch(event->type)
	{
	  case KeyPress:
	  case KeyRelease:
	  case ButtonPress:
	  case ButtonRelease:
	  case MotionNotify:
	  case FocusIn:
	  case EnterNotify:
	    w = event->xany.window;
	    while((w != None) && (w != root))
	    {
		/* This event is for the modal window */
		if(w == modal_w)
		    return;

		w = OSWGetWindowParent(w);
	    }
	    break;
	}

	/* The event is for a window that is not the modal window and
	 * not a child of the modal window
	 *
	 * Translate the event, based on its type, to be relative to
	 * the modal window
	 */
	switch(event->type)
	{
	  case KeyPress:
	  case KeyRelease:
	    key = &event->xkey;
	    XTranslateCoordinates(
		key->display,
		key->window,
		modal_w,
		key->x, key->y,
		&x, &y,
		&key->subwindow
	    );
	    key->x = x;
	    key->y = y;
	    key->window = modal_w;
	    break;

	  case ButtonPress:
	  case ButtonRelease:
	    button = &event->xbutton;
	    XTranslateCoordinates(
		button->display,
		button->window,
		modal_w,
		button->x, button->y,
		&x, &y,
		&button->subwindow
	    );
	    button->x = x; 
	    button->y = y;
	    button->window = modal_w;
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    XTranslateCoordinates(
		motion->display,
		motion->window,
		modal_w,
		motion->x, motion->y,
		&x, &y,
		&motion->subwindow
	    );
	    motion->x = x;
	    motion->y = y;
	    motion->window = modal_w;
	    break;

	  case FocusIn:
	    focus = &event->xfocus;
	    focus->window = modal_w;
	    break;

	  case EnterNotify:
	    crossing = &event->xcrossing;
	    XTranslateCoordinates(
		crossing->display,
		crossing->window,
		modal_w,
		crossing->x, crossing->y,
		&x, &y,
		&crossing->subwindow
	    );
	    crossing->x = x;
	    crossing->y = y;
	    crossing->window = modal_w;
	    break;
	}
}

/*
 *	Sets the window as modal.
 *
 *	When this is set, key and button events for windows other than
 *	the modal window will be discarded unless the pointer is
 *	grabbed by the modal window, in which case all events will be
 *	translated to be relative to the modal window.
 */
void OSWSetModal(win_t w)
{
	osw_gui[0].modal_w = w;
}

/*
 *	Maps the window.
 */
void OSWMapWindow(win_t w)
{
	if(w == None)
	    return;

	XMapWindow(
	    osw_gui[0].display,
	    w
	);
}

/*
 *	Map and raise the window.
 */
void OSWMapRaised(win_t w)
{
	if(w == None)
	    return;

	XMapRaised(
	    osw_gui[0].display,
	    w
	);
}

/*
 *	Maps all the child windows.
 */
void OSWMapSubwindows(win_t w)
{
	if(w == None)
	    return; 

	XMapSubwindows(osw_gui[0].display, w);
}

/*
 *	Unmap window.
 */
void OSWUnmapWindow(win_t w)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	/* If this window is the modal window then clear it */
	if(w == gui->modal_w)
	    gui->modal_w = None;

	XUnmapWindow(
	    gui->display,
	    w
	);
}

/*
 *	Iconifies the toplevel window.
 */
void OSWIconifyWindow(win_t w)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	/* Unset modal window as needed */
	if(w == gui->modal_w)
	    gui->modal_w = None;

	XIconifyWindow(
	    gui->display,
	    w,
	    gui->scr_num
	);
}

/*
 *	Reparent window.
 */
void OSWReparentWindow(
	win_t w,
	win_t parent
)
{
	if((w == None) || (parent == None))
	    return;

	XReparentWindow(
	    osw_gui[0].display,
	    w,
	    parent,
	    0, 0
	);
}

/*
 *	Move window.
 */
void OSWMoveWindow(
	win_t w,
	const int x, const int y
)
{
	if(w == None)
	    return;

	XMoveWindow(
	    osw_gui[0].display,
	    w,
	    x, y
	);
}

/*
 *	Resize window.
 */
void OSWResizeWindow(
	win_t w,
	const unsigned int width, const unsigned int height
)
{
	if((w == None) || (width == 0) || (height == 0))
	    return;

	XResizeWindow(
	    osw_gui[0].display,
	    w,
	    width, height
	);
}

/*
 *	Move and resize window in one step.
 */
void OSWMoveResizeWindow(
	win_t w,
	const int x, const int y,
	const unsigned int width, const unsigned int height)
{
	if(w == None)
	    return;

	if((width == 0) || (height == 0))
	    XMoveWindow(
		osw_gui[0].display,
		w,
		x, y
	    );
	else
	    XMoveResizeWindow(
		osw_gui[0].display,
		w,
		x, y,
		width, height
	    );
}


/*
 *	Creates a bitmap
 */
int OSWCreateBitmap(
	bitmap_t *bitmap,
	const unsigned int width, const unsigned int height
)
{
	osw_gui_struct *gui = &osw_gui[0];

	if((gui->root_win == None) || (bitmap == NULL) ||
	   (width == 0) || (height == 0)
	)
	    return(-1);

	/* Create a 1 bit depth Pixmap (which is a bitmap) */
	*bitmap = XCreatePixmap(
	    gui->display, gui->root_win,
	    width,
	    height,
	    1			/* 1 bit depth */
	);
	return((*bitmap != None) ? 0 : -1);
}

/*
 *	Destroys the bitmap.
 */
void OSWDestroyBitmap(bitmap_t *bitmap)
{
	if(bitmap == NULL)
	    return;

	if(*bitmap != None)
	{
	    XFreePixmap(osw_gui[0].display, *bitmap);
	    *bitmap = None;
	}
}


/*
 *	Creates a pixmap.
 */
int OSWCreatePixmap(pixmap_t *pixmap, unsigned int width, unsigned int height)
{
	osw_gui_struct *gui = &osw_gui[0];

	if((gui->root_win == None) || (pixmap == NULL) ||
	   (width == 0) || (height == 0)
	)
	    return(-1);

	*pixmap = XCreatePixmap(
	    gui->display, gui->root_win,
	    width,
	    height,
	    gui->depth
	);
	return((*pixmap != None) ? 0 : -1);
}

/*
 *	Destroy the pixmap.
 */
void OSWDestroyPixmap(pixmap_t *pixmap)
{
	if(pixmap == NULL)
	    return;

	if(*pixmap != None)
	{
	    XFreePixmap(osw_gui[0].display, *pixmap);
	    *pixmap = None;
	}
}

/*
 *	Checks if the drawable is a pixmap.
 */
bool_t OSWDrawableIsPixmap(drawable_t d)
{
	int status;
	osw_gui_struct *gui = &osw_gui[0];
	win_attr_t wattr;

	if(d == None)
	    return(False);

	/* Use simple error handler for this operation */
	gui->last_error = 0;
	XSetErrorHandler(OSWGUISimpleErrorHandler);

	/* Get window attributes of d, if this fails then d is a
	 * pixmap_t
	 *
	 * If this is successful then d is a win_t
	 */
	XGetWindowAttributes(
	    gui->display,
	    d,
	    &wattr
	);
	status = gui->last_error;

	/* Set back error handler */
	XSetErrorHandler(OSWGUIErrorHandler);

	return(status ? True : False);
}

/*
 *	Gets the pixmap's geometry.
 */
bool_t OSWGetPixmapGeometry(
	pixmap_t pixmap,
	unsigned int *width, unsigned int *height,
	unsigned int *depth
)
{
	int cx = 0, cy = 0;
	unsigned int	cwidth = 0, cheight = 0,
			cborder = 0, cdepth = 0;
	Status status;
	win_t root = None;

	if(width != NULL)
	    *width = cwidth;
	if(height != NULL)
	    *height = cheight;
	if(depth != NULL)
	    *depth = cdepth;

	if(pixmap == None)
	    return(False);

	status = XGetGeometry(
	    osw_gui[0].display, (Drawable)pixmap,
	    &root,
	    &cx, &cy, &cwidth, &cheight, &cborder, &cdepth 
	);
	if(!status)
	    return(False);

	if(width != NULL)
	    *width = cwidth;
	if(height != NULL)
	    *height = cheight;
	if(depth != NULL)
	    *depth = cdepth;

	return(True);
}

/*
 *	Gets attributes of pixmap.
 */
bool_t OSWGetPixmapAttributes(pixmap_t pixmap, pixmap_attr_t *pattr)
{
	int x, y, status;
	osw_gui_struct *gui = &osw_gui[0];

	if(pattr == NULL)
	    return(False);

	if(pixmap == None)
	{
	    memset(pattr, 0x00, sizeof(pixmap_attr_t));
	    return(False);
	}

	/* Use simple error handler for this operation */
	gui->last_error = 0;
	XSetErrorHandler(OSWGUISimpleErrorHandler);

	/* Get pixmap attributes */
	XGetGeometry(
	    gui->display,
	    pixmap,
	    &(pattr->root_win),		/* Root window */
	    &x, &y,			/* Coordinates? */
	    &(pattr->width), &(pattr->height),	/* Size */
	    &(pattr->border_width),	/* Border width */
	    reinterpret_cast<unsigned int *>(&(pattr->depth))		/* Operative depth */
	);
	status = gui->last_error;

	/* Set back to use standard error handler */
	XSetErrorHandler(OSWGUIErrorHandler);

	return(status ? False : True);
}

/*
 *	Clear a graphics buffer / Pixmap.
 */
void OSWClearPixmap(
	pixmap_t pixmap,
	unsigned int width, unsigned int height,
	pixel_t pix
)
{
	if((pixmap == None) || (width == 0) || (height == 0))
	    return;

	OSWSetForeground(pix);
	OSWDrawSolidRectangle(
	    (drawable_t)pixmap,
	    0, 0,
	    width, height
	);
}


/*
 *	Creates an image from the specified XImage.
 */
image_t *OSWCreateImageFromXImage(XImage *ximg)
{
	image_t *img;

	if(ximg == NULL)
	    return(NULL);

	img = (image_t *)calloc(1, sizeof(image_t));
	img->ref_count = 1;
	img->byte_order = ximg->byte_order;
	img->xoffset = ximg->xoffset;
	img->format = ximg->format;
	img->data = (u_int8_t *)ximg->data;
	img->bits_per_pixel = ximg->bits_per_pixel;
	img->bitmap_pad = ximg->bitmap_pad;
	img->bytes_per_pixel = ximg->bits_per_pixel / 8;
	img->bytes_per_line = ximg->bytes_per_line;
	img->width = ximg->width;
	img->height = ximg->height;
	img->depth = ximg->depth;
	img->red_mask = ximg->red_mask;
	img->green_mask = ximg->green_mask;
	img->blue_mask = ximg->blue_mask;
	img->ximage = ximg;

	return(img);
}

/*
 *	Create an image.
 */
int OSWCreateImage(
	image_t **img_rtn,
	unsigned int width, unsigned int height
)
{
	XImage *ximg;
	visual_t *vis;
	osw_gui_struct *gui = &osw_gui[0];

	if((img_rtn == NULL) || (width == 0) || (height == 0))
	    return(-2);

	*img_rtn = NULL;

	vis = gui->visual;
	if(vis == NULL)
	    return(-2);

	/* Width and height must be even number */
	if(IS_NUM_ODD(width))
	    width += 1;
	if(IS_NUM_ODD(height))
	    height += 1;

	/* Create the X image */
	ximg = XCreateImage(
	    gui->display,
	    vis,		/* Visual */
	    gui->depth,		/* Operative bit depth */
	    ZPixmap,		/* Format */
	    0,			/* X offset */
	    NULL,		/* Image data */
	    width, height,	/* Size (in pixels) */
	    BitmapPad(gui->display),	/* Bitmap pad; 8, 16, or 32 */
	    0			/* Bytes per line (0 = auto calculated) */
	);
	if(ximg == NULL)
	{
	    fprintf(
		stderr,
"OSWCreateImage(): Warning: Unable to create an image:\
 size=%ix%i depth=%i\n",
		width, height, gui->depth
	    );
	    return(-3);
	}

	/* Allocate the image data as needed */
	if(ximg->data == NULL)
	{
	    const size_t data_len = (size_t)(
		ximg->height * ximg->bytes_per_line
	    );
	    if(data_len > 0)
		ximg->data = (char *)malloc(data_len);
	}

	if(ximg->xoffset != 0)
	{
	    fprintf(
		stderr,
"OSWCreateImage(): Warning: ximg->xoffset %i is not 0.\n",
		ximg->xoffset
	    );
	}

	*img_rtn = OSWCreateImageFromXImage(ximg);

	return(0);
}

/*
 *	Destroys an image.
 */
void OSWDestroyImage(image_t **img_rtn)
{
	image_t *img;

	if(img_rtn == NULL)
	    return;

	img = *img_rtn;
	if(img == NULL)
	    return;

	if(img->ximage != NULL)
	    XDestroyImage(img->ximage);

	free(img);
	*img_rtn = NULL;
}


/*
 *	Create shared memory image.
 */
int OSWCreateSharedImage(
	shared_image_t **img_rtn,
	unsigned int width, unsigned int height
)
{
#ifdef USE_XSHM
	osw_gui_struct *gui = &osw_gui[0];

	if(gui->def_use_shm)
	{
	    size_t data_len;
	    XImage *ximg;
	    XShmSegmentInfo *shm_seg_ptr;
	    Visual *vis;
	    shared_image_t *simg;

	    if((img_rtn == NULL) || (width == 0) || (height == 0))
		return(-1);

	    *img_rtn = NULL;

	    /* Width and height must be even number */
	    if(IS_NUM_ODD(width))
	        width += 1;
	    if(IS_NUM_ODD(height))
	        height += 1;

	    /* Get pointer to visual */
	    vis = gui->visual;
	    if(vis == NULL)
		return(-1);

	    /* Create the shared image */
	    *img_rtn = simg = (shared_image_t *)malloc(
	        sizeof(shared_image_t)
	    );
	    if(simg == NULL)
	        return(-3);

	    /* Get pointer to shm segment info */
	    shm_seg_ptr = &simg->shminfo;

	    /* Clear the shared memory info */
	    memset(
	        shm_seg_ptr,
		0x00,
	        sizeof(XShmSegmentInfo)
	    );

	    /* Create the X shared image */
	    simg->ximage = ximg = XShmCreateImage(
		gui->display,
		vis,			/* Visual */
		gui->depth,		/* Operative bit depth */
		XShmPixmapFormat(gui->display),	/* Format */
		NULL,			/* Actual data */
		shm_seg_ptr,
		width, height		/* Size (in pixels) */
	    );
	    if(ximg == NULL)
	    {
		fprintf(
		    stderr,
"OSWCreateSharedImage(): Warning: Unable to create a shared image:\
 size=%ix%i depth=%i\n",
		    width, height, gui->depth
		);
	        free(simg);
	        *img_rtn = NULL;
		return(-3);
	    }

	    /* Copy values from ximg to shared image base struct */
	    simg->ref_count = 1;
	    simg->in_progress = False;

	    simg->byte_order = ximg->byte_order;
	    simg->xoffset = ximg->xoffset;

	    if(simg->xoffset != 0)
	    {
		fprintf(
		    stderr,
"OSWCreateSharedImage(): Warning: simg->xoffset %i is not 0.\n",
		    simg->xoffset
		);
	    }

	    simg->format = ximg->format;
	    simg->data = (u_int8_t *)ximg->data;

	    simg->bits_per_pixel = ximg->bits_per_pixel;
	    simg->bitmap_pad = ximg->bitmap_pad;
	    simg->bytes_per_pixel = ximg->bits_per_pixel / 8;
	    simg->bytes_per_line = ximg->bytes_per_line;

	    simg->width = ximg->width;
	    simg->height = ximg->height;
	    simg->depth = ximg->depth;

	    simg->red_mask = ximg->red_mask;
	    simg->green_mask = ximg->green_mask;
	    simg->blue_mask = ximg->blue_mask;

	    /* Check for potential incompatable values */
	    if((simg->red_mask != vis->red_mask) ||
	       (simg->green_mask != vis->green_mask) ||
	       (simg->blue_mask != vis->blue_mask)
	    )
		fprintf(
		    stderr,
"OSWCreateSharedImage(): Warning: Incompatable plane masks between shared image and Visual.\n"
		);


	    /* Calculate total size of the image's data in bytes */
	    data_len = (size_t)(
		simg->height * simg->bytes_per_line
	    );

	    /* Get shared memory ID */
	    shm_seg_ptr->shmid = shmget(
		IPC_PRIVATE,
		data_len,		/* Size in bytes */
		IPC_CREAT | 0777
	    );
	    if(shm_seg_ptr->shmid < 0)
	    {
		const int error_code = errno;
		char *s = STRDUP(strerror(error_code));
		if(s != NULL)
		{
		    *s = toupper(*s);
		    fprintf(
			stderr,
"OSWCreateSharedImage(): shmget(): %s\n",
			s
		    );
		}
		free(s);
		free(simg);
		*img_rtn = NULL;
		return(-1);
	    }

	    /* Attach shared memory and check for errors */
	    shm_seg_ptr->shmaddr = ximg->data = (char *)shmat(
		shm_seg_ptr->shmid, 0, 0
	    );
	    if(shm_seg_ptr->shmaddr == (char *)-1)
	    {
		const int error_code = errno;
		char *s = STRDUP(strerror(error_code));
		if(s != NULL)
		{
		    *s = toupper(*s);
		    fprintf(
			stderr,
"OSWCreateSharedImage(): shmat(): %s\n",
			s
		    );
		}
		free(s);
		free(simg);
		*img_rtn = NULL;
		return(-1);
	    }

	    /* Set shared memory as read/write */
	    shm_seg_ptr->readOnly = False;

	    /* Attached the shared memory to the shared image */
	    XShmAttach(gui->display, shm_seg_ptr);
/*	    XSync(gui->display, False); */

	    /* Need to update pointer to shared image data */
	    simg->data = (u_int8_t *)ximg->data;


	    /* Mark this shm segment for deletion at once. The segment will
	     * automatically become released when both the server and our
	     * program deatches from it.
	     */
	    shmctl(shm_seg_ptr->shmid, IPC_RMID, NULL);
	}
	else
	{
#endif  /* USE_XSHM */
	    XImage *ximg;
	    visual_t *vis;
	    shared_image_t *simg;

	    if((img_rtn == NULL) || (width == 0) || (height == 0))
		return(-2);

	    *img_rtn = NULL;

	    vis = gui->visual;
	    if(vis == NULL)
		return(-2);

	    /* Width and height must be even number */
	    if(IS_NUM_ODD(width))
		width += 1;
	    if(IS_NUM_ODD(height))
		height += 1;

	    /* Create a regular X image */
	    ximg = XCreateImage(
		gui->display,
		vis,			/* Vixual */
		gui->depth,		/* Operative bit depth */
		ZPixmap,		/* Format */
		0,			/* X offset */
		NULL,			/* Image data */
		width, height,		/* Size */
		BitmapPad(gui->display),	/* Bitmap pad; 8, 16, or 32 */
		0			/* Bytes per line (0 = auto calculated) */
	    );
	    if(ximg == NULL)
	    {
		fprintf(
		    stderr,
"OSWCreateSharedImage(): Warning: Unable to create an image:\
 size=%ix%i depth=%i\n",
		    width, height, gui->depth
		);
		return(-3);
	    }

	    /* Allocate image data as needed */
	    if(ximg->data == NULL)
	    {
		const size_t data_len = (size_t)(
		    ximg->height * ximg->bytes_per_line
		);
		if(data_len > 0)
		    ximg->data = (char *)malloc(data_len);
	    }

	    if(ximg->xoffset != 0)
	    {
		fprintf(
		    stderr,
"OSWCreateSharedImage(): Warning: ximg->xoffset %i is not 0.\n",
		    ximg->xoffset
		);
	    }

	    *img_rtn = simg = (shared_image_t *)calloc(
		1, sizeof(shared_image_t)
	    );
	    simg->ref_count = 1;
	    simg->in_progress = False;
	    simg->byte_order = ximg->byte_order;
	    simg->xoffset = ximg->xoffset;
	    simg->format = ximg->format;
	    simg->data = (u_int8_t *)ximg->data;
	    simg->bytes_per_line = ximg->bytes_per_line;
	    simg->bits_per_pixel = ximg->bits_per_pixel;
	    simg->bitmap_pad = ximg->bitmap_pad; 
	    simg->width = ximg->width;
	    simg->height = ximg->height;
	    simg->depth = ximg->depth;
	    simg->red_mask = ximg->red_mask;
	    simg->green_mask = ximg->green_mask;
	    simg->blue_mask = ximg->blue_mask;
	    simg->ximage = ximg;

#ifdef USE_XSHM
	}
#endif  /* USE_XSHM */

	return(0);
}

/*
 *	Destroys a shared image.
 */
void OSWDestroySharedImage(shared_image_t **img_rtn)
{
#ifdef USE_XSHM
	XShmSegmentInfo *shm_seg_ptr;
#endif	/* USE_XSHM */
	shared_image_t *simg;
	XImage *ximg;
	osw_gui_struct *gui = &osw_gui[0];

#ifdef USE_XSHM
	if(osw_gui[0].def_use_shm)
	{
	    if(img_rtn == NULL)
		return;

	    simg = *img_rtn;
	    if(simg == NULL)
	        return;

	    ximg = simg->ximage;
	    shm_seg_ptr = &simg->shminfo;

#if 0
	    /* Get statistics of shared memory segment */
	    memset(&shm_stats_buf, 0x00, sizeof(struct shmid_ds));
	    shmctl(shm_seg_ptr->shmid, IPC_STAT, &shm_stats_buf);
#endif

	    /* Detach shared memory from GUI */
	    XShmDetach(gui->display, shm_seg_ptr);

	    /* Destroy the image */
	    if(ximg != NULL)
		XDestroyImage(ximg);

	    /* Detach shared memory from our side */
	    shmdt(shm_seg_ptr->shmaddr);

/*	    shm_stats_buf.shm_nattch = 0; */	/* Explicit set # attachs to 0 */

#if 0
	    /* Mark shm segment as destroyed */
	    shmctl(shm_seg_ptr->shmid, IPC_RMID, &shm_stats_buf);
#endif

	    free(simg);
	    *img_rtn = NULL;
	}
	else
	{
#endif	/* USE_XSHM */
	    if(img_rtn == NULL)
		return;

	    simg = *img_rtn;
	    if(simg == NULL)
		return;
  
	    ximg = simg->ximage;
	    if(ximg != NULL)
		XDestroyImage(ximg);

	    free(simg);
	    *img_rtn = NULL;
#ifdef USE_XSHM
	}
#endif  /* USE_XSHM */
}


/*
 *	Creates a pixmap from the specified image.
 */
pixmap_t OSWCreatePixmapFromImage(image_t *img)
{
	pixmap_t pixmap;

	if(img == NULL)
	    return(0);

	/* Create the pixmap */
	if(!OSWCreatePixmap(
	    &pixmap, img->width, img->height
	))
	{
	    /* Copy the image to the pixmap */
	    OSWPutImageToDrawable(img, pixmap);
	}

	return(pixmap);
}


/*
 *	Creates a mask from the specified image.
 */
bitmap_t OSWCreateMaskFromImage(image_t *img)
{
	int x, y, bpp, bpl;
	unsigned int width, height;
	unsigned long pixel_mask;
	u_int8_t *data;
	u_int8_t *ptr8;
	u_int16_t *ptr16;
	u_int32_t *ptr32;
	pixel_t white_pix, black_pix;
	bitmap_t mask;
	Display *dpy;
	gc_t gc;
	gc_val_t gcv;
	osw_gui_struct *gui = &osw_gui[0];

	if(img == NULL)
	    return(0);

	dpy = gui->display;
	width = img->width;
	height = img->height;
	pixel_mask = img->red_mask | img->green_mask | img->blue_mask;
	data = (u_int8_t *)img->data;
	if(data == NULL)
	    return(0);

	/* Create the mask */
	if(OSWCreateBitmap(&mask, width, height))
	    return(mask);

	/* Create a graphics context for drawing the mask */
	gcv.function = GXclear;
	gcv.plane_mask = 1;		/* Monochrome */
	gcv.foreground = white_pix = gui->white_pix;
	gcv.background = black_pix = gui->black_pix;
	gc = XCreateGC(
	    dpy,
	    (Drawable)mask,
	    GCFunction | GCPlaneMask |
	    GCForeground | GCBackground,
	    &gcv
	);
	if(gc == None)
	    return(mask);

	switch(img->depth)
	{
	  case 8:
	    bpp = 1;
	    bpl = img->bytes_per_line;
	    for(y = 0; y < (int)height; y++)
	    {
		for(x = 0; x < (int)width; x++)
		{
		    ptr8 = (u_int8_t *)(data +
			(y * bpl) + (x * bpp)
		    );
		    if(*ptr8 & (u_int8_t)pixel_mask)
			XSetFunction(dpy, gc, GXset);
		    else  
			XSetFunction(dpy, gc, GXclear);
		    XDrawPoint(dpy, mask, gc, x, y);
		}
	    }
	    break; 

	  case 15:
	  case 16:
	    bpp = 2;
	    bpl = img->bytes_per_line;
	    for(y = 0; y < (int)height; y++)
	    {
		for(x = 0; x < (int)width; x++)  
		{
		    ptr16 = (u_int16_t *)(data +
			(y * bpl) + (x * bpp)
		    );
		    if(*ptr16 & (u_int16_t)pixel_mask)
			XSetFunction(dpy, gc, GXset);
		    else
			XSetFunction(dpy, gc, GXclear);
		    XDrawPoint(dpy, mask, gc, x, y);
		}
	    }    
	    break;

	  case 24:
	    bpp = 3;
	    bpl = img->bytes_per_line;
	    for(y = 0; y < (int)height; y++)
	    {
		for(x = 0; x < (int)width; x++)
		{
		    ptr8 = (u_int8_t *)(data +
			(y * bpl) + (x * bpp)
		    );
		    if(ptr8[0] || ptr8[1] || ptr8[2])
			XSetFunction(dpy, gc, GXset);
		    else  
			XSetFunction(dpy, gc, GXclear);
		    XDrawPoint(dpy, mask, gc, x, y);
		}
	    }
	    break;

	  case 32:
	    bpp = 4;
	    bpl = img->bytes_per_line;
	    for(y = 0; y < (int)height; y++)
	    {
		for(x = 0; x < (int)width; x++)
		{
		    ptr32 = (u_int32_t *)(data +
			(y * bpl) + (x * bpp)
		    );
		    if(*ptr32 & (u_int32_t)pixel_mask)
			XSetFunction(dpy, gc, GXset);
		    else
			XSetFunction(dpy, gc, GXclear);
		    XDrawPoint(dpy, mask, gc, x, y);
		}    
	    }
	    break;
	}

	XFreeGC(dpy, gc);

	return(mask);
}

/*
 *	Gets an image from drawable d.
 */
image_t *OSWGetImage(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	XImage *ximg;
	osw_gui_struct *gui = &osw_gui[0];

	if((d == None) || (width == 0) || (height == 0))
	    return(NULL);

	if(IS_NUM_ODD(width))
	    width += 1;
	if(IS_NUM_ODD(height))
	    height += 1;

	ximg = XGetImage(
	    gui->display,
	    d,
	    x, y,
	    width, height,
	    AllPlanes,
	    ZPixmap
	);
	if(ximg == NULL)
	    return(NULL);

	/* XGetImage() sets the inverse of the pixel masks that
	 * were retrieved but we need to "reinvert" them by setting
	 * them to the values of the specified pixel mask
	 */
	if(ximg->red_mask == 0l)
	    ximg->red_mask = gui->red_mask;
	if(ximg->green_mask == 0l)
	    ximg->green_mask = gui->green_mask;
	if(ximg->blue_mask == 0l)
	    ximg->blue_mask = gui->blue_mask;
 
	return(OSWCreateImageFromXImage(ximg));
}

/*
 *	Puts an image to a drawable.
 */
void OSWPutImageToDrawable(image_t *img, drawable_t d)
{
	XImage *ximg;

	if((img == NULL) || (d == None))
	    return;

	ximg = img->ximage;
	if(ximg == NULL)
	    return;

	XPutImage(
	    osw_gui[0].display,
	    d,
	    osw_gui[0].gc,
	    ximg,
	    0, 0,
	    0, 0,
	    ximg->width, ximg->height
	);
}

/*
 *	Puts image to drawable at a specific position.
 */
void OSWPutImageToDrawablePos(
	image_t *img, drawable_t d,
	int tar_x, int tar_y
)
{
	XImage *ximg;

	if((img == NULL) || (d == None))
	    return;

	ximg = img->ximage;
	if(ximg == NULL)
	    return;

	XPutImage(
	    osw_gui[0].display,
	    d,
	    osw_gui[0].gc,
	    ximg,
	    0, 0,
	    tar_x, tar_y,
	    ximg->width, ximg->height
	);
}

/*
 *	Puts section of image to drawable at a specific position.
 */
void OSWPutImageToDrawableSect(
	image_t *img, drawable_t d,
	int tar_x, int tar_y,
	int src_x, int src_y,
	unsigned int width, unsigned int height
)
{
	XImage *ximg;

	if((img == NULL) || (d == None))
	    return;

	ximg = img->ximage;
	if(ximg == NULL)
	    return;
    
	XPutImage(
	    osw_gui[0].display,
	    d,
	    osw_gui[0].gc,
	    ximg,
	    src_x, src_y,
	    tar_x, tar_y,
	    width, height
	);
}  

/*
 *	Puts a shared image to a drawable.
 *
 *	If USE_XSHM is not defined then this function calls
 *	OSWPutImageToDrawable()
 */
void OSWPutSharedImageToDrawable(shared_image_t *img, drawable_t d)
{
#ifdef USE_XSHM
	XImage *ximg;
	event_t event;
	osw_gui_struct *gui = &osw_gui[0];
	if(gui->def_use_shm)
	{
	    if((gui->display == NULL) || (img == NULL) || (d == None))
	        return;

	    ximg = img->ximage;
	    if(ximg == NULL)
	        return;

	    if(gui->def_sync_shm)
	    {
		gui->last_error = 0;
		XSetErrorHandler(OSWGUISimpleErrorHandler);

		/* Put and sync */
		XShmPutImage(
		    gui->display,
		    d,
		    gui->gc,
		    ximg,
		    0, 0,
		    0, 0,
		    ximg->width, ximg->height,
		    True
		);

		/* Wait untill image is completly put */
		while(!gui->last_error)
		{
		    if(XCheckTypedWindowEvent(
			gui->display,
			d,
			gui->shm_completion_event_code,
			&event
		    ))
		        break;
		}
		XSetErrorHandler(OSWGUIErrorHandler);
	    }
	    else
	    {
		XShmPutImage(
		    gui->display,
		    d,
		    gui->gc,
		    ximg,
		    0, 0,
		    0, 0,
		    ximg->width, ximg->height,
		    False
		);
	    }
	}
	else
	{
#endif	/* USE_XSHM */
	    XImage *ximg;

	    if(img == NULL)
		return;

	    ximg = img->ximage;
	    if(ximg == NULL)
		return;

	    XPutImage(
		osw_gui[0].display,
		d,
		osw_gui[0].gc,
		ximg,
		0, 0,
		0, 0,
		ximg->width, ximg->height
	    );
#ifdef USE_XSHM
	}
#endif /* USE_XSHM */
}


/*
 *      Puts image to drawable at a specific position.
 *
 *      If USE_XSHM is not defined then this function calls
 *      OSWPutImageToDrawablePos()
 */
void OSWPutSharedImageToDrawablePos(
	shared_image_t *img, drawable_t d,
	int tar_x, int tar_y
)
{
#ifdef USE_XSHM
	XImage *ximg;
	event_t event;
	osw_gui_struct *gui = &osw_gui[0];
	if(gui->def_use_shm)
	{ 
	    if((gui->display == NULL) || (img == NULL) || (d == None))
		return;

	    ximg = img->ximage;
	    if(ximg == NULL)
		return;

	    if(gui->def_sync_shm)
	    {
		gui->last_error = 0;
		XSetErrorHandler(OSWGUISimpleErrorHandler);

		/* Put and sync */
		XShmPutImage(
		    gui->display,
		    d,
		    gui->gc,
		    ximg,
		    0, 0,
		    tar_x, tar_y,
		    ximg->width, ximg->height,
		    True
		);

		/* Wait untill image is completly put */
		while(!gui->last_error)
		{   
		    if(XCheckTypedWindowEvent(
			gui->display,
			d,
			gui->shm_completion_event_code,
			&event
		    ))
			break;
		}
		XSetErrorHandler(OSWGUIErrorHandler);
	    }
	    else
	    {
		XShmPutImage(
		    gui->display,
		    d,
		    gui->gc,
		    ximg,
		    0, 0,
		    tar_x, tar_y,
		    ximg->width, ximg->height,
		    False
		);
	    }
	}
	else
	{
#endif	/* USE_XSHM */
	    XImage *ximg;

	    if((img == NULL) || (d == None))
		return;

	    ximg = img->ximage;
	    if(ximg == NULL)
		return;

	    XPutImage(
		osw_gui[0].display,
		d,
		osw_gui[0].gc,
		ximg,
		0, 0,
		tar_x, tar_y,
		ximg->width, ximg->height
	    );
#ifdef USE_XSHM
	}
#endif /* USE_XSHM */
}


/*
 *      Puts section of image to drawable at a specific position.
 *
 *      If USE_XSHM is not defined then this function calls
 *      OSWPutImageToDrawableSect()
 */
void OSWPutSharedImageToDrawableSect(
	shared_image_t *img, drawable_t d,
	int tar_x, int tar_y,
	int src_x, int src_y,
	unsigned int width, unsigned int height
)       
{
#ifdef USE_XSHM
	XImage *ximg;
	event_t event;
	osw_gui_struct *gui = &osw_gui[0];
	if(gui->def_use_shm)
	{
	    if((gui->display == NULL) || (img == NULL) || (d == None))
	        return;

	    ximg = img->ximage;
	    if(ximg == NULL)
		return;

	    if(gui->def_sync_shm)
	    {
		gui->last_error = 0;
		XSetErrorHandler(OSWGUISimpleErrorHandler);

		/* Put and sync */
		XShmPutImage(
		    gui->display,
		    d,
		    gui->gc,
		    ximg,
		    src_x, src_y,
		    tar_x, tar_y,
		    width, height,
		    True
		);

		/* Wait untill image is completly put */
		while(!gui->last_error)
		{
		    if(XCheckTypedWindowEvent(
			gui->display,
			d,
			gui->shm_completion_event_code,
			&event
		    ))
			break;
		}
		XSetErrorHandler(OSWGUIErrorHandler);
	    }
	    else
	    {
		/* Just put, no sync */
		XShmPutImage(
		    gui->display,
		    d,
		    gui->gc,
		    ximg,
		    src_x, src_y,
		    tar_x, tar_y,
		    width, height,
		    False
		);
	    }
	}
	else
	{  
#endif	/* USE_XSHM */
	    XImage *ximg;

	    if((img == NULL) || (d == None))
		return;

	    ximg = img->ximage;
	    if(ximg == NULL)
		return;

	    XPutImage(
		osw_gui[0].display,
		d,
		osw_gui[0].gc,
		ximg,
		src_x, src_y,
		tar_x, tar_y,
		width, height
	    );
#ifdef USE_XSHM
	}
#endif /* USE_XSHM */
}


/*
 *	Syncronizes shared image with current thread of execution.
 *
 *	A short delay may occure due to the time needed to sync.
 */
void OSWSyncSharedImage(shared_image_t *img, drawable_t d)
{
#ifdef USE_XSHM
	XImage *ximg;
	event_t event;
	osw_gui_struct *gui = &osw_gui[0];
	if(gui->def_use_shm)
	{
	    if((gui->display == NULL) || (img == NULL) || (d == None))
		return;

	    ximg = img->ximage;
	    if(ximg == NULL)
	        return;

	    gui->last_error = 0;
	    XSetErrorHandler(OSWGUISimpleErrorHandler);

	    /* Put shared image to drawable (entire size) */
	    XShmPutImage(
		gui->display,
		d,
		gui->gc,
		ximg,
		0, 0,
		0, 0,
		ximg->width,
		ximg->height,
		True
	    );

	    /* Wait until the image has been completely put */
	    while(!gui->last_error)
	    {
		if(XCheckTypedWindowEvent(
		    gui->display,
		    d,
		    gui->shm_completion_event_code,
		    &event
		))
		    break;
	    }
	    XSetErrorHandler(OSWGUIErrorHandler);
	}
#endif /* USE_XSHM */
}


/*
 *	Puts the drawable to the window.
 */
void OSWPutDrawableToWindow(win_t w, drawable_t d)
{
	unsigned int width, height;

	if((w == None) || (d == None))
	    return;

	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);
	OSWCopyDrawables(
	    (drawable_t)w,
	    (drawable_t)d,
	    width, height
	);
}

/*
 *	Copy drawables.
 */
void OSWCopyDrawables(
	drawable_t tar_d, drawable_t src_d,
	unsigned int width, unsigned int height
)
{
	if((tar_d == None) || (src_d == None) ||
	   (width == 0) || (height == 0)
	)
	    return;

	XCopyArea(
	    osw_gui[0].display,
	    src_d, tar_d,
	    osw_gui[0].gc,
	    0, 0, width, height, 0, 0
	);
}

/*
 *	Copy drawables with coordinates.
 */
void OSWCopyDrawablesCoord(
	drawable_t tar_d,
	drawable_t src_d,
	int tar_x, int tar_y,
	unsigned int width, unsigned int height,
	int src_x, int src_y
)
{
	if((tar_d == None) || (src_d == None) ||
	   (width == 0) || (height == 0)
	)
	    return;

	XCopyArea(
	    osw_gui[0].display,
	    src_d, tar_d,
	    osw_gui[0].gc,
	    src_x, src_y,
	    width, height,
	    tar_x, tar_y
	);
}


/*
 *	Sets the graphic context's clip rectangle.
 */
void OSWSetClipRectangle(const rectangle_t *rect)
{
	osw_gui_struct *gui = &osw_gui[0];
	if(rect != NULL)
	{
	    rectangle_t r;
	    memcpy(&r, rect, sizeof(rectangle_t));
	    XSetClipRectangles(
		gui->display,
		gui->gc,
		0, 0,
		&r, 1,
		Unsorted
	    );
	}
	else
	{
	    XSetClipMask(gui->display, gui->gc, None);
	}
}

/*
 *	Sets the graphic context's clip mask.
 */
void OSWSetClipMask(bitmap_t mask)
{
	osw_gui_struct *gui = &osw_gui[0];
	XSetClipMask(gui->display, gui->gc, mask);
}

/*
 *	Sets the graphic context's clip origin.
 */
void OSWSetClipOrigin(int x, int y)
{
	osw_gui_struct *gui = &osw_gui[0];
	XSetClipOrigin(gui->display, gui->gc, x, y);
}

/*
 *	Draws the string.
 */
void OSWDrawString(drawable_t d, int x, int y, const char *s)
{
	if((d == None) || (s == NULL))
	    return;

	XDrawString(
	    osw_gui[0].display, d, osw_gui[0].gc,
	    x, y, s, strlen(s)
	);
}

/*
 *	Draws the string with a limited number of characters.
 */
void OSWDrawStringLimited(
	drawable_t d, int x, int y, const char *s, int len
)
{
	int real_len;

	if((d == None) || (s == NULL) || (len < 1))
	    return;

	real_len = strlen(s);
	if(len > real_len)
	    len = real_len;

	XDrawString(
	    osw_gui[0].display, d, osw_gui[0].gc,
	    x, y, s, len
	);
}

/*
 *	Draw line.
 */
void OSWDrawLine(
	drawable_t d,
	int start_x, int start_y,
	int end_x, int end_y
)
{
	if(d == None)
	    return;

	XDrawLine(
	    osw_gui[0].display, d, osw_gui[0].gc,
	    start_x, start_y,
	    end_x, end_y
	);
}

/*
 *	Draw rectangle.
 */
void OSWDrawRectangle(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	if((d == None) || (width == 0) || (height == 0)) 
	    return;

	XDrawRectangle(
	    osw_gui[0].display, d, osw_gui[0].gc,
	    x, y, width, height
	);
}

/*
 *	Draw solid rectangle.
 */
void OSWDrawSolidRectangle(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	if((d == None) || (width == 0) || (height == 0))
	    return;

	XFillRectangle(
	    osw_gui[0].display, d, osw_gui[0].gc,
	    x, y, width, height
	);
}

/*
 *	Draw arc, angles are in radians.
 */
void OSWDrawArc(
	drawable_t d, int x, int y,
	unsigned int width, unsigned int height,
	float position_angle,			/* In radians */
	float terminal_angle			/* In radians */
)
{
	if((d == None) || (width == 0) || (height == 0))
	    return;

	XDrawArc(
	    osw_gui[0].display, d, osw_gui[0].gc,
	    x, y,
	    width, height,
	    (int)(position_angle * 180.0f / PI * 64.0f),
	    (int)(terminal_angle * 180.0f / PI * 64.0f)
	);
}

/* 
 *	Draw solid arc, angles are in radians.
 */
void OSWDrawSolidArc(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height,
	float position_angle,			/* In radians */
	float terminal_angle			/* In radians */
)
{
	if((d == None) || (width == 0) || (height == 0)) 
	    return;

	XFillArc(
	    osw_gui[0].display, d, osw_gui[0].gc,
	    x, y,
	    width, height,
	    (int)(position_angle * 180.0f / PI * 64.0f),
	    (int)(terminal_angle * 180.0f / PI * 64.0f)
	);
}


/*
 *	GUI Error handler, prints GUI error to stdout and sets
 *	gui->last_error to 0 if there was no error or sets to non-zero
 *	to indicate error.
 */
static int OSWGUIErrorHandler(Display *dpy, XErrorEvent *error_event)
{
	int request_code;
	const char *s;
	osw_gui_struct *gui = &osw_gui[0];

	if((dpy == NULL) || (error_event == NULL))
	    return(0);

	/* Ignore success error code */
	if(error_event->error_code == Success)
	    return(0);

	/* Ignore these errors */
	request_code = error_event->request_code;
	if((request_code == X_GetImage) ||
	   (request_code == X_SetInputFocus) ||
	   (request_code == X_QueryTree)
	)
	    return(0);

	/* Set the error code */
	gui->last_error = 1;


	/* Print error */

	fprintf(
	    stderr,
"\nSubsystems message: X server error:\n\n"
	);

	/* Error type */
	fprintf(
	    stderr,
"        Error type: %i\n", error_event->type
	);

	/* Display */
	fprintf(
	    stderr,
"        X server display connection pointer: %p\n",
	    error_event->display
	);

	/* Serial number */
	fprintf(
	    stderr,
"        Serial number of failed request: %i\n",
	    (int32_t)error_event->serial
	);

	/* Print error code */
	fprintf(
	    stderr,
"        Error code of failed request: %i (",
	    (int32_t)error_event->error_code
	);
	switch(error_event->error_code)
	{
	  case BadRequest:
	    s = "BadRequest";
	    break;
	  case BadValue:
	    s = "BadValue";
	    break;
	  case BadWindow:
	    s = "BadWindow";
	    break;
	  case BadPixmap:
	    s = "BadPixmap";
	    break;
	  case BadAtom:
	    s = "BadAtom";
	    break;
	  case BadCursor:
	    s = "BadCursor";
	    break;
	  case BadFont:
	    s = "BadFont";
	    break;
	  case BadMatch:
	    s = "BadMatch";
	    break;
	  case BadDrawable:
	    s = "BadDrawable"; 
	    break;
	  case BadAccess:
	    s = "BadAccess";
	    break;
	  case BadAlloc:
	    s = "BadAlloc";
	    break;
	  case BadColor:
	    s = "BadColor";
	    break;
	  case BadGC:
	    s = "BadGC";
	    break;
	  case BadIDChoice:
	    s = "BadIDChoice";
	    break;
	  case BadName:
	    s = "BadName";
	    break;
	  case BadLength:
	    s = "BadLength";
	    break;
	  case BadImplementation:
	    s = "BadImplementation";
	    break;
	  default:
	    s = "Unknown";
	    break;
	}
	fprintf(stderr, "%s)\n", s);

	/* Major and minor operation code */
	fprintf(
	    stderr,
"        Major op-code of failed request: %i (", error_event->request_code
	);
	switch(error_event->request_code)
	{
	  case X_CreateWindow:
	    s = "CreateWindow";
	    break;
	  case X_ChangeWindowAttributes:
	    s = "ChangeWindowAttributes";
	    break;
	  case X_GetWindowAttributes:
	    s = "GetWindowAttributes";
	    break;
	  case X_DestroyWindow:
	    s = "DestroyWindow";
	    break;
	  case X_DestroySubwindows:
	    s = "DestroySubwindows";
	    break;
	  case X_ChangeSaveSet:
	    s = "ChangeSaveSet";
	    break;
	  case X_ReparentWindow:
	    s = "ReparentWindow";
	    break;
	  case X_MapWindow:
	    s = "MapWindow";
	    break;
	  case X_MapSubwindows:
	    s = "MapSubwindows";
	    break;
	  case X_UnmapWindow:
	    s = "UnmapWindow";
	    break;
	  case X_UnmapSubwindows:
	    s = "UnmapSubwindows";
	    break;
	  case X_ConfigureWindow:
	    s = "ConfigureWindow";
	    break;
	  case X_CirculateWindow:
	    s = "CirculateWindow";
	    break;
	  case X_GetGeometry:
	    s = "GetGeometry";
	    break;
	  case X_QueryTree:
	    s = "QueryTree";
	    break;
	  case X_InternAtom:
	    s = "InternAtom";
	    break;
	  case X_GetAtomName:
	    s = "GetAtomName";
	    break;
	  case X_ChangeProperty:
	    s = "ChangeProperty";
	    break;
	  case X_DeleteProperty:
	    s = "DeleteProperty";
	    break;
	  case X_GetProperty:
	    s = "GetProperty";
	    break;
	  case X_ListProperties:
	    s = "ListProperties";
	    break;
	  case X_SetSelectionOwner:
	    s = "SetSelectionOwner";
	    break;
	  case X_GetSelectionOwner:
	    s = "GetSelectionOwner";
	    break;
	  case X_ConvertSelection:
	    s = "ConvertSelection";
	    break;
	  case X_SendEvent:
	    s = "SendEvent";
	    break;
	  case X_GrabPointer:
	    s = "GrabPointer";
	    break;
	  case X_UngrabPointer:
	    s = "UngrabPointer";
	    break;
	  case X_GrabButton:
	    s = "GrabButton";
	    break;
	  case X_ChangeActivePointerGrab:
	    s = "ChangeActivePointerGrab";
	    break;
	  case X_GrabKeyboard:
	    s = "GrabKeyboard";
	    break;
	  case X_UngrabKeyboard:
	    s = "UngrabKeyboard";
	    break;
	  case X_GrabKey:
	    s = "GrabKey";
	    break;
	  case X_UngrabKey:
	    s = "UngrabKey";
	    break;
	  case X_AllowEvents:
	    s = "AllowEvents";
	    break;
	  case X_GrabServer:
	    s = "GrabServer";
	    break;
	  case X_UngrabServer:
	    s = "UngrabServer";
	    break;
	  case X_QueryPointer:
	    s = "QueryPointer";
	    break;
	  case X_GetMotionEvents:
	    s = "GetMotionEvents";
	    break;
	  case X_TranslateCoords:
	    s = "TranslateCoords";
	    break;
	  case X_WarpPointer:
	    s = "WarpPointer";
	    break;
	  case X_SetInputFocus:
	    s = "SetInputFocus";
	    break;
	  case X_GetInputFocus:
	    s = "GetInputFocus";
	    break;
	  case X_QueryKeymap:
	    s = "QueryKeymap";
	    break;
	  case X_OpenFont:
	    s = "OpenFont";
	    break;
	  case X_CloseFont:
	    s = "CloseFont";
	    break;
	  case X_QueryFont:
	    s = "QueryFont";
	    break;
	  case X_QueryTextExtents:
	    s = "QueryTextExtents";
	    break;
	  case X_ListFonts:
	    s = "ListFonts";
	    break;
	  case X_ListFontsWithInfo:
	    s = "ListFontsWithInfo";
	    break;
	  case X_SetFontPath:
	    s = "SetFontPath";
	    break;
	  case X_GetFontPath:
	    s = "GetFontPath";
	    break;
	  case X_CreatePixmap:
	    s = "CreatePixmap";
	    break;
	  case X_FreePixmap:
	    s = "FreePixmap";
	    break;
	  case X_CreateGC:
	    s = "CreateGC";
	    break;
	  case X_ChangeGC:
	    s = "ChangeGC";
	    break;
	  case X_CopyGC:
	    s = "CopyGC";
	    break;
	  case X_SetDashes:
	    s = "SetDashes";
	    break;
	  case X_SetClipRectangles:
	    s = "SetClipRectangles";
	    break;
	  case X_FreeGC:
	    s = "FreeGC";
	    break;
	  case X_ClearArea:
	    s = "ClearArea";
	    break;
	  case X_CopyArea:
	    s = "CopyArea";
	    break;
	  case X_CopyPlane:
	    s = "CopyPlane";
	    break;
	  case X_PolyPoint:
	    s = "PolyPoint";
	    break;
	  case X_PolyLine:
	    s = "PolyLine";
	    break;
	  case X_PolySegment:
	    s = "PolySegment";
	    break;
	  case X_PolyRectangle:
	    s = "PolyRectangle";
	    break;
	  case X_PolyArc:
	    s = "PolyArc";
	    break;
	  case X_FillPoly:
	    s = "FillPoly";
	    break;
	  case X_PolyFillRectangle:
	    s = "PolyFillRectangle";
	    break;
	  case X_PolyFillArc:
	    s = "PolyFillArc";
	    break;
	  case X_PutImage:
	    s = "PutImage";
	    break;
	  case X_GetImage:
	    s = "GetImage";
	    break;
	  case X_PolyText8:
	    s = "PolyText8";
	    break;
	  case X_PolyText16:
	    s = "PolyText16";
	    break;
	  case X_ImageText8:
	    s = "ImageText8";
	    break;
	  case X_ImageText16:
	    s = "ImageText16";
	    break;
	  case X_CreateColormap:
	    s = "CreateColormap";
	    break;
	  case X_FreeColormap:
	    s = "FreeColormap";
	    break;
	  case X_CopyColormapAndFree:
	    s = "CopyColormapAndFree";
	    break;
	  case X_InstallColormap:
	    s = "InstallColormap";
	    break;
	  case X_UninstallColormap:
	    s = "UninstallColormap";
	    break;
	  case X_ListInstalledColormaps:
	    s = "ListInstalledColormaps";
	    break;
	  case X_AllocColor:
	    s = "AllocColor";
	    break;
	  case X_AllocNamedColor:
	    s = "AllocNamedColor";
	    break;
	  case X_AllocColorCells:
	    s = "AllocColorCells";
	    break;
	  case X_AllocColorPlanes:
	    s = "AllocColorPlanes";
	    break;
	  case X_FreeColors:
	    s = "FreeColors";
	    break;
	  case X_StoreColors:
	    s = "StoreColors";
	    break;
	  case X_StoreNamedColor:
	    s = "StoreNamedColor";
	    break;
	  case X_QueryColors:
	    s = "QueryColors";
	    break;
	  case X_LookupColor:
	    s = "LookupColor";
	    break;
	  case X_CreateCursor:
	    s = "CreateCursor";
	    break;
	  case X_CreateGlyphCursor:
	    s = "CreateGlyphCursor";
	    break;
	  case X_FreeCursor:
	    s = "FreeCursor";
	    break;
	  case X_RecolorCursor:
	    s = "RecolorCursor";
	    break;
	  case X_QueryBestSize:
	    s = "QueryBestSize";
	    break;
	  case X_QueryExtension:
	    s = "QueryExtension";
	    break;
	  case X_ListExtensions:
	    s = "ListExtensions";
	    break;
	  case X_ChangeKeyboardMapping:
	    s = "ChangeKeyboardMapping";
	    break;
	  case X_GetKeyboardMapping:
	    s = "GetKeyboardMapping";
	    break;
	  case X_ChangeKeyboardControl:
	    s = "ChangeKeyboardControl";
	    break;
	  case X_GetKeyboardControl:
	    s = "GetKeyboardControl";
	    break;
	  case X_Bell:
	    s = "Bell";
	    break;
	  case X_ChangePointerControl:
	    s = "ChangePointerControl";
	    break;
	  case X_GetPointerControl:
	    s = "GetPointerControl";
	    break;
	  case X_SetScreenSaver:
	    s = "SetScreenSaver";
	    break;
	  case X_GetScreenSaver:
	    s = "GetScreenSaver";
	    break;
	  case X_ChangeHosts:
	    s = "ChangeHosts";
	    break;
	  case X_ListHosts:
	    s = "ListHosts";
	    break;
	  case X_SetAccessControl:
	    s = "SetAccessControl";
	    break;
	  case X_SetCloseDownMode:
	    s = "SetCloseDownMode";
	    break;
	  case X_KillClient:
	    s = "KillClient";
	    break;
	  case X_RotateProperties:
	    s = "RotateProperties";
	    break;
	  case X_ForceScreenSaver:
	    s = "ForceScreenSaver";
	    break;
	  case X_SetPointerMapping:
	    s = "SetPointerMapping";
	    break;
	  case X_GetPointerMapping:
	    s = "GetPointerMapping";
	    break;
	  case X_SetModifierMapping:
	    s = "SetModifierMapping";
	    break;
	  case X_GetModifierMapping:
	    s = "GetModifierMapping";
	    break;
	  case X_NoOperation:
	    s = "NoOperation";
	    break;
	  default:	/* Assume unsupported and treat as extended */
	    s = "Unsupported or Extended";
	    break;
	}
	fprintf(stderr, "%s)\n", s);

	fprintf(
	    stderr,
 "        Minor op-code of failed request: %i\n",
	    error_event->minor_code
	);

	/* Resource ID */
	fprintf(
	    stderr,
 "        Resource ID: %i\n",
	    (int32_t)error_event->resourceid
	);

	return(0);
}

/*
 *	GUI Error handler for quick and simple error detection.
 *
 *	Sets gui->last_error to 0 if there was no error or non-zero to
 *	indicate error.
 */
static int OSWGUISimpleErrorHandler(Display *dpy, XErrorEvent *error_event)
{
	osw_gui_struct *gui = &osw_gui[0];
	int request_code;

	if((dpy == NULL) || (error_event == NULL)) 
	    return(0);

	/* Ignore success error code */
	if(error_event->error_code == Success)
	    return(0);

	/* Ignore errors for these types of requests */
	request_code = error_event->request_code;
	if((request_code == X_GetImage) ||
	   (request_code == X_SetInputFocus)
	)
	    return(0);


	/* Set the error code */
	gui->last_error = 1;

	return(0);
}


